/**
 *  Copyright (C) 2005
 *  Sebastian Schaffert and Rupert Westenthaler
 *  Salzburg Research Forschungsgesellschaft m.b.H.
 *
 *  IkeWiki is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU Lesser General Public
 * 	License as published by the Free Software Foundation; either
 * 	version 2 of the License, or (at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * 	Lesser General Public License for more details.
 *
 * 	You should have received a copy of the GNU Lesser General Public
 * 	License along with this library; if not, write to the
 * 	Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * 	Boston, MA 02111-1307, USA.
 * 
 */
package at.srfg.ikewiki.actions;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Collection;
import java.util.Hashtable;

import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import at.srfg.ikewiki.Main;
import at.srfg.ikewiki.data.RequestData;
import at.srfg.ikewiki.data.User;
import at.srfg.ikewiki.data.WikiResource;
import at.srfg.ikewiki.db.DBException;
import at.srfg.ikewiki.layout.perspectives.Perspective;

public class ActionRegistry {

	private static Log log = LogFactory.getLog(ActionRegistry.class);

	private Main main;
	
	
	private static final String[] ACTIONS = {
			"at.srfg.ikewiki.actions.ViewAction",
			"at.srfg.ikewiki.actions.ContextAction",
			"at.srfg.ikewiki.actions.DiscussionAction",
			"at.srfg.ikewiki.actions.DeleteAction",
			"at.srfg.ikewiki.actions.UndeleteAction",
			"at.srfg.ikewiki.actions.EditMozileAction",
			"at.srfg.ikewiki.actions.EditContentAction",
			"at.srfg.ikewiki.actions.EditFormAction",
			"at.srfg.ikewiki.actions.EditMetadataAction",
			"at.srfg.ikewiki.actions.EditNamespacesAction",
			"at.srfg.ikewiki.actions.EditVariablesAction",
			"at.srfg.ikewiki.actions.EditReferencesAction",
			"at.srfg.ikewiki.actions.EditTypesAction",
			"at.srfg.ikewiki.actions.EditDomainAction",
			"at.srfg.ikewiki.actions.EditRangeAction",
			"at.srfg.ikewiki.actions.EditInverseAction",
			"at.srfg.ikewiki.actions.EditSuperClassesAction",
			"at.srfg.ikewiki.actions.EditSuperPropertiesAction",
			"at.srfg.ikewiki.actions.EditUsersAction",
			"at.srfg.ikewiki.actions.EditRolesAction",
			"at.srfg.ikewiki.actions.HistoryAction",
			"at.srfg.ikewiki.actions.LoginAction",
			"at.srfg.ikewiki.actions.LogoutAction",
			"at.srfg.ikewiki.actions.FlushAction",
			"at.srfg.ikewiki.actions.NewResourceAction",
			"at.srfg.ikewiki.actions.NewImageAction",
			"at.srfg.ikewiki.actions.NewClassAction",
			"at.srfg.ikewiki.actions.NewPropertyAction",
			"at.srfg.ikewiki.actions.RecentChangesAction",
			"at.srfg.ikewiki.actions.SearchAction",
			"at.srfg.ikewiki.actions.RelatedAction",
			"at.srfg.ikewiki.actions.SparqlAction",
			"at.srfg.ikewiki.actions.UploadRDFAction"
	};
	
	
	private Hashtable<String,Action> registry;

	
	public ActionRegistry(Main main) {
		super();
		
		long start = System.currentTimeMillis();
		
		this.main = main;
		
		registry = new Hashtable<String,Action>();
		
		for(String name : ACTIONS) {
			registerAction(name);
		}
		
		long end = System.currentTimeMillis();
		
		log.info("ActionRegistry initialised ("+(end-start)+" ms).");
	}
	
	private void registerAction(String classname) {
		try {
			Class cls = Class.forName(classname);
			Constructor cons = cls.getConstructor(Main.class);
			
			Action instance = (Action)cons.newInstance(this.main);
			
			registry.put(instance.getIdentifier(),instance);
			
		} catch(ClassNotFoundException ex) {
			log.error("could not register action "+classname+" (no class definition found)");
		} catch(NoSuchMethodException ex) {
			log.error("could not register action "+classname+" (class definition contains errors)",ex);
		} catch(InstantiationException ex) {
			log.error("could not register action "+classname+" (class is abstract)",ex);
		} catch(IllegalAccessException ex) {
			log.error("could not register action "+classname+ " (access not allowed)",ex);
		} catch(InvocationTargetException ex) {
			log.error("could not register action "+classname+" ("+ex.getCause().getLocalizedMessage()+")");
		}

	}
	
	public Action getAction(String id) {
		return registry.get(id);
	}

	
	public Perspective handleAction(String action_id, RequestData parameters, HttpServletResponse response, HttpSession session) throws DBException {
		Action action = registry.get(action_id);
		if(action != null) {
			
			User user = (User)session.getAttribute("user");
			if(user == null)
				user = main.getPermissionManager().getAnonymousUser();
			
			// check whether all required parameters are set
			for(String s : action.getRequiredParameters()) {
				if(parameters.get(s) == null) {
					return main.getErrorPerspective("required parameter \""+s+"\" not specified.");
				}
			}
			
			// check whether no more than the acceptable parameters are set
			Collection<String> acceptable = action.getAcceptableParameters();
			for(String s : parameters.keySet()) {
				if(!acceptable.contains(s) && !s.equals("action")) {
					log.warn("parameter \""+s+"\" not acceptable for action "+action_id);
				}
			}

			WikiResource ref = null;
			if(parameters.get("title") != null) {
				ref = main.getResourceByQTitle(parameters.get("title"));
			} else {
				ref = main.getDefaultContext();
			}
			if(main.getPermissionManager().hasPermission(user,ref,action)) {
				return action.handleAction(parameters,response,session);
			} else {
				log.error("action "+action.getIdentifier()+": insufficient privileges for user "+user.getLogin());
				return main.getErrorPerspective("insufficient privileges for action \""+action.getIdentifier()+"\".");
			}
		} else {
			log.error("no action handler registered for action \""+action_id+"\"");
			return main.getErrorPerspective("no action handler registered for action \""+action_id+"\". Please ask your system administrator for assistance.");
		}
	}
	
	
	public Collection<Action> listActions() {
		return registry.values();
	}
	

}
