/**
 *  Copyright (C) 2005
 *  Sebastian Schaffert and Rupert Westenthaler
 *  Salzburg Research Forschungsgesellschaft m.b.H.
 *
 *  IkeWiki is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU Lesser General Public
 * 	License as published by the Free Software Foundation; either
 * 	version 2 of the License, or (at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * 	Lesser General Public License for more details.
 *
 * 	You should have received a copy of the GNU Lesser General Public
 * 	License along with this library; if not, write to the
 * 	Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * 	Boston, MA 02111-1307, USA.
 * 
 */
package at.srfg.ikewiki.actions;

import java.net.URISyntaxException;
import java.util.Arrays;
import java.util.Collection;

import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import at.srfg.ikewiki.Main;
import at.srfg.ikewiki.data.Article;
import at.srfg.ikewiki.data.RequestData;
import at.srfg.ikewiki.data.User;
import at.srfg.ikewiki.data.WikiResource;
import at.srfg.ikewiki.db.DBException;
import at.srfg.ikewiki.layout.perspectives.EditContentPerspective;
import at.srfg.ikewiki.layout.perspectives.Perspective;

/**
 * Handle a request for editing page content. When applied, this action will create an EditContentPerspective,
 * and if form data is available either store or preview the data.
 * 
 * @see EditContentPerspective
 * 
 * @author sschaffe
 *
 */
public class EditContentAction extends BaseAction implements Action {

	private static final String[] required_parameters = { "title" };
	private static final String[] acceptable_parameters = 
		{ "title", "submit", "wpSave", "template", "remove-type", "add-type", 
		  "wpSummary", "wpTextbox1", "wpLanguage", "wpURI" };

	/**
	 * @param main
	 */
	public EditContentAction(Main main) {
		super(main);
	}

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.actions.Action#getIdentifier()
	 */
	public String getIdentifier() {
		return "edit_content";
	}

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.actions.Action#getLabel()
	 */
	public String getLabel() {
		return "Edit Content";
	}

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.actions.Action#getAcceptableParameters()
	 */
	public Collection<String> getAcceptableParameters() {
		return Arrays.asList(acceptable_parameters);
	}

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.actions.Action#getRequiredParameters()
	 */
	public Collection<String> getRequiredParameters() {
		return Arrays.asList(required_parameters);
	}
	
	
	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.actions.Action#handleAction(at.srfg.ikewiki.data.RequestData, javax.servlet.http.HttpServletResponse, javax.servlet.http.HttpSession)
	 */
	public Perspective handleAction(RequestData parameters, HttpServletResponse response, HttpSession session) throws DBException {
		// edit the content of the current article
		
		String article_ns    = getArticleNS(parameters);
		String article_title = getArticleTitle(parameters); 

		User user = (User)session.getAttribute("user");

		// TODO: permission handling!
		
		// recognised parameters:
		// - wpPreview displays the edited content in a preview
		// - wpSave    stores the edited content in the database as a new revision
	
		try {
			if(parameters.get("submit") != null) {
				Article a = main.constructArticle(article_ns,article_title,parameters.getStringData());
				if(user != null) {
					a.setCreator(user.getLogin());
				}
				
				EditContentPerspective perspective = new EditContentPerspective(main,a);
				if(parameters.get("wpSave") != null) {
					main.getStorageBackend().storeArticle(a);				
					//perspective.storeArticle();
					
					// apply template (see at.srfg.ikewiki.editing.templates.*)
					if(parameters.get("template") != null) {
						main.getTemplate(parameters.get("template")).applyTemplateRules(a,parameters.getStringData());
					}
					
					// process removed types (needs to be done before added types)
					String[] removetypes = parameters.getList("remove-type");
					if(removetypes != null) {
						for(String t : removetypes) {
							WikiResource type = main.getResourceByQTitle(t);
							main.getStorageBackend().removeType(a,type);
						}
					}
					
					// process added types (see edit_content.jspf and edit_content.js)
					String[] addtypes = parameters.getList("add-type");
					if(addtypes != null) {
						for(String t : addtypes) {
							WikiResource type = main.getResourceByQTitle(t);
							main.getStorageBackend().storeType(a,type);
						}
					}
					
					return main.getRedirectPerspective("index.jsp?action=view&title="+Main.encodeURL(a.getQualifiedTitle()));
				} else {
					return perspective;
				}
			} else {
				return main.getEditContentPerspective(article_ns,article_title);
			}
		} catch(URISyntaxException ex) {
			throw new DBException(ex);
		}
	}

}
