/**
 *  Copyright (C) 2006
 *  Sebastian Schaffert and Rupert Westenthaler
 *  Salzburg Research Forschungsgesellschaft m.b.H.
 *
 *  IkeWiki is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU Lesser General Public
 * 	License as published by the Free Software Foundation; either
 * 	version 2 of the License, or (at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * 	Lesser General Public License for more details.
 *
 * 	You should have received a copy of the GNU Lesser General Public
 * 	License along with this library; if not, write to the
 * 	Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * 	Boston, MA 02111-1307, USA.
 * 
 */
package at.srfg.ikewiki.actions;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.StringWriter;
import java.util.AbstractCollection;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xml.serialize.OutputFormat;
import org.apache.xml.serialize.XMLSerializer;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import at.srfg.ikewiki.Main;
import at.srfg.ikewiki.data.Article;
import at.srfg.ikewiki.data.Namespace;
import at.srfg.ikewiki.data.RequestData;
import at.srfg.ikewiki.db.DBException;
import at.srfg.ikewiki.layout.perspectives.EditFormPerspective;
import at.srfg.ikewiki.layout.perspectives.Perspective;

/**
 * @author sschaffe
 *
 */
public class EditFormAction extends BaseAction implements Action {

	private static final String[] required_parameters = { "title" };
	private static final String[] acceptable_parameters = { "title", "submit" };


	private static Log log = LogFactory.getLog(EditFormAction.class);
	
	/**
	 * @param main
	 */
	public EditFormAction(Main main) {
		super(main);
	}

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.actions.Action#getIdentifier()
	 */
	public String getIdentifier() {
		return "edit_form";
	}

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.actions.Action#getLabel()
	 */
	public String getLabel() {
		return "Edit Form";
	}

	
	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.actions.Action#getAcceptableParameters()
	 */
	public Collection<String> getAcceptableParameters() {
		// accepts all parameters!
		return new AbstractCollection<String>() {

			@Override
			public Iterator<String> iterator() {
				return Arrays.asList(acceptable_parameters).iterator();
			}

			@Override
			public int size() {
				return acceptable_parameters.length;
			}

			// contains every parameter
			@Override
			public boolean contains(Object o) {
				return true;
			}
		};
	}

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.actions.Action#getRequiredParameters()
	 */
	public Collection<String> getRequiredParameters() {
		return Arrays.asList(required_parameters);
	}

	private String xmlToString(Document d) {
		return xmlToString(d,null);
	}
	
	private String xmlToString(Document d, String[] cdata_elements) {
		StringWriter out = new StringWriter();
		try {
			OutputFormat format = new OutputFormat(d);
			format.setOmitXMLDeclaration(true);
			format.setIndenting(true);			
			if(cdata_elements != null) {
				format.setCDataElements(cdata_elements);
			}
			XMLSerializer output = new XMLSerializer(out, format);
			output.setNamespaces(true);		
			output.serialize(d);
		}
		catch (IOException e) {
			log.error(e);
		}
		
		
		return out.toString();
	}
	
	
	private static boolean hasContent(Node n) {
		
		NodeList nl = n.getChildNodes();
		for(int i=0; i<nl.getLength(); i++) {
			if(nl.item(i).getNodeType() == Node.ELEMENT_NODE) {
				return true;
			} else if(nl.item(i).getNodeType() == Node.ATTRIBUTE_NODE) {
				log.debug("attribute: "+nl.item(i).getNodeName());
			}
		}
		
		return false;
	}
	
	
	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.actions.Action#handleAction(at.srfg.ikewiki.data.RequestData, javax.servlet.http.HttpServletResponse, javax.servlet.http.HttpSession)
	 */
	public Perspective handleAction(RequestData parameters,	HttpServletResponse response, HttpSession session) throws DBException {
		String article_ns    = getArticleNS(parameters);
		String article_title = getArticleTitle(parameters); 

		if(parameters.get("submit") != null) {
					
			try {
				DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
				factory.setNamespaceAware(true);
				
				DocumentBuilder xml_parser = factory.newDocumentBuilder();
				

				// retrieve the article corresponding to the form
				Article a = main.getStorageBackend().getArticleByNamespaceTitle(article_ns,article_title);
				
				StringBuffer buf = new StringBuffer();

				// find the <form> parts in the content; if they contain instance data, remove them
				Pattern pat = Pattern.compile("<form.+?</form>",Pattern.DOTALL | Pattern.MULTILINE);
				Matcher m = pat.matcher(a.getText());
			
				while(m.find()) {
					ByteArrayInputStream is = new ByteArrayInputStream(m.group().getBytes());
					
					Document form = xml_parser.parse(is);

					NodeList nl = form.getElementsByTagNameNS(Namespace.NS_XFORMS,"instance");
					for(int i=0; i<nl.getLength(); i++) {
						Node inst_node = nl.item(i);
						
						NodeList childnodes = inst_node.getChildNodes();
						for(int j=0; j<childnodes.getLength(); j++) {
							if(childnodes.item(j).getNodeType() == Node.ELEMENT_NODE &&
							   childnodes.item(j).getLocalName().equals("data")) {
								inst_node.removeChild(childnodes.item(j));
							}
						}
						inst_node.normalize();
						if(!hasContent(inst_node)) {
							if(inst_node.getParentNode() != null) {
								inst_node.getParentNode().removeChild(inst_node);
							}
						}
					}
					
					form.normalize();
					if(!hasContent(form.getDocumentElement())) {
						m.appendReplacement(buf,"");						
					} else {
						m.appendReplacement(buf,xmlToString(form));
					}
					log.debug("removed instance data from existing form");
				}
				m.appendTail(buf);
				
				
				// add to the end of the content a <form> element with just the instance data
				Document instance = xml_parser.newDocument();
				Element el_form = instance.createElementNS(Namespace.NS_IKEWIKI_EXT,"form");
				instance.appendChild(el_form);
				
				Element el_instance = instance.createElementNS(Namespace.NS_XFORMS,"instance");
				el_form.appendChild(el_instance);
				
				Element el_data = instance.createElement("data");
				el_instance.appendChild(el_data);
				
				Vector<String> cdata_elements = new Vector<String>();
				
				for(String key : parameters.keySet()) {
					if(key.startsWith("fv-")) {
						String input_name = key.substring(3);

						log.debug("Form Data: "+input_name+" = "+parameters.get(key));
						
						Element el_value = instance.createElement(input_name);
						el_data.appendChild(el_value);
						for(String v : parameters.getList(key)) {
							el_value.appendChild(instance.createTextNode(v+" "));
							//el_value.appendChild(instance.createCDATASection(v+" "));
						}
						cdata_elements.add(input_name);
					}
				}
				
				String[] _cdata_elements = new String[cdata_elements.size()];
				for(int i = 0; i < cdata_elements.size(); i++) {
					_cdata_elements[i] = cdata_elements.get(i);
				}
				
				buf.append("\n");
				buf.append(xmlToString(instance,_cdata_elements));
				buf.append("\n");
				
				a.setText(buf.toString());

				//log.debug(buf.toString());

				
				main.getStorageBackend().storeArticle(a);
				
				return new EditFormPerspective(main,a);
			} catch(SAXException ex) {
				throw new DBException("form data is not well-formed XML; please change the content",ex);
			} catch(ParserConfigurationException ex) {
				throw new DBException("failed to initialise XML parser",ex);				
			} catch(IOException ex) {
				// should never happen for byte array input stream
				throw new DBException("serious error in XML parser",ex);				
			}
		}
				
		return main.getEditFormPerspective(article_ns,article_title);
	}

}
