/**
 *  Copyright (C) 2006
 *  Sebastian Schaffert and Rupert Westenthaler
 *  Salzburg Research Forschungsgesellschaft m.b.H.
 *
 *  IkeWiki is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU Lesser General Public
 * 	License as published by the Free Software Foundation; either
 * 	version 2 of the License, or (at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * 	Lesser General Public License for more details.
 *
 * 	You should have received a copy of the GNU Lesser General Public
 * 	License along with this library; if not, write to the
 * 	Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * 	Boston, MA 02111-1307, USA.
 * 
 */
package at.srfg.ikewiki.actions;

import java.util.Arrays;
import java.util.Collection;

import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import at.srfg.ikewiki.Main;
import at.srfg.ikewiki.data.Permission;
import at.srfg.ikewiki.data.RequestData;
import at.srfg.ikewiki.data.Role;
import at.srfg.ikewiki.data.User;
import at.srfg.ikewiki.db.DBException;
import at.srfg.ikewiki.layout.perspectives.EditUsersPerspective;
import at.srfg.ikewiki.layout.perspectives.Perspective;

/**
 * @author wastl
 *
 */
public class EditUsersAction extends BaseAction {

	private static final String[] required_parameters = { };
	private static final String[] acceptable_parameters = 
		{ "subaction", "subsubaction", "login", "role", "firstName", "lastName", "email", "submit",
		  "permission", "pattern", "type", "precedence" };

	/**
	 * @param main
	 */
	public EditUsersAction(Main main) {
		super(main);
	}

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.actions.Action#getIdentifier()
	 */
	public String getIdentifier() {
		return "edit_users";
	}

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.actions.Action#getLabel()
	 */
	public String getLabel() {
		return "edit users";
	}

	
	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.actions.Action#getAcceptableParameters()
	 */
	public Collection<String> getAcceptableParameters() {
		return Arrays.asList(acceptable_parameters);
	}

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.actions.Action#getRequiredParameters()
	 */
	public Collection<String> getRequiredParameters() {
		return Arrays.asList(required_parameters);
	}

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.actions.Action#handleAction(at.srfg.ikewiki.data.RequestData, javax.servlet.http.HttpServletResponse, javax.servlet.http.HttpSession)
	 */
	public Perspective handleAction(RequestData parameters, HttpServletResponse response, HttpSession session) throws DBException {
		String subaction = parameters.get("subaction");
		String subsubaction = parameters.get("subsubaction");
		
		if(subaction != null && subaction.equals("edit")) {
			String login = parameters.get("login");
			if(login != null) {
				User user = main.getStorageBackend().getUser(login);
				
				// add a role for the user
				if(subsubaction != null && subsubaction.equals("add_role")) {
					String r_name = parameters.get("role");
					
					if(r_name == null) {
						return main.getErrorPerspective("edit users: role name not given");
					}
					
					Role role = main.getStorageBackend().getRole(r_name);
					
					main.getStorageBackend().assignRole(user,role);
					
				//	retract a role for the user
				} else if(subsubaction != null && subsubaction.equals("remove_role")) {
					String r_name = parameters.get("role");
					
					if(r_name == null) {
						return main.getErrorPerspective("edit users: role name not given");
					}
					
					Role role = main.getStorageBackend().getRole(r_name);
					
					main.getStorageBackend().retractRole(user,role);
					
					
				// update user
				} else if(subsubaction != null && subsubaction.equals("update")) {
					String firstName = parameters.get("firstName");
					String lastName  = parameters.get("lastName");
					String email     = parameters.get("email");
					
					user.setFirstName(firstName);
					user.setLastName(lastName);
					user.setEmail(email);
					
					main.getStorageBackend().updateUser(user);
				
				// add permission
				} else if(subsubaction != null && subsubaction.equals("add_permission")) {
					String p_rpat = parameters.get("resource_pattern");
					String p_apat = parameters.get("action_pattern");
					String p_type = parameters.get("type");
					String p_prec = parameters.get("precedence");									
					String p_desc = parameters.get("description");
					
					if(p_rpat == null || p_apat == null || p_type == null || p_prec == null) {
						return main.getErrorPerspective("add user permission: insufficient parameters (resource pattern, action pattern, precedence, type)");
					}
					
					// parse precedence
					int precedence = 0;
					try {
						precedence = Integer.parseInt(p_prec);
					} catch(NumberFormatException ex) {
						return main.getErrorPerspective("wrong number format for precedence");
					}
					
					// parse type
					Permission.Type type;
					switch(p_type.toLowerCase().charAt(0)) {
					case 'r': type = Permission.Type.RESTRICTION; break;
					case 'p': type = Permission.Type.PERMISSION; break;
					default : type = Permission.Type.NONE; break;
					}
					
					Permission p = new Permission(type,precedence,p_rpat,p_apat,p_desc);
					
					main.getStorageBackend().addPermission(user,p);
					
				
				// remove permission
				} else if(subsubaction != null && subsubaction.equals("remove_permission")) {
					String p_rpat = parameters.get("resource_pattern");
					String p_apat = parameters.get("action_pattern");
					String p_type = parameters.get("type");
					String p_prec = parameters.get("precedence");									
					String p_desc = parameters.get("description");
					
					if(p_rpat == null || p_apat == null || p_type == null || p_prec == null) {
						return main.getErrorPerspective("remove user permission: insufficient parameters (resource pattern, action pattern, precedence, type)");
					}

					// parse precedence
					int precedence = 0;
					try {
						precedence = Integer.parseInt(p_prec);
					} catch(NumberFormatException ex) {
						return main.getErrorPerspective("wrong number format for precedence");
					}

					// parse type
					Permission.Type type;
					switch(p_type.toLowerCase().charAt(0)) {
					case 'r': type = Permission.Type.RESTRICTION; break;
					case 'p': type = Permission.Type.PERMISSION; break;
					default : type = Permission.Type.NONE; break;
					}					
					
					Permission p = new Permission(type,precedence,p_rpat,p_apat,p_desc);
					
					main.getStorageBackend().removePermission(user,p);
				}
				
			} else {
				return main.getErrorPerspective("edit users: user login not given");
			}
			
		} else if(subaction != null && subaction.equals("delete")) {
			String login = parameters.get("login");
			
			if(login != null) {
				User user = main.getStorageBackend().getUser(login);
				main.getStorageBackend().deleteUser(user);
			} else {
				return main.getErrorPerspective("edit users: user login not given");
			}
			
			
		}
		
		
		return new EditUsersPerspective(main);
	}

}
