/**
 *  Copyright (C) 2005
 *  Sebastian Schaffert and Rupert Westenthaler
 *  Salzburg Research Forschungsgesellschaft m.b.H.
 *
 *  IkeWiki is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU Lesser General Public
 * 	License as published by the Free Software Foundation; either
 * 	version 2 of the License, or (at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * 	Lesser General Public License for more details.
 *
 * 	You should have received a copy of the GNU Lesser General Public
 * 	License along with this library; if not, write to the
 * 	Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * 	Boston, MA 02111-1307, USA.
 * 
 */
package at.srfg.ikewiki.db;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.net.URISyntaxException;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.text.Collator;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;

import at.srfg.ikewiki.Main;
import at.srfg.ikewiki.data.Article;
import at.srfg.ikewiki.data.Namespace;
import at.srfg.ikewiki.data.Permission;
import at.srfg.ikewiki.data.Role;
import at.srfg.ikewiki.data.User;
import at.srfg.ikewiki.data.Variable;
import at.srfg.ikewiki.data.WikiException;
import at.srfg.ikewiki.data.WikiImage;
import at.srfg.ikewiki.data.WikiResource;
import at.srfg.ikewiki.transactions.Transaction;
import at.srfg.ikewiki.transactions.TransactionItem;

import com.hp.hpl.jena.rdf.model.Literal;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.NodeIterator;
import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.ResIterator;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Statement;

/**
 * @author sschaffe
 *
 */
public class DBStore extends DBCommon {

	private static Log log = LogFactory.getLog(DBStore.class);

	Connection database = null;
	DocumentBuilderFactory factory;
	DocumentBuilder builder;
	
	public DBStore(Main main) throws DBException {
		super(main);

		factory = DocumentBuilderFactory.newInstance();
		factory.setNamespaceAware(true);
		
		try {
			builder = factory.newDocumentBuilder();
		} catch(ParserConfigurationException ex) {
			log.error("error while initialising XML parser",ex);
			throw new DBException(ex);
		}
		
		// verify whether database connection succeeds
		getDatabaseConnection();
	}
	
	
	public void close() {
		try {
			if(database != null && !database.isClosed()) {
				database.close();
			}
		} catch(SQLException ex) {
			log.error("error while closing database connection",ex);
		}
	}
	
	
	public Connection getDatabaseConnection() throws DBException {
		try {
			if(database == null || database.isClosed()) {
				database = DriverManager.getConnection(super.getConnectionString(),DB_USER,DB_PASS);
			}
			return database;
		} catch(SQLException ex) {
			throw new DBException("Error while connecting to database",ex);
		}
	}
	

	/**
	 * Retrieve an article identified by namespace and title from the database.
	 * 
	 * @param namespace
	 * @param title
	 * @return
	 * @throws DBException in case the article cannot be retrieved or the content of the article is invalid
	 */
	private int getArticleIdByNamespaceTitle(String ns_prefix, String title) throws DBException {
		try {
			if(stmt_exists_article_namespace_title == null) {
				String sql_string = 
						"SELECT P.id " +
						"FROM pages P, revisions R, namespaces N " +
						"WHERE N.prefix = ? " +
						"  AND P.namespace = N.id "+
						"  AND P.title = ? " +
						"  AND P.cur_revision = R.id "+
						"  AND R.deleted = FALSE";				
				stmt_exists_article_namespace_title = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_exists_article_namespace_title.clearParameters();
			stmt_exists_article_namespace_title.setString(1,ns_prefix);
			stmt_exists_article_namespace_title.setString(2,title);
			
			ResultSet r = stmt_exists_article_namespace_title.executeQuery();
			if(r.next()) {
				return r.getInt("id");				
			} else {
				return -1;				
			}
			
		} catch(SQLException ex) {
			throw new DBException("Database Error while retrieving document from database",ex);
		}
	}

	
	/**
	 * Retrieve an article identified by namespace and title from the database.
	 * 
	 * @param namespace
	 * @param title
	 * @return
	 * @throws DBException in case the article cannot be retrieved or the content of the article is invalid
	 */
	public boolean existsArticleByNamespaceTitle(String ns_prefix, String title) throws DBException {
		return getArticleIdByNamespaceTitle(ns_prefix,title) > -1;
	}
	
	
	/**
	 * Retrieve an article identified by namespace and title from the database.
	 * 
	 * @param namespace
	 * @param title
	 * @return
	 * @throws DBException in case the article cannot be retrieved or the content of the article is invalid
	 */
	public Article getArticleByNamespaceTitle(String ns_prefix, String title) throws DBException {
		try {
			if(stmt_get_article_namespace_title == null) {
				String sql_string = 
						"SELECT revision_id, ns_prefix, ns_uri, creator, title, uri, timestamp, comment, language, text, text_xml " +
						"FROM current_page " +
						"WHERE ns_prefix = ? " +
						"  AND title = ?";				
				stmt_get_article_namespace_title = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_get_article_namespace_title.clearParameters();
			stmt_get_article_namespace_title.setString(1,ns_prefix);
			stmt_get_article_namespace_title.setString(2,title);
			
			ResultSet r = stmt_get_article_namespace_title.executeQuery();
			if(r.next()) {
				Article a = new Article(new Namespace(r.getString("ns_prefix").trim(),r.getString("ns_uri").trim()),
							            r.getString("title").trim(),
										r.getString("comment").trim(),r.getString("text").trim(),
										r.getString("language"),r.getTimestamp("timestamp"));
				a.setRevisionId(r.getInt("revision_id"));
				a.setCreator(r.getString("creator"));
				a.setRenderer(main.getConfig().getString("render.engine","at.srfg.ikewiki.layout.engine.WikipediaRenderEngine"));
				
				if(r.getString("text_xml") != null) {
					a.setBaseDom(parseXML(r.getString("text_xml")));
				}
				return a;
			} else {
				throw new ArticleNotFoundException("Article with namespace '"+ns_prefix+"' and title '"+title+"' does not exist in the database");
			}
			
		} catch(SQLException ex) {
			throw new DBException("Database Error while retrieving document from database",ex);
		}
	}


	
	public List<Article> getRevisionsByNamespaceTitle(String ns_prefix, String title, int limit, int offset) throws DBException {
		try {
			if(stmt_get_revisions_namespace_title == null) {
				String sql_string = 
						"SELECT revision_id, ns_prefix, ns_uri, creator, title, uri, timestamp, comment, language, text, text_xml " +
						"FROM page_revisions " +
						"WHERE ns_prefix = ? " +
						"  AND title = ? " +
						"ORDER BY revision_id DESC "+
						"LIMIT ? OFFSET ?";				
				stmt_get_revisions_namespace_title = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_get_revisions_namespace_title.clearParameters();
			stmt_get_revisions_namespace_title.setString(1,ns_prefix);
			stmt_get_revisions_namespace_title.setString(2,title);
			stmt_get_revisions_namespace_title.setInt(3,limit);
			stmt_get_revisions_namespace_title.setInt(4,offset);
			
			ResultSet r = stmt_get_revisions_namespace_title.executeQuery();
			
			
			List<Article> result = new ArrayList<Article>();
			while(r.next()) {
				Article a = new Article(new Namespace(r.getString("ns_prefix").trim(),r.getString("ns_uri").trim()),
							            r.getString("title").trim(),
										r.getString("comment").trim(),r.getString("text").trim(),
										r.getString("language"),r.getTimestamp("timestamp"));
				a.setRevisionId(r.getInt("revision_id"));
				a.setCreator(r.getString("creator"));
				a.setRenderer(main.getConfig().getString("render.engine","at.srfg.ikewiki.layout.engine.WikipediaRenderEngine"));
				if(r.getString("text_xml") != null) {
					a.setBaseDom(parseXML(r.getString("text_xml")));
				}
				result.add(a);
			}
			log.debug("retrieved "+result.size()+" revisions of "+title+" LIMIT " + limit + " OFFSET "+ offset);
			return result;
			
		} catch(SQLException ex) {
			throw new DBException("Database Error while retrieving revisions from database",ex);
		}
		
	}
	
	public Article getArticleByRevisionId(String ns_prefix, String title, long revision_id) throws DBException {
		try {
			if(stmt_get_article_revision == null) {
				String sql_string = 
						"SELECT revision_id, ns_prefix, ns_uri, creator,title, uri, timestamp, comment, language, text, text_xml " +
						"FROM page_revisions " +
						"WHERE ns_prefix = ? " +
						"  AND title = ? " +
						"  AND revision_id = ? ";				
				stmt_get_article_revision = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_get_article_revision.clearParameters();
			stmt_get_article_revision.setString(1,ns_prefix);
			stmt_get_article_revision.setString(2,title);
			stmt_get_article_revision.setLong(3,revision_id);
			
			ResultSet r = stmt_get_article_revision.executeQuery();
			
			
			if(r.next()) {
				Article a = new Article(new Namespace(r.getString("ns_prefix").trim(),r.getString("ns_uri").trim()),
							            r.getString("title").trim(),
										r.getString("comment").trim(),r.getString("text").trim(),
										r.getString("language"),r.getTimestamp("timestamp"));
				a.setRevisionId(r.getInt("revision_id"));
				a.setCreator(r.getString("creator"));
				a.setRenderer(main.getConfig().getString("render.engine","at.srfg.ikewiki.layout.engine.WikipediaRenderEngine"));
				if(r.getString("text_xml") != null) {
					a.setBaseDom(parseXML(r.getString("text_xml")));
				}
				log.debug("retrieved revision "+revision_id+" of article "+ns_prefix+":"+title);
				return a;
			} else {
				log.error("revision "+revision_id+" of article "+ns_prefix+":"+title+" does not exist in the database");
				throw new ArticleNotFoundException("revision "+revision_id+" of article "+ns_prefix+":"+title+" does not exist in the database");				
			}
			
		} catch(SQLException ex) {
			throw new DBException("Database Error while retrieving revisions from database",ex);
		}
		
	}
	
	
	/**
	 * Store an article in the database.
	 * 
	 * In case the article already exists, a new revision is created.
	 * 
	 * @param a
	 */
	public void storeArticle(Article a) throws DBException {
		try {
			// undelete article in case it has been deleted previously ...
			try {
				undeleteArticle(a);
			} catch(DBException ex) {
				log.error("error while restoring previous revisions of "+a.getQualifiedTitle(),ex);
			}
			
			// first look whether the page already exists in the database; 
			int article_id;
			if(!existsArticleByNamespaceTitle(a.getNamespace().getPrefix(),a.getTitle())) {
				// page does not exist in the database; add a complete new entry
				
				// create uri entry in URI table if not yet there
				ensureURI(a.getUri().toString());
				
				// create new page descriptor in pages table
				if(stmt_store_article_page == null) {
					String sql_string = 
						"INSERT INTO pages (id,namespace,title,creator) VALUES (nextval('page_ids'),(SELECT id FROM namespaces WHERE prefix=?),?,(SELECT id FROM users WHERE login = ?))";
					stmt_store_article_page = getDatabaseConnection().prepareStatement(sql_string);
				}
				stmt_store_article_page.clearParameters();
				stmt_store_article_page.setString(1,a.getNamespace().getPrefix());
				stmt_store_article_page.setString(2,a.getTitle());
				if(a.getCreator() != null) {
					stmt_store_article_page.setString(3,a.getCreator());
					log.debug("storing article "+a.getQualifiedTitle()+" with user information "+a.getCreator());
				} else {
					stmt_store_article_page.setString(3,null);
				}
				stmt_store_article_page.executeUpdate();
				
				// retrieve last page id created in this session
				if(stmt_store_article_lastid == null) {
					String sql_string = 
						"SELECT currval('page_ids')";
					stmt_store_article_lastid = getDatabaseConnection().prepareStatement(sql_string);
				}
				ResultSet r2 = stmt_store_article_lastid.executeQuery();
				if(r2.next()) {
					article_id = r2.getInt(1);
				} else {
					throw new DBException("Serious Database Error: no sequence value found.");
				}
				log.debug("stored article "+a.getQualifiedTitle());
				
			} else {
				// page already exists, simply use the article id
				
				article_id = getArticleIdByNamespaceTitle(a.getNamespace().getPrefix(),a.getTitle());
			}

			
			// first insert text into texts table
			if(stmt_update_article_text == null) {
				String sql_string = 
					"INSERT INTO texts (id,language,text,text_xml) VALUES (nextval('text_ids'),?,?,?)";
				stmt_update_article_text = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_update_article_text.clearParameters();
			stmt_update_article_text.setString(1,a.getLanguage().getLanguage());
			stmt_update_article_text.setString(2,a.getText());
			
			try {
				main.getRenderingPipeline().createDOM(a);
				stmt_update_article_text.setString(3,a.getXML());
			} catch(WikiException ex) {
				log.error("error transforming XML to String",ex);
				stmt_update_article_text.setString(3,null);
			}
			
			stmt_update_article_text.executeUpdate();
			
			// then insert new revision into revisions table
			if(stmt_update_article_revision == null) {
				String sql_string =
					"INSERT INTO revisions (page,text,comment,timestamp,creator) " +
					"VALUES (?,currval('text_ids'),?,?,(SELECT id FROM users WHERE login = ?))";
				stmt_update_article_revision = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_update_article_revision.clearParameters();
			stmt_update_article_revision.setInt(1,article_id);
			stmt_update_article_revision.setString(2,a.getComment());
			stmt_update_article_revision.setTimestamp(3,new Timestamp(a.getTimestamp().getTime()));
			stmt_update_article_revision.setString(4,a.getCreator());
			stmt_update_article_revision.executeUpdate();
				
				
		} catch(SQLException ex) {
			throw new DBException("Database Error while retrieving document from database",ex);
		}
		
	}

	
	/**
	 * Mark an article in the database as deleted. 
	 * 
	 * This method will not really delete the article from the database, it will just set the 
	 * "deleted" field of all revisions to "true". Such articles can be restored via undeleteArticle().
	 * 
	 * The database can be cleaned up using the method purgeDatabase(). All articles marked as "deleted"
	 * will be removed then.
	 * 
	 * 
	 * @param a
	 */
	public void deleteArticle(WikiResource a) throws DBException {
		try {
			// first look whether the page already exists in the database; 
			int article_id;
			if(!existsArticleByNamespaceTitle(a.getNamespace().getPrefix(),a.getTitle())) {

				throw new DBException("Could not delete article. "+a.getQualifiedTitle()+" does not exist in the database.");
			} else {
				// page already exists, simply use the article id
				
				article_id = getArticleIdByNamespaceTitle(a.getNamespace().getPrefix(),a.getTitle());
			}

			
			
			// first insert text into texts table (required reference by DB Schema)
			
			// then insert new revision into revisions table
			if(stmt_delete_article == null) {
				String sql_string =
					"UPDATE revisions SET deleted = TRUE where page = ?";
				stmt_delete_article = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_delete_article.clearParameters();
			stmt_delete_article.setInt(1,article_id);
			stmt_delete_article.executeUpdate();

			log.info("deleted article "+a.getQualifiedTitle());
				
		} catch(SQLException ex) {
			throw new DBException("Database Error while retrieving document from database",ex);
		}
		
	}

	
	/**
	 * Mark an article in the database as deleted. 
	 * 
	 * This method will not really delete the article from the database, it will just set the 
	 * "deleted" field of all revisions to "true". Such articles can be restored via undeleteArticle().
	 * 
	 * The database can be cleaned up using the method purgeDatabase(). All articles marked as "deleted"
	 * will be removed then.
	 * 
	 * 
	 * @param a
	 */
	public void undeleteArticle(WikiResource a) throws DBException {
		try {
			
			// update relevant revisions
			if(stmt_undelete_article == null) {
				String sql_string =
					"UPDATE revisions SET deleted = FALSE "+
					"WHERE page IN (SELECT pages.id FROM pages, namespaces "+
											"WHERE pages.namespace = namespaces.id "+ 
											"AND   namespaces.prefix = ? "+
											"AND   title = ?" +
											"AND   deleted = TRUE) ";
				stmt_undelete_article = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_undelete_article.clearParameters();
			stmt_undelete_article.setString(1,a.getNamespace().getPrefix());
			stmt_undelete_article.setString(2,a.getTitle());
			int r = stmt_undelete_article.executeUpdate();

			if( r > 0 ) {
				log.info("restored article "+a.getQualifiedTitle());
			}
		} catch(SQLException ex) {
			throw new DBException("Database Error while retrieving document from database",ex);
		}
		
	}
	
	public Iterable<Article> listArticles() throws DBException {
		try {
			if(stmt_list_articles == null) {
				String sql_string = 
						"SELECT revision_id, ns_prefix, ns_uri, creator, title, uri, timestamp, comment, language, text, text_xml " +
						"FROM current_page ";
				stmt_list_articles = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_list_articles.clearParameters();
			
			final ResultSet r = stmt_list_articles.executeQuery();
			final boolean myHasNext = r.next();
			
			return new Iterable<Article>() {
				public Iterator<Article> iterator() {
					return new Iterator<Article>() {

						private boolean _hasNext = myHasNext;
						
						public boolean hasNext() {
							return _hasNext;
						}

						public Article next() {
							try {
								Article a = new Article(new Namespace(r.getString("ns_prefix").trim(),r.getString("ns_uri").trim()),
										r.getString("title").trim(),
										r.getString("comment").trim(),r.getString("text").trim(),
										r.getString("language"),r.getTimestamp("timestamp"));
								a.setRevisionId(r.getInt("revision_id"));
								a.setCreator(r.getString("creator"));
								a.setRenderer(main.getConfig().getString("render.engine","at.srfg.ikewiki.layout.engine.WikipediaRenderEngine"));
								if(r.getString("text_xml") != null) {
									a.setBaseDom(parseXML(r.getString("text_xml")));
								}
								
								_hasNext = r.next();
								
								return a;
							} catch(SQLException ex) {
								log.error("database error while listing articles",ex);
								return null;
							}
						}

						public void remove() {
						}
						
					};
				}
				
			};
			
		} catch(SQLException ex) {
			log.error("database error while listing articles",ex);
			throw new DBException("database error while listing articles",ex);
		}
		
	}
	
	public List<Article> searchArticles(String keyword, int limit, int offset) throws DBException {
		try {
			long start = System.currentTimeMillis();
			if(stmt_search_article == null) {
				String sql_string = 
						"SELECT revision_id, ns_prefix, ns_uri, creator, title, uri, timestamp, comment, language, text, text_xml " +
						"FROM current_page " +
						"WHERE text ILIKE ? " +
						"   OR (ns_prefix || ':' || title) ILIKE ? "+
						"ORDER BY f_rank_page(?,(ns_prefix || ':' || title),text) DESC "+
						"LIMIT ? OFFSET ?";
				stmt_search_article = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_search_article.clearParameters();
			stmt_search_article.setString(1,"%"+keyword+"%");
			stmt_search_article.setString(2,"%"+keyword+"%");
			stmt_search_article.setString(3,keyword);
			stmt_search_article.setInt(4,limit);
			stmt_search_article.setInt(5,offset);
			
			ResultSet r = stmt_search_article.executeQuery();
			
			
			List<Article> result = new ArrayList<Article>();
			while(r.next()) {
				Article a = new Article(new Namespace(r.getString("ns_prefix").trim(),r.getString("ns_uri").trim()),
							            r.getString("title").trim(),
										r.getString("comment").trim(),r.getString("text").trim(),
										r.getString("language"),r.getTimestamp("timestamp"));
				a.setRevisionId(r.getInt("revision_id"));
				a.setCreator(r.getString("creator"));
				a.setRenderer(main.getConfig().getString("render.engine","at.srfg.ikewiki.layout.engine.WikipediaRenderEngine"));
				if(r.getString("text_xml") != null) {
					a.setBaseDom(parseXML(r.getString("text_xml")));
				}
				main.getStorageBackend().cacheArticle(a);
				result.add(a);
			}
			long end = System.currentTimeMillis();
			log.debug("retrieved "+result.size()+" articles matching "+keyword + " in "+(end-start)+" ms");
			return result;
			
		} catch(SQLException ex) {
			log.error("database error while searching for articles in database",ex);
			throw new DBException("Database Error while searching for articles in database",ex);
		}
		
	}
	
	public List<Article> listRecentChanges(int limit, int offset) throws DBException {
		try {
			long start = System.currentTimeMillis();
			if(stmt_get_recent == null) {
				String sql_string = 
						"SELECT revision_id, ns_prefix, ns_uri, creator, title, uri, timestamp, comment, language, text, text_xml " +
						"FROM current_page " +
						"ORDER BY timestamp DESC "+
						"LIMIT ? OFFSET ?";
				stmt_get_recent = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_get_recent.clearParameters();
			stmt_get_recent.setInt(1,limit);
			stmt_get_recent.setInt(2,offset);
			
			ResultSet r = stmt_get_recent.executeQuery();
			
			
			List<Article> result = new ArrayList<Article>();
			while(r.next()) {
				Article a = new Article(new Namespace(r.getString("ns_prefix").trim(),r.getString("ns_uri").trim()),
							            r.getString("title").trim(),
										r.getString("comment").trim(),r.getString("text").trim(),
										r.getString("language"),r.getTimestamp("timestamp"));
				a.setRevisionId(r.getInt("revision_id"));
				a.setCreator(r.getString("creator"));
				a.setRenderer(main.getConfig().getString("render.engine","at.srfg.ikewiki.layout.engine.WikipediaRenderEngine"));
				if(r.getString("text_xml") != null) {
					a.setBaseDom(parseXML(r.getString("text_xml")));
				}				
				main.getStorageBackend().cacheArticle(a);
				result.add(a);
			}
			long end = System.currentTimeMillis();
			log.debug("retrieved "+result.size()+" recent articles in "+(end-start)+" ms");
			return result;
			
		} catch(SQLException ex) {
			log.error("database error while searching for articles in database",ex);
			throw new DBException("Database Error while searching for articles in database",ex);
		}
		
	}
	
	/**
	 * Add a URI to the database
	 * @param uri
	 * @throws DBException
	 */
	private void addURI(String uri) throws DBException {
		try {
			if(stmt_store_uri == null) {
				String sql_string = "INSERT INTO uris (uri) VALUES (?)";
				stmt_store_uri = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_store_uri.clearParameters();
			stmt_store_uri.setString(1,uri.toString());
			stmt_store_uri.executeUpdate();
			
		} catch(SQLException ex) {
			throw new DBException("Database Error while adding URI to database",ex);
		}
	}
	
	
	private boolean hasURI(String uri) throws DBException {
		try {
			if(stmt_get_uri == null) {
				String sql_string = "SELECT uri FROM uris WHERE uri = ?";
				stmt_get_uri = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_get_uri.clearParameters();
			stmt_get_uri.setString(1,uri.toString());
			ResultSet s = stmt_get_uri.executeQuery();
			return s.next();			
		} catch(SQLException ex) {
			throw new DBException("Database Error while adding URI to database",ex);
		}
		
	}
	
	private void ensureURI(String uri) throws DBException {
		if(!hasURI(uri)) {
			addURI(uri);
		}
	}
	
	
	public Namespace getNamespaceByPrefix(String ns_prefix) throws DBException {
		try {
			if(stmt_get_ns_prefix == null) {
				String sql_string = 
						"SELECT uri FROM namespaces " +
						"WHERE prefix = ? ";				
				stmt_get_ns_prefix = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_get_ns_prefix.clearParameters();
			stmt_get_ns_prefix.setString(1,ns_prefix);
			
			ResultSet r = stmt_get_ns_prefix.executeQuery();
			if(r.next()) {
				return new Namespace(ns_prefix,r.getString("uri"));				
			} else {
				throw new NamespaceNotFoundException("Namespace prefix '"+ns_prefix+"' does not exist in database; you need to create it before you can add articles to it.");				
			}
			
		} catch(SQLException ex) {
			throw new DBException("Database error while looking up namespace",ex);
		}
		
	}

	public Namespace getNamespaceByUri(String uri) throws DBException {
		try {
			if(stmt_get_ns_uri == null) {
				String sql_string = 
						"SELECT prefix FROM namespaces " +
						"WHERE uri = ? ";				
				stmt_get_ns_uri = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_get_ns_uri.clearParameters();
			stmt_get_ns_uri.setString(1,uri);
			
			ResultSet r = stmt_get_ns_uri.executeQuery();
			if(r.next()) {
				return new Namespace(r.getString("prefix"),uri);				
			} else {
				throw new NamespaceNotFoundException("Namespace uri '"+uri+"' does not exist in database; you need to create it before you can add articles to it.");				
			}
			
		} catch(SQLException ex) {
			throw new DBException("Database error while looking up namespace",ex);
		}
		
	}


	/**
	 * Add a URI to the database
	 * @param uri
	 * @throws DBException
	 */
	private void addNamespace(Namespace n) throws DBException {
		try {
			if(stmt_store_ns == null) {
				String sql_string = "INSERT INTO namespaces (prefix,uri) VALUES (?,?)";
				stmt_store_ns = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_store_ns.clearParameters();
			stmt_store_ns.setString(1,n.getPrefix());
			stmt_store_ns.setString(2,n.getUri());
			stmt_store_ns.executeUpdate();
			
		} catch(SQLException ex) {
			log.error("database error while adding namespace "+n,ex);
			throw new DBException("database error while adding namespace "+n,ex);
		}
	}
	
	
	private boolean hasNamespace(Namespace n) throws DBException {
		try {
			if(stmt_get_ns == null) {
				String sql_string = "SELECT prefix,uri FROM namespaces WHERE uri = ? AND prefix = ?";
				stmt_get_ns = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_get_ns.clearParameters();
			stmt_get_ns.setString(1,n.getUri());
			stmt_get_ns.setString(2,n.getPrefix());
			ResultSet s = stmt_get_ns.executeQuery();
			
			return s.next();
		
		} catch(SQLException ex) {
			log.error("database error while retrieving namespace",ex);
			throw new DBException("database error while retrieving namespace",ex);
		}
		
	}

	private boolean hasNamespacePrefix(String ns_prefix) throws DBException {
		try {
			if(stmt_get_ns_prefix == null) {
				String sql_string = 
						"SELECT uri FROM namespaces " +
						"WHERE prefix = ? ";				
				stmt_get_ns_prefix = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_get_ns_prefix.clearParameters();
			stmt_get_ns_prefix.setString(1,ns_prefix);
			
			ResultSet r = stmt_get_ns_prefix.executeQuery();

			return r.next();
			
		} catch(SQLException ex) {
			throw new DBException("Database error while looking up namespace",ex);
		}
		
	}

	private boolean hasNamespaceUri(String uri) throws DBException {
		try {
			if(stmt_get_ns_uri == null) {
				String sql_string = 
						"SELECT prefix FROM namespaces " +
						"WHERE uri = ? ";				
				stmt_get_ns_uri = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_get_ns_uri.clearParameters();
			stmt_get_ns_uri.setString(1,uri);
			
			ResultSet r = stmt_get_ns_uri.executeQuery();
			
			return r.next();
			
		} catch(SQLException ex) {
			throw new DBException("Database error while looking up namespace",ex);
		}
		
	}
	
	
	
	/**
	 * Ensure that a namespace exists.
	 * 
	 * This method tries to handle errors gracefully:
	 * - if neither the prefix nor the URI exist, a new namespace is added and the prefix is returned
	 * - if the uri already exists with a different prefix, the prefix is returned
	 * - if the prefix already exists with a different URI, a new prefix is generated and added
	 * 
	 * @param n
	 * @return the prefix of the namespace (possibly different to the provided prefix)
	 * @throws DBException
	 */
	public String ensureNamespace(Namespace n) throws DBException {
		if(hasNamespace(n)) {
			// namespace already exists
			return n.getPrefix();
		} else if(hasNamespaceUri(n.getUri())) {
			// retrieve the prefix of the already existing URI
			Namespace p = getNamespaceByUri(n.getUri());
			return p.getPrefix();
		} else if(hasNamespacePrefix(n.getPrefix())) {
			// generate a new prefix and try with that
			Pattern p = Pattern.compile("(.*?)([0-9]+)$");
			Matcher m = p.matcher(n.getPrefix());
			if(m.matches()) {
				String real_prefix = m.group(1);
				int counter;
				try {
					counter = Integer.parseInt(m.group(2));
				} catch(NumberFormatException ex) {
					log.error("error, could not parse number '"+m.group(2)+"'",ex);
					counter = 0;
				}
				counter ++;
				return ensureNamespace(new Namespace(real_prefix+counter,n.getUri()));				
			} else if(n.getPrefix().length() == 0){				
				return ensureNamespace(new Namespace("ns1",n.getUri()));
			} else {
				return ensureNamespace(new Namespace(n.getPrefix()+"1",n.getUri()));
			}
			
		} else {
			// add the namespace, neither the uri nor the prefix already exist
			addNamespace(n);
			return n.getPrefix();			
		}
	}

	/**
	 * Modify a namespace mapping in the database
	 * @param uri
	 * @throws DBException
	 */
	public void updateNamespace(Namespace n) throws DBException {
		try {
			if(stmt_upd_ns == null) {
				String sql_string = "UPDATE namespaces SET prefix = ? WHERE uri = ?";
				stmt_upd_ns = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_upd_ns.clearParameters();
			stmt_upd_ns.setString(1,n.getPrefix());
			stmt_upd_ns.setString(2,n.getUri());
			stmt_upd_ns.executeUpdate();
			
		} catch(SQLException ex) {
			log.error("database error while updating namespace in database",ex);
			throw new DBException("database error while updating namespace in database",ex);
		}
	}
	
	/**
	 * Delete a namespace mapping from the database
	 * @param uri
	 * @throws DBException
	 */
	public void deleteNamespace(String uri) throws DBException {
		try {
			if(stmt_del_ns == null) {
				String sql_string = "DELETE FROM namespaces WHERE uri = ?";
				stmt_del_ns = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_del_ns.clearParameters();
			stmt_del_ns.setString(1,uri);
			stmt_del_ns.executeUpdate();
			
		} catch(SQLException ex) {
			log.error("database error while deleting namespace from database",ex);
			throw new DBException("database error while deleting namespace from database",ex);
		}
	}
	
	
	/**
	 * Retrieve a list of all namespaces.
	 * 
	 * @return
	 * @throws DBException
	 */
	public Collection<Namespace> listNamespaces() throws DBException {
		try {
			if(stmt_get_namespaces == null) {
				String sql_string = 
						"SELECT uri, prefix FROM namespaces";				
				stmt_get_namespaces = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_get_namespaces.clearParameters();
			
			ResultSet r = stmt_get_namespaces.executeQuery();
			
			TreeSet<Namespace> result = new TreeSet<Namespace>();
			
			while(r.next()) {
				result.add(new Namespace(r.getString("prefix"),r.getString("uri")));
			}

			return result;
			
		} catch(SQLException ex) {
			log.error("database error while looking up namespace",ex);
			throw new DBException("database error while looking up namespace",ex);
		}			
	}
	
	/**
	 * Store a RDF graph completely in the database. Only non-existing triples are added.
	 * 
	 * Needs to be refined to delete missing triples.
	 * 
	 * @param g
	 */
	@Deprecated
	public void storeTriples(Model m, User u) throws DBException {
		
		try {
			//getDatabaseConnection().setAutoCommit(false);
			
			// first ensure that all namespaces exist
			Map namespaces = m.getNsPrefixMap();
			for(Object op : namespaces.keySet()) {
				String prefix = (String)op;
				String uri    = (String)namespaces.get(prefix);
				String nprefix = ensureNamespace(new Namespace(prefix,uri));
				if(!nprefix.equals(prefix)) {
					// prefix has changed, update model
					m.removeNsPrefix(prefix);
					m.setNsPrefix(nprefix,uri);
					log.info("updated namespace prefix '"+prefix+"' to '"+nprefix+"'");
				}
			}
			
			//m.commit();
			
			// finally, create a page entry for every resource in the model
			// use dc:description as article content and rdfs:comment as article comment
			Property rdfs_comment = m.createProperty("http://www.w3.org/2000/01/rdf-schema#","comment");
			Property dc_description = m.createProperty("http://purl.org/dc/elements/1.1/description");
			ResIterator ri = m.listSubjects();
			while(ri.hasNext()) {
				Resource r = ri.nextResource();
				
				String content="", comment="";
				NodeIterator n = m.listObjectsOfProperty(r,dc_description);
				while(n.hasNext()) {
					RDFNode node = n.nextNode();
					if(node instanceof Literal) {
						content+=((Literal)node).getLexicalForm();
					}
				}

				n = m.listObjectsOfProperty(r,rdfs_comment);
				while(n.hasNext()) {
					RDFNode node = n.nextNode();
					if(node instanceof Literal) {
						comment+=((Literal)node).getLexicalForm();
					}
				}
								
				// if there is no dc:description, we use the comment as content and set the comment empty
				if(content.length() == 0) {
					content = comment;
					comment = "";
				}
				
				
				
				try {
					String prefix = m.getNsURIPrefix(r.getNameSpace());
					String uri    = r.getNameSpace(); 
					Article a;
					if(uri == null) {
						// an anonymous bNode, create a name in the "internal" namespace
						a = new Article(main.getNamespaceSpecial(),"Anonymous"+r.getId().toString().replace(':','_'),"RDF import",content,"en",main.getNamespaceSpecial().constructURI("Anonymous"+r.getId().toString().replace(':','_')).toString(),new Date());
						log.info("anonymous bNode: <'"+r.getId()+"'>");
					} else {
						Namespace ns = new Namespace(prefix,uri);
						a = new Article(ns,r.getLocalName(),comment,content,"en",r.getURI(),new Date());
					}
					if(u != null) {
						a.setCreator(u.getLogin());
					}
					storeArticle(a);
				} catch(URISyntaxException ex) {
					ex.printStackTrace();
					// do not add article
				}
			}

			getDatabaseConnection().commit();
		} catch(Exception ex) {
			try {
				getDatabaseConnection().rollback();
				log.info("gracefully recovered from database error; rollback succeeded");
			} catch(SQLException ex2) {
				log.error("error during rollback",ex2);
			}
			throw new DBException("Database Error while adding URI to database",ex);
		} finally {
			try {
				getDatabaseConnection().setAutoCommit(true);
			} catch(SQLException ex2) {
				log.error("database error",ex2);
				throw new DBException("database error",ex2);
			}
		}
		
	}
	
	
	public void updateUser(User u) throws DBException {
		try {
			if(stmt_upd_user == null) {
				String sql_string = 
					"UPDATE users " +
					"SET first_name  = ?, " +
					"    last_name   = ?," +
					"    email       = ?, " +
					"    last_login  = ?, " +
					"    login_count = ? " +
					"WHERE login = ?";
				stmt_upd_user = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_upd_user.clearParameters();
			stmt_upd_user.setString(1,u.getFirstName());
			stmt_upd_user.setString(2,u.getLastName());
			stmt_upd_user.setString(3,u.getEmail());			
			stmt_upd_user.setTimestamp(4,new Timestamp(u.getLastLogin().getTime()));			
			stmt_upd_user.setInt(5,u.getLoginCount());			
			stmt_upd_user.setString(6,u.getLogin());
			stmt_upd_user.executeUpdate();
			
		} catch(SQLException ex) {
			throw new DBException("Database Error while updating user in database",ex);
		}
	}
	
	public void storeUser(User u, String password) throws DBException {
		if(password != null) {
			try {
				if(stmt_store_user == null) {
					String sql_string = 
						"INSERT INTO users (login,password,first_name,last_name,email) " +
						"VALUES (?,md5(?),?,?,?)";
					stmt_store_user = getDatabaseConnection().prepareStatement(sql_string);
				}
				stmt_store_user.clearParameters();
				stmt_store_user.setString(1,u.getLogin());
				stmt_store_user.setString(2,password);
				stmt_store_user.setString(3,u.getFirstName());
				stmt_store_user.setString(4,u.getLastName());
				stmt_store_user.setString(5,u.getEmail());			
				stmt_store_user.executeUpdate();
				
			} catch(SQLException ex) {
				throw new DBException("Database Error while adding user to database",ex);
			}
		} else {
			// system users cannot log in
			try {
				if(stmt_store_user_nopass == null) {
					String sql_string = 
						"INSERT INTO users (login,password,first_name,last_name,email) " +
						"VALUES (?,'!',?,?,?)";
					stmt_store_user_nopass = getDatabaseConnection().prepareStatement(sql_string);
				}
				stmt_store_user_nopass.clearParameters();
				stmt_store_user_nopass.setString(1,u.getLogin());
				stmt_store_user_nopass.setString(2,u.getFirstName());
				stmt_store_user_nopass.setString(3,u.getLastName());
				stmt_store_user_nopass.setString(4,u.getEmail());			
				stmt_store_user_nopass.executeUpdate();
				
			} catch(SQLException ex) {
				throw new DBException("Database Error while adding user to database",ex);
			}
			
		}
	}

	
	
	public User getUser(String login) throws DBException {
		try {
			if(stmt_get_user_noauth == null) {
				String sql_string = 
					"SELECT login,first_name,last_name,email,first_login,last_login,login_count " +
					"FROM users WHERE login = ?";
				stmt_get_user_noauth = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_get_user_noauth.clearParameters();
			stmt_get_user_noauth.setString(1,login);
			ResultSet s = stmt_get_user_noauth.executeQuery();
			if(s.next()) {
				User u = new User(s.getString("login"),s.getString("first_name"),s.getString("last_name"),
								  s.getString("email"),s.getTimestamp("first_login"),s.getTimestamp("last_login"),
								  s.getInt("login_count"));
				return u;
			} else {
				log.warn("could not find user with login "+login);
				return null;
			}
		} catch(SQLException ex) {
			throw new DBException("Database Error while adding URI to database",ex);
		}
	}

	public User getUser(String login, String password) throws DBException {
		try {
			if(stmt_get_user == null) {
				String sql_string = 
					"SELECT login,first_name,last_name,email,first_login,last_login,login_count " +
					"FROM users WHERE login = ? AND password = md5(?)";
				stmt_get_user = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_get_user.clearParameters();
			stmt_get_user.setString(1,login);
			stmt_get_user.setString(2,password);
			ResultSet s = stmt_get_user.executeQuery();
			if(s.next()) {
				User u = new User(s.getString("login"),s.getString("first_name"),s.getString("last_name"),
								  s.getString("email"),s.getTimestamp("first_login"),s.getTimestamp("last_login"),
								  s.getInt("login_count"));
				return u;
			} else {
				log.warn("could not authenticate user with login "+login);
				throw new AuthenticationFailure("could not authenticate user with login "+login);
			}
		} catch(SQLException ex) {
			throw new DBException("Database Error while adding URI to database",ex);
		}
		
	}
	
	
	public void deleteUser(User user) throws DBException {
		try {
			if(stmt_delete_user == null) {
				String sql_string = 
					"DELETE FROM users WHERE login = ?";
				stmt_delete_user = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_delete_user.clearParameters();
			stmt_delete_user.setString(1,user.getLogin());
			stmt_delete_user.executeUpdate();
			log.info("deleted user "+user.getLogin());
			
		} catch(SQLException ex) {
			throw new DBException("Database Error while delete user from database",ex);
		}
		
	}
	
	
	public Collection<User> listUsers() throws DBException {
		try {
			if(stmt_list_users == null) {
				String sql_string = 
					"SELECT login,first_name,last_name,email,first_login,last_login,login_count " +
					"FROM users";
				stmt_list_users = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_list_users.clearParameters();
			ResultSet s = stmt_list_users.executeQuery();
			
			TreeSet<User> result = new TreeSet<User>(new Comparator<User>() {
				Collator myCollator = Collator.getInstance();
				public int compare(User o1, User o2) {
					return myCollator.compare(o1.getLogin(),o2.getLogin());
				}
				
			});
			
			while(s.next()) {
				User u = new User(s.getString("login"),s.getString("first_name"),s.getString("last_name"),
								  s.getString("email"),s.getTimestamp("first_login"),s.getTimestamp("last_login"),
								  s.getInt("login_count"));
				result.add(u);
			}
			return result;
		} catch(SQLException ex) {
			throw new DBException("Database Error while retrieving list of users",ex);
		}
				
	}
	
	
	public Collection<User> listUsers(Role role) throws DBException {
		try {
			if(stmt_list_users_role == null) {
				String sql_string = 
					"SELECT login,first_name,last_name,email,first_login,last_login,login_count " +
					"FROM users, roles, user_roles " +
					"WHERE users.id = user_roles.user_id " +
					"  AND roles.id = user_roles.role_id " +
					"  AND roles.name = ?";
				stmt_list_users_role = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_list_users_role.clearParameters();
			stmt_list_users_role.setString(1,role.getName());
			ResultSet s = stmt_list_users_role.executeQuery();
			
			TreeSet<User> result = new TreeSet<User>(new Comparator<User>() {
				Collator myCollator = Collator.getInstance();
				public int compare(User o1, User o2) {
					return myCollator.compare(o1.getLogin(),o2.getLogin());
				}
				
			});
			
			while(s.next()) {
				User u = new User(s.getString("login"),s.getString("first_name"),s.getString("last_name"),
								  s.getString("email"),s.getTimestamp("first_login"),s.getTimestamp("last_login"),
								  s.getInt("login_count"));
				result.add(u);
			}
			return result;
		} catch(SQLException ex) {
			throw new DBException("Database Error while retrieving list of users",ex);
		}		
	}
	
	
	public void storeRole(Role role) throws DBException {
		try {
			if(stmt_store_role == null) {
				String sql_string = 
					"INSERT INTO roles (name,description) " +
					"VALUES (?,?)";
				stmt_store_role = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_store_role.clearParameters();
			stmt_store_role.setString(1,role.getName());
			stmt_store_role.setString(2,role.getDescription());
			stmt_store_role.executeUpdate();
			
			log.info("stored role "+role.getName());
		} catch(SQLException ex) {
			throw new DBException("Database Error while adding role to database",ex);
		}
	}

	public void updateRole(Role role) throws DBException {
		try {
			if(stmt_update_role == null) {
				String sql_string = 
					"UPDATE roles SET description = ? WHERE name = ?";
				stmt_update_role = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_update_role.clearParameters();
			stmt_update_role.setString(1,role.getDescription());
			stmt_update_role.setString(2,role.getName());
			stmt_update_role.executeUpdate();
			
			log.info("updated role "+role.getName());
		} catch(SQLException ex) {
			throw new DBException("Database Error while upading role in database",ex);
		}
	}

	public void deleteRole(Role role) throws DBException {
		try {
			if(stmt_delete_role == null) {
				String sql_string = 
					"DELETE FROM roles WHERE name = ?";
				stmt_delete_role = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_delete_role.clearParameters();
			stmt_delete_role.setString(1,role.getName());
			stmt_delete_role.executeUpdate();
			
			log.info("deleted role "+role.getName());
		} catch(SQLException ex) {
			throw new DBException("Database Error while deleting role from database",ex);
		}
	}
	
	
	public Role getRole(String name) throws DBException {
		try {
			if(stmt_get_role == null) {
				String sql_string = 
					"SELECT name,description " +
					"FROM roles WHERE name = ?";
				stmt_get_role = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_get_role.clearParameters();
			stmt_get_role.setString(1,name);
			ResultSet s = stmt_get_role.executeQuery();
			if(s.next()) {
				Role r = new Role(s.getString("name"),s.getString("description"));
				return r;
			} else {
				log.warn("could not find role with name "+name);
				return null;
			}
		} catch(SQLException ex) {
			throw new DBException("Database Error while retrieving role from database",ex);
		}
		
	}
	
	public WikiImage getImageByNamespaceTitle(String ns_prefix, String title) throws DBException {
		try {
			if(stmt_get_image_namespace_title == null) {
				String sql_string = 
						"SELECT image_id, ns_prefix, ns_uri, creator, title, timestamp, comment, mimetype, image " +
						"FROM current_image " +
						"WHERE ns_prefix = ? " +
						"  AND title = ?";				
				stmt_get_image_namespace_title = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_get_image_namespace_title.clearParameters();
			stmt_get_image_namespace_title.setString(1,ns_prefix);
			stmt_get_image_namespace_title.setString(2,title);
			
			ResultSet r = stmt_get_image_namespace_title.executeQuery();
			if(r.next()) {
				WikiImage img = 
					new WikiImage(new Namespace(r.getString("ns_prefix").trim(),r.getString("ns_uri").trim()),
					              r.getString("title").trim(),
						  		  r.getString("comment").trim(),
						  		  r.getString("mimetype").trim(),
						  		  r.getBytes("image"),
								  r.getTimestamp("timestamp"));
				img.setCreator(r.getString("creator"));
				return img;
			} else {
				log.error("database error while retrieving image "+ns_prefix+":"+title+" from database");
				throw new ArticleNotFoundException("Image with namespace '"+ns_prefix+"' and title '"+title+"' does not exist in the database");
			}
			
		} catch(SQLException ex) {
			throw new DBException("Database Error while retrieving image from database",ex);
		}
		
	}
	
	public boolean existsImageByNamespaceTitle(String ns_prefix, String title) throws DBException {
		try {
			if(stmt_exists_image_namespace_title == null) {
				String sql_string = 
						"SELECT image_id " +
						"FROM current_image " +
						"WHERE ns_prefix = ? " +
						"  AND title = ?";				
				stmt_exists_image_namespace_title = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_exists_image_namespace_title.clearParameters();
			stmt_exists_image_namespace_title.setString(1,ns_prefix);
			stmt_exists_image_namespace_title.setString(2,title);
			
			ResultSet r = stmt_exists_image_namespace_title.executeQuery();
			return r.next();
			
		} catch(SQLException ex) {
			throw new DBException("Database Error while retrieving image from database",ex);
		}
		
	}
	
	
	public void storeImage(WikiImage img) throws DBException {
		deleteImage(img); // delete first so we can overwrite
		try {
			if(stmt_store_image == null) {
				String sql_string = 
					"INSERT INTO images (namespace,title,comment,mimetype,image,creator) " +
					"VALUES (	(SELECT id FROM namespaces WHERE prefix=?)," +
					"			?," +
					"			?," +
					"			?," +
					"			?," +
					"			(SELECT id FROM users WHERE login = ?))";
				stmt_store_image = getDatabaseConnection().prepareStatement(sql_string);
			}
			
			ensureNamespace(img.getNamespace());
			
			stmt_store_image.clearParameters();
			stmt_store_image.setString(1,img.getNamespace().getPrefix());
			stmt_store_image.setString(2,img.getTitle());
			stmt_store_image.setString(3,img.getComment());
			stmt_store_image.setString(4,img.getMimeType());
			stmt_store_image.setBytes(5,img.getBytes());			
			stmt_store_image.setString(6,img.getCreator());
			stmt_store_image.executeUpdate();
			
		} catch(SQLException ex) {
			log.error("database error while adding image "+img.getQualifiedTitle()+" to database",ex);
			throw new DBException("database error while adding image to database",ex);
		}
	}
	

	
	
	public void deleteImage(WikiImage img) throws DBException {
		try {
			if(stmt_delete_image == null) {
				String sql_string = 
					"DELETE FROM images " +
					"WHERE namespace = (SELECT id FROM namespaces WHERE prefix=?) " +
					"  AND title = ?";
				stmt_delete_image = getDatabaseConnection().prepareStatement(sql_string);
			}
			
			stmt_delete_image.clearParameters();
			stmt_delete_image.setString(1,img.getNamespace().getPrefix());
			stmt_delete_image.setString(2,img.getTitle());
			stmt_delete_image.executeUpdate();
			
		} catch(SQLException ex) {
			log.error("database error while deleting image "+img.getQualifiedTitle()+" from database",ex);
			throw new DBException("database error while deleting image from database",ex);
		}
		
	}

	public void storeVariable(String name, String value) throws DBException {
		deleteVariable(name); // delete first so we can overwrite
		try {
			if(stmt_store_variable == null) {
				String sql_string = 
					"INSERT INTO variables (name,value) " +
					"VALUES (?,?)";
				stmt_store_variable = getDatabaseConnection().prepareStatement(sql_string);
			}
			
			
			stmt_store_variable.clearParameters();
			stmt_store_variable.setString(1,name.toLowerCase());
			stmt_store_variable.setString(2,value);
			stmt_store_variable.executeUpdate();
			
		} catch(SQLException ex) {
			log.error("database error while adding variable "+name+" to database",ex);
			throw new DBException("database error while adding variable to database",ex);
		}
	}
	

	public String getVariable(String name) throws DBException {
		try {
			if(stmt_get_variable == null) {
				String sql_string = 
						"SELECT name, value " +
						"FROM variables " +
						"WHERE name = ? ";				
				stmt_get_variable = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_get_variable.clearParameters();
			stmt_get_variable.setString(1,name.toLowerCase());
			
			ResultSet r = stmt_get_variable.executeQuery();
			if(r.next()) {
				return r.getString("value");
			} else {
				log.error("database error while retrieving variable "+name+" from database");
				throw new ArticleNotFoundException("variable '"+name+"' does not exist in the database");
			}
			
		} catch(SQLException ex) {
			throw new DBException("Database Error while retrieving image from database",ex);
		}
		
	}
	
	
	
	public void deleteVariable(String name) throws DBException {
		try {
			if(stmt_del_variable == null) {
				String sql_string = 
					"DELETE FROM variables " +
					"WHERE name = ?";
				stmt_del_variable = getDatabaseConnection().prepareStatement(sql_string);
			}
			
			stmt_del_variable.clearParameters();
			stmt_del_variable.setString(1,name.toLowerCase());
			stmt_del_variable.executeUpdate();
			
		} catch(SQLException ex) {
			log.error("database error while deleting variable "+name+" from database",ex);
			throw new DBException("database error while deleting variable from database",ex);
		}
		
	}
	
	
	public List<Variable> listVariables() throws DBException {
		try {
			if(stmt_list_variables == null) {
				String sql_string = 
						"SELECT name, value " +
						"FROM variables ";				
				stmt_list_variables = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_list_variables.clearParameters();
			
			ResultSet r = stmt_list_variables.executeQuery();
			
			List<Variable> result = new ArrayList<Variable>();
			
			while(r.next()) {
				result.add(new Variable(r.getString("name"),r.getString("value")));
			}
			return result;
						
		} catch(SQLException ex) {
			log.error("database error while retrieving list of variables from database");
			throw new DBException("Database Error while retrieving list of variables from database",ex);
		}
		
	}
	
	
	public void logMetadataAdd(Collection<Statement> stmts, User user) throws DBException {
		logMetadataUpdate(stmts,user,"A");
	}
	
	public void logMetadataDelete(Collection<Statement> stmts, User user) throws DBException {
		logMetadataUpdate(stmts,user,"D");
	}


	/**
	 * Log a metadata update in the metadata_journal table of the database using the given user and action. 
	 * 
	 * Each Statement of the update is stored as a triple in the metadata_journal table. Since the nodes in triples
	 * may be of different types, nodes are encoded in a special format by RDFNodeEncoder (see below). For performance
	 * reasons, large literal objects are stored in a separate column that is NULL in all other cases.
	 * 
	 * The "persistency" field is used to mark whether a triple is only in the memory model or has already been stored 
	 * persistently in the database model. In the future, it can thus be used to recover smoothly from system crashes 
	 * as it allows to perform all remaining transactions. Initially, journal entries are marked as "not persistent". 
	 * When the memory model is written to the database, RDFStore callsflushMetadataJournal() (see below), marking all 
	 * entries as persistent.  
	 * 
	 * @param stmts the collection of statements added or deleted in this update 
	 * @param user
	 * @param action either "A" (for add) or "D" (for delete)
	 * @throws DBException
	 */
	protected void logMetadataUpdate(Collection<Statement> stmts, User user, String action) throws DBException {
		assert(action.equals("A") || action.equals("D"));
		
		try {
			int revision = -1;
			
			if(stmt_log_metadata_nextrev == null) {
				String sql_string = "SELECT nextval('metadata_journal_transaction_ids')";
				stmt_log_metadata_nextrev = getDatabaseConnection().prepareStatement(sql_string);
				
			}
			ResultSet r_nextid = stmt_log_metadata_nextrev.executeQuery();
			if(r_nextid.next()) {
				revision = r_nextid.getInt(1);
			} else {
				throw new DBException("could not retrieve transaction id for next metadata journal transaction");
			}

			
			
			// each statement is stored as a triple in the metadata_journal table
			for(Statement stmt : stmts) {
				Resource subject   = stmt.getSubject();
				Property predicate = stmt.getPredicate();
				RDFNode  object    = stmt.getObject();
				
				RDFNodeEncoder subject_enc = new RDFNodeEncoder(subject);
				RDFNodeEncoder predicate_enc = new RDFNodeEncoder(predicate);
				RDFNodeEncoder object_enc = new RDFNodeEncoder(object);
				
				
				if(stmt_log_metadata == null) {
					String sql_string = 
						"INSERT INTO metadata_journal (transaction,uid,action,persistent,subject,predicate,object,object_overflow) "+
						"VALUES (?,(SELECT id FROM users WHERE login = ?),?,?,?,?,?,?)";
					stmt_log_metadata = getDatabaseConnection().prepareStatement(sql_string);						
				}
				stmt_log_metadata.clearParameters();
				stmt_log_metadata.setInt(1,revision);
				stmt_log_metadata.setString(2,user!=null?user.getLogin():null);
				stmt_log_metadata.setString(3,action);
				stmt_log_metadata.setBoolean(4,false);
				stmt_log_metadata.setString(5,subject_enc.getEncoded());
				stmt_log_metadata.setString(6,predicate_enc.getEncoded());
				stmt_log_metadata.setString(7,object_enc.getEncoded());
				stmt_log_metadata.setString(8,object_enc.getOverflow());
				
				stmt_log_metadata.executeUpdate();
										
			}
			
		
		
		} catch(SQLException ex) {
			log.error("database error while logging metadata changes");
			throw new DBException("Database Error while logging metadata changes",ex);
		}
		
	}
	
	/**
	 * Mark all metadata entries as stored.
	 * 
	 * This method should only be called if the system can ensure that all metadata entries
	 * have been stored persistently to the database.
	 * 
	 * @throws DBException
	 */
	public void flushMetadataJournal() throws DBException {
		try {
			if(stmt_flush_metadata == null) {
				String sql_string = "UPDATE metadata_journal SET persistent = TRUE WHERE persistent = FALSE";
				stmt_flush_metadata = getDatabaseConnection().prepareStatement(sql_string);						
			}
			stmt_flush_metadata.executeUpdate();
		} catch(SQLException ex) {
			log.error("database error while flushing metadata changes");
			throw new DBException("Database Error while flushing metadata changes",ex);
		}
	}
	
	/**
	 * Recover all journal entries that are not marked as persistent and return a list of
	 * corresponding statements that allows to perform the changes.
	 * 
	 * TODO: needs to be implemented, needs to differentiate between additions and deletions, ...
	 * also needed: restore specific revision for a resource, ...
	 * 
	 * @return
	 * @throws DBException
	 */
	public List<Statement> recoverJournal() throws DBException {
		
		// TODO
		
		return null;
	}
	
	
	public List<Transaction> listMetadataJournal(Model m, int offset, int limit) throws DBException {
		try {
			if(stmt_list_metadata == null) {
				String sql_string = 
					"SELECT transaction,date,login,action,subject,predicate,object,object_overflow," +
					"       (SELECT login from users where id = uid) AS user " +
					"FROM   metadata_journal " +
					"LIMIT  ? " +
					"OFFSET ? " +
					"ORDER BY transaction";
				stmt_list_metadata = getDatabaseConnection().prepareStatement(sql_string);						
			}
			stmt_list_metadata.clearParameters();
			stmt_list_metadata.setInt(1,limit);
			stmt_list_metadata.setInt(2,offset);
			
			ResultSet r = stmt_list_metadata.executeQuery();
			
			LinkedList<Transaction> transactions = new LinkedList<Transaction>();
			int previous_transaction_id = -1;
			Transaction current = null;
			while(r.next()) {
				int transaction_id = r.getInt("transaction");
				if(transaction_id != previous_transaction_id) {
					current = new Transaction(transaction_id);
					transactions.add(current);
					previous_transaction_id = transaction_id;
				}
				
				RDFNodeDecoder dec_subject = new RDFNodeDecoder(r.getString("subject"),null);
				RDFNodeDecoder dec_predicate = new RDFNodeDecoder(r.getString("predicate"),null);
				RDFNodeDecoder dec_object  = new RDFNodeDecoder(r.getString("object"), r.getString("object_overflow"));
				
				Statement stmt = m.createStatement((Resource)dec_subject.asJenaNode(m),
												   m.createProperty(((Resource)dec_predicate.asJenaNode(m)).getURI()),
												   dec_object.asJenaNode(m));
				
				User u = getUser(r.getString("user"));
				Date d = r.getTimestamp("timestamp");
				String action = r.getString("action");
								
				TransactionItem ti;
			
				if(action.equals("A")) {
					ti = new TransactionItem(stmt,TransactionItem.KINDS.ADD,u,d);
				} else if(action.equals("D")) {
					ti = new TransactionItem(stmt,TransactionItem.KINDS.REMOVE,u,d);					
				} else {
					log.error("while retrieving journal data: unsupported action "+action);
					throw new DBException("while retrieving journal data: unsupported action "+action);
				}
				
				current.addTransactionItem(ti);
			}
			
			return transactions;
		} catch(SQLException ex) {
			log.error("database error while retrieving journal data");
			throw new DBException("Database Error while retrieving journal data",ex);
		}
	}
	
	
	public List<Transaction> listMetadataJournal(Model m, WikiResource context, int offset, int limit) throws DBException {
		return null;
		
	}
	
	public void assignRole(User user, Role role) throws DBException {
		try {
			if(stmt_assign_role == null) {
				String sql_string = 
					"INSERT INTO user_roles (user_id,role_id) " +
					"VALUES ((SELECT id FROM users WHERE login = ?),(SELECT id FROM roles WHERE name=?))";
				stmt_assign_role = getDatabaseConnection().prepareStatement(sql_string);
			}
			
			
			stmt_assign_role.clearParameters();
			stmt_assign_role.setString(1,user.getLogin());
			stmt_assign_role.setString(2,role.getName());
			stmt_assign_role.executeUpdate();
			
			log.info("assigned role "+role.getName()+" to user "+user.getLogin());
		} catch(SQLException ex) {
			log.error("database error while adding role assignment ("+user.getLogin()+","+role.getName()+") to database",ex);
			throw new DBException("database error while adding role assignment ("+user.getLogin()+","+role.getName()+") to database",ex);
		}
	}
	

	
	
	public void retractRole(User user, Role role) throws DBException {
		try {
			if(stmt_retract_role == null) {
				String sql_string = 
					"DELETE FROM user_roles " +
					"WHERE user_id = (SELECT id FROM users WHERE login = ?) " +
					"  AND role_id = (SELECT id FROM roles WHERE name = ?)";
				stmt_retract_role = getDatabaseConnection().prepareStatement(sql_string);
			}
			
			stmt_retract_role.clearParameters();
			stmt_retract_role.setString(1,user.getLogin());
			stmt_retract_role.setString(2,role.getName());
			stmt_retract_role.executeUpdate();
			
			log.info("retracted role "+role.getName()+" of user "+user.getLogin());
		} catch(SQLException ex) {
			log.error("database error while retracting role assignment ("+user.getLogin()+","+role.getName()+") to database",ex);
			throw new DBException("database error while retracting role assignment ("+user.getLogin()+","+role.getName()+") to database",ex);
		}
		
	}
	
	
	public Collection<Role> listRoles() throws DBException {
		try {
			if(stmt_list_roles == null) {
				String sql_string = 
						"SELECT name,description " +
						"FROM roles ";				
				stmt_list_roles = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_list_roles.clearParameters();
			
			ResultSet r = stmt_list_roles.executeQuery();
			
			TreeSet<Role> result = new TreeSet<Role>(new Comparator<Role>() {
				Collator myCollator = Collator.getInstance();
				public int compare(Role o1, Role o2) {
					return myCollator.compare(o1.getName(),o2.getName());
				}
				
			});
			
			while(r.next()) {
				result.add(new Role(r.getString("name"),r.getString("description")));
			}
			return result;
						
		} catch(SQLException ex) {
			log.error("database error while retrieving list of roles from database");
			throw new DBException("Database Error while retrieving list of roles from database",ex);
		}
		
	}
	
	/**
	 * Retrieve the role assignments for a given user from the database
	 * 
	 * @param user
	 * @return
	 * @throws DBException
	 */
	public Collection<Role> listRoles(User user) throws DBException {
		try {
			if(stmt_list_roles_user == null) {
				String sql_string = 
						"SELECT roles.name,roles.description " +
						"FROM  roles,user_roles,users " +
						"WHERE roles.id = user_roles.role_id " +
						"  AND users.id = user_roles.user_id " +
						"  AND users.login = ?";				
				stmt_list_roles_user = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_list_roles_user.clearParameters();
			stmt_list_roles_user.setString(1,user.getLogin());
			
			ResultSet r = stmt_list_roles_user.executeQuery();
			
			TreeSet<Role> result = new TreeSet<Role>(new Comparator<Role>() {
				Collator myCollator = Collator.getInstance();
				public int compare(Role o1, Role o2) {
					return myCollator.compare(o1.getName(),o2.getName());
				}
				
			});
			
			while(r.next()) {
				result.add(new Role(r.getString("name"),r.getString("description")));
			}
			return result;
						
		} catch(SQLException ex) {
			log.error("database error while retrieving list of roles from database");
			throw new DBException("Database Error while retrieving list of roles from database",ex);
		}
		
	}

	/**
	 * Check the permissions of a user to a given resource using a given action.
	 * 
	 * The database stores resources using wildcards (like '%' or 'http://www.example.com/%'), which are
	 * matched when selecting.
	 * 
	 * This method checks the permission tables for matching patterns and orders them by precedence (in decending order). The
	 * entry with the highest precedence decides whether we grant or deny access.
	 * 
	 * @param user
	 * @param resource
	 * @param action
	 * @return
	 * @throws DBException
	 */
	public boolean hasPermission(User user, String resource, String action) throws DBException {
		try {
			// check permission table, ordered by precedence; the first entry beats all others and decides whether we deny or grant access
			if(stmt_check_permission == null) {
				String sql_string = 
						"SELECT * " +
						"FROM  all_user_permissions " +
						"WHERE login = ? AND ? LIKE resource_pattern AND ? LIKE action_pattern " +
						"ORDER BY precedence DESC " +
						"LIMIT 1";				
				stmt_check_permission = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_check_permission.clearParameters();
			stmt_check_permission.setString(1,user.getLogin());
			stmt_check_permission.setString(2,resource);
			stmt_check_permission.setString(3,action);
			
			ResultSet r = stmt_check_permission.executeQuery();
			
			if(r.next()) {
				if(r.getString("kind").equals("p")) {
					// entry is of type permission
					return true;
				} else if(r.getString("kind").equals("r")) {
					// entry is of type restriction
					return false;
				} else {
					// unknown type
					return false;
				}
			} else {
				// deny access by default
				return false;
			}
			
						
		} catch(SQLException ex) {
			log.error("database error while checking permissions for "+user.getLogin(),ex);
			throw new DBException("database error while checking permissions for "+user.getLogin(),ex);
		}
		
	}
	
	
	/**
	 * List all permissions that are associated with the given user. Both direct and inferred permissions are returned.
	 * 
	 * @param user
	 * @return
	 * @throws DBException
	 */
	public Collection<Permission> listPermissions(User user) throws DBException {
		try {
			if(stmt_list_permissions_user == null) {
				String sql_string = 
						"SELECT action_pattern,resource_pattern,precedence,indirect,kind,description " +
						"FROM   all_user_permissions " +
						"WHERE  login = ?";				
				stmt_list_permissions_user = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_list_permissions_user.clearParameters();
			stmt_list_permissions_user.setString(1,user.getLogin());
			
			ResultSet r = stmt_list_permissions_user.executeQuery();
			
			TreeSet<Permission> result = new TreeSet<Permission>(new PermissionComparator());
			
			while(r.next()) {
				
				Permission.Type t;
				switch(r.getString("kind").toLowerCase().charAt(0)) {
				case 'r': t = Permission.Type.RESTRICTION; break;
				case 'p': t = Permission.Type.PERMISSION; break;
				default : t = Permission.Type.NONE; break;
				}
				
				Permission p = new Permission(t,r.getInt("precedence"),r.getString("resource_pattern"),r.getString("action_pattern"),r.getString("description")); 
				p.setDirect(!r.getBoolean("indirect"));
				
				result.add(p);
			}
			return result;
						
		} catch(SQLException ex) {
			log.error("database error while retrieving list of permissions from database");
			throw new DBException("Database Error while retrieving list of permissions from database",ex);
		}
	}
	

	/**
	 * List all permissions that are associated with the given role.
	 * 
	 * @param role
	 * @return
	 * @throws DBException
	 */
	public Collection<Permission> listPermissions(Role role) throws DBException {
		try {
			if(stmt_list_permissions_role == null) {
				String sql_string = 
						"SELECT resource_pattern, action_pattern, precedence, kind, role_permissions.description " +
						"FROM   role_permissions, roles " +
						"WHERE  roles.name = ? " +
						"  AND  roles.id = role_permissions.role_id ";				
				stmt_list_permissions_role = getDatabaseConnection().prepareStatement(sql_string);
			}
			stmt_list_permissions_role.clearParameters();
			stmt_list_permissions_role.setString(1,role.getName());
			
			ResultSet r = stmt_list_permissions_role.executeQuery();
			
			TreeSet<Permission> result = new TreeSet<Permission>(new PermissionComparator());
			
			while(r.next()) {

				Permission.Type t;
				switch(r.getString("kind").toLowerCase().charAt(0)) {
				case 'r': t = Permission.Type.RESTRICTION; break;
				case 'p': t = Permission.Type.PERMISSION; break;
				default : t = Permission.Type.NONE; break;
				}

				Permission p = new Permission(t, r.getInt("precedence"),r.getString("resource_pattern"),r.getString("action_pattern"),r.getString("description")); 
				
				result.add(p);
			}
			return result;
						
		} catch(SQLException ex) {
			log.error("database error while retrieving list of permissions from database");
			throw new DBException("Database Error while retrieving list of permissions from database",ex);
		}
	}

	
	
	/**
	 * Add a permission specified by name and pattern to the given user.
	 * 
	 * @param user
	 * @param p
	 * @throws DBException
	 */
	public void addPermission(User user, Permission p) throws DBException {
		
		try {
			if(stmt_add_permission_user == null) {
				String sql_string = 
					"INSERT INTO user_permissions (user_id,resource_pattern,action_pattern,kind,precedence,description) " +
					"VALUES ((SELECT id FROM users WHERE login = ?),?,?,?,?,?)";
				stmt_add_permission_user = getDatabaseConnection().prepareStatement(sql_string);
			}
			
			
			stmt_add_permission_user.clearParameters();
			stmt_add_permission_user.setString(1,user.getLogin());
			stmt_add_permission_user.setString(2,p.getResourcePattern());
			stmt_add_permission_user.setString(3,p.getActionPattern());
			stmt_add_permission_user.setString(4,p.getType().shortLabel());
			stmt_add_permission_user.setInt(5,p.getPrecedence());
			stmt_add_permission_user.setString(6,p.getDescription());
			stmt_add_permission_user.executeUpdate();
			
			log.info("added permission "+p.getType()+" on resource "+p.getResourcePattern()+", action "+p.getActionPattern()+" to user "+user.getLogin());
		} catch(SQLException ex) {
			log.error("database error while adding permission ("+user.getLogin()+", resource "+p.getResourcePattern()+", action "+p.getActionPattern()+") to database",ex);
			throw new DBException("database error while adding permission ("+user.getLogin()+", resource "+p.getResourcePattern()+", action "+p.getActionPattern()+") to database",ex);
		}
		
	}
	
	
	/**
	 * Add a permission specified by name and pattern to the given role.
	 * 
	 * @param role
	 * @param p
	 * @throws DBException
	 */
	public void addPermission(Role role, Permission p) throws DBException {

		try {
			if(stmt_add_permission_role == null) {
				String sql_string = 
					"INSERT INTO role_permissions (role_id,resource_pattern,action_pattern,kind,precedence,description) " +
					"VALUES ((SELECT id FROM roles WHERE name = ?),?,?,?,?,?)";
				stmt_add_permission_role = getDatabaseConnection().prepareStatement(sql_string);
			}
			
			
			stmt_add_permission_role.clearParameters();
			stmt_add_permission_role.setString(1,role.getName());
			stmt_add_permission_role.setString(2,p.getResourcePattern());
			stmt_add_permission_role.setString(3,p.getActionPattern());
			stmt_add_permission_role.setString(4,p.getType().shortLabel());
			stmt_add_permission_role.setInt(5,p.getPrecedence());
			stmt_add_permission_role.setString(6,p.getDescription());
			stmt_add_permission_role.executeUpdate();
			
			log.info("added permission "+p.getType()+" on resource "+p.getResourcePattern()+", action "+p.getActionPattern()+" to role "+role.getName());
		} catch(SQLException ex) {
			log.error("database error while adding permission ("+role.getName()+", resource "+p.getResourcePattern()+", action "+p.getActionPattern()+") to database",ex);
			throw new DBException("database error while adding permission ("+role.getName()+", resource "+p.getResourcePattern()+", action "+p.getActionPattern()+") to database",ex);
		}		
	}
	
	
	/**
	 * Remove a permission specified by name and pattern from the given user.
	 * 
	 * @param user
	 * @param p
	 * @throws DBException
	 */
	public void removePermission(User user, Permission p) throws DBException {

		try {
			if(stmt_remove_permission_user == null) {
				String sql_string = 
					"DELETE FROM user_permissions " +
					"WHERE user_id = (SELECT id FROM users WHERE login = ?) " +
					"  AND resource_pattern = ? " +
					"  AND action_pattern = ? " +
					"  AND precedence = ? " +
					"  AND kind = ?";
				stmt_remove_permission_user = getDatabaseConnection().prepareStatement(sql_string);
			}
			
			stmt_remove_permission_user.clearParameters();
			stmt_remove_permission_user.setString(1,user.getLogin());
			stmt_remove_permission_user.setString(2,p.getResourcePattern());
			stmt_remove_permission_user.setString(3,p.getActionPattern());
			stmt_remove_permission_user.setInt(4,p.getPrecedence());
			stmt_remove_permission_user.setString(5,p.getType().shortLabel());
			stmt_remove_permission_user.executeUpdate();
			
			log.info("removed permission "+p.getType()+" on resource "+p.getResourcePattern()+", action "+p.getActionPattern()+" to user "+user.getLogin());
		} catch(SQLException ex) {
			log.error("database error while removing permission ("+user.getLogin()+", resource "+p.getResourcePattern()+", action "+p.getActionPattern()+") to database",ex);
			throw new DBException("database error while removing permission ("+user.getLogin()+", resource "+p.getResourcePattern()+", action "+p.getActionPattern()+") to database",ex);
		}
		
	}
	
	/**
	 * Remove a permission specified by name and pattern from the given role.
	 * 
	 * @param role
	 * @param p
	 * @throws DBException
	 */
	public void removePermission(Role role, Permission p) throws DBException {

		try {
			if(stmt_remove_permission_role == null) {
				String sql_string = 
					"DELETE FROM role_permissions " +
					"WHERE role_id = (SELECT id FROM roles WHERE name = ?) " +
					"  AND resource_pattern = ? " +
					"  AND action_pattern = ? " +
					"  AND precedence = ? " +
					"  AND kind = ?";
				stmt_remove_permission_role = getDatabaseConnection().prepareStatement(sql_string);
			}
			
			stmt_remove_permission_role.clearParameters();
			stmt_remove_permission_role.setString(1,role.getName());
			stmt_remove_permission_role.setString(2,p.getResourcePattern());
			stmt_remove_permission_role.setString(3,p.getActionPattern());
			stmt_remove_permission_role.setInt(4,p.getPrecedence());
			stmt_remove_permission_role.setString(5,p.getType().shortLabel());
			stmt_remove_permission_role.executeUpdate();
			
			log.info("removed permission "+p.getType()+" on resource "+p.getResourcePattern()+", action "+p.getActionPattern()+" to role "+role.getName());
		} catch(SQLException ex) {
			log.error("database error while removing permission ("+role.getName()+", resource "+p.getResourcePattern()+", action "+p.getActionPattern()+") to database",ex);
			throw new DBException("database error while removing permission ("+role.getName()+", resource "+p.getResourcePattern()+", action "+p.getActionPattern()+") to database",ex);
		}		
		
	}
		
	
	
	public Document parseXML(String content) {		
		try {
			long start = System.currentTimeMillis();
			ByteArrayInputStream is = new ByteArrayInputStream(content.getBytes());
			Document d = builder.parse(is);
			long end = System.currentTimeMillis();
			//log.debug("XML content parsed in "+(end-start)+" ms");
			
			return d;
		} catch(SAXException ex) {
			log.error("error parsing XML content of article",ex);
			return null;
		} catch(IOException ex) {
			// never happens for ByteArrayInputStream
			return null;
		}
	}
	
		
	PreparedStatement stmt_exists_article_namespace_title;
	PreparedStatement stmt_get_article_namespace_title;
	PreparedStatement stmt_get_article_revision;
	PreparedStatement stmt_search_article, stmt_list_articles;
	PreparedStatement stmt_get_revisions_namespace_title;
	
	PreparedStatement stmt_update_article_text, stmt_update_article_revision;
	PreparedStatement stmt_delete_article, stmt_undelete_article;
	PreparedStatement stmt_store_article_page, stmt_store_article_lastid;
	
	PreparedStatement stmt_get_recent;
	
	PreparedStatement stmt_store_uri, stmt_get_uri;
	PreparedStatement stmt_store_ns, stmt_get_ns, stmt_upd_ns, stmt_del_ns;
	
	PreparedStatement stmt_get_ns_prefix, stmt_get_ns_uri, stmt_get_namespaces;
	
	PreparedStatement stmt_get_user, stmt_get_user_noauth, stmt_store_user, stmt_store_user_nopass, stmt_upd_user, stmt_delete_user;
	PreparedStatement stmt_list_users, stmt_list_users_role;
	
	PreparedStatement stmt_get_image_namespace_title;
	PreparedStatement stmt_exists_image_namespace_title;
	PreparedStatement stmt_store_image;
	PreparedStatement stmt_delete_image;
	
	PreparedStatement stmt_get_variable, stmt_list_variables, stmt_del_variable, stmt_store_variable;
	
	PreparedStatement stmt_log_metadata, stmt_log_metadata_literal, stmt_log_metadata_nextrev;
	PreparedStatement stmt_flush_metadata;
	PreparedStatement stmt_list_metadata, stmt_list_metadata_ctx;
	
	PreparedStatement stmt_store_role, stmt_get_role, stmt_update_role, stmt_delete_role;
	PreparedStatement stmt_assign_role, stmt_retract_role, stmt_list_roles, stmt_list_roles_user;
	
	PreparedStatement stmt_check_permission, stmt_add_permission, stmt_remove_permission;
	PreparedStatement stmt_list_permissions, stmt_list_permissions_user, stmt_list_permissions_role;
	PreparedStatement stmt_add_permission_user, stmt_add_permission_role, stmt_remove_permission_user, stmt_remove_permission_role;
	
	
	private static class PermissionComparator implements Comparator<Permission> {
		Collator myCollator = Collator.getInstance();
		
		public int compare(Permission o1, Permission o2) {
			// compare by precedence, pattern, and name
			int c_precedence = Integer.signum(o2.getPrecedence()-o1.getPrecedence());
			int c_pattern = c_precedence==0?myCollator.compare(o1.getResourcePattern(),o2.getResourcePattern()):c_precedence;
			return c_pattern==0?myCollator.compare(o1.getActionPattern(),o2.getActionPattern()):c_pattern;
		}
		
	}
}

