/**
 * 
 */
package at.srfg.ikewiki.db;

import java.io.File;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.index.Term;
import org.apache.lucene.queryParser.MultiFieldQueryParser;
import org.apache.lucene.queryParser.ParseException;
import org.apache.lucene.queryParser.QueryParser;
import org.apache.lucene.search.Hits;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;

import at.srfg.ikewiki.Main;
import at.srfg.ikewiki.data.Article;

/**
 * @author sschaffe
 *
 */
public class LuceneStore {

	private static Log log = LogFactory.getLog(LuceneStore.class);

	private Main main;
	
	private static final String INDEX = ".ikewiki_index";
	
	
	private StandardAnalyzer analyzer = null;
	private QueryParser parser = null;
	
	/**
	 * 
	 */
	public LuceneStore(Main main) {
		super();
		this.main = main;
				
	}

	public void initIndexer() {
		//reIndex();
	}
		
	
	// initialise the index with all articles from the database if the index is empty/does not exist
	public void reIndex() {
		try {
			for(Article a : main.getStorageBackend().listArticles()) {
				storeArticle(a);
			}
		} catch(DBException ex) {
			log.error("error while indexing articles",ex);
		}		
	}
	
	public void storeArticle(Article a) throws DBException {
		if(analyzer == null) {
			analyzer = new StandardAnalyzer();
		}
		
		
		try {
			File f = new File(INDEX);

			if(f.exists() && !f.isDirectory()) {
				log.error("cannot index article; file "+INDEX+" is not an index directory");
				throw new DBException("cannot index article; file "+INDEX+" is not an index directory");
			} else if(f.exists() && (!f.canRead() || !f.canWrite())) {
				log.error("cannot index article; file "+INDEX+" has wrong permissions");
				throw new DBException("cannot index article; file "+INDEX+" has wrong permissions");
			}
			
			int deleted = 0;
			try {
				// first delete possible previous entries
				IndexReader reader = IndexReader.open(f);
				deleted = reader.deleteDocuments(new Term("uri",a.getUri()));
				reader.close();
			} catch(IOException ex) {
				log.warn("failed reading index; it probably does not exist");
			}
				
			
			boolean create = f.exists()?false:true;
			
			IndexWriter writer = new IndexWriter(f,analyzer,create);
			
			
			Document doc = new Document();
			doc.add(new Field("uri",a.getUri().trim(),Field.Store.YES,Field.Index.UN_TOKENIZED));
			doc.add(new Field("qtitle",a.getQualifiedTitle(),Field.Store.YES,Field.Index.UN_TOKENIZED));
			doc.add(new Field("namespace",a.getNamespace().getPrefix(),Field.Store.YES,Field.Index.UN_TOKENIZED));
			doc.add(new Field("title",a.getTitle(),Field.Store.YES,Field.Index.TOKENIZED));
			 // TODO: could make use of info in RDFStore and list all authors
			if(a.getCreator() != null) {
				doc.add(new Field("author",a.getCreator(),Field.Store.YES,Field.Index.TOKENIZED));
			} else {
				doc.add(new Field("author","Anonymous",Field.Store.YES,Field.Index.TOKENIZED));				
			}
			doc.add(new Field("content",a.getText(),Field.Store.NO,Field.Index.TOKENIZED));
			
			writer.addDocument(doc);
			
			if(deleted > 0) {
				writer.optimize();
			}
			
			writer.close();
			
			log.info("indexed article "+a.getQualifiedTitle());
		} catch(IOException ex) {
			log.error(ex);
		}
	}
	
	
	public List<Article> searchArticles(String search_terms, int limit, int offset) throws DBException {
		String[] fields = { "title", "content" };
		
		if(analyzer == null) {
			analyzer = new StandardAnalyzer();
		}

		if(parser == null) {
			parser = new MultiFieldQueryParser(fields,analyzer);
		}
		
		try {
						
			File f = new File(INDEX);
			IndexReader reader = IndexReader.open(f);
			
			Query query = parser.parse(search_terms);
			IndexSearcher searcher = new IndexSearcher(reader);
			
			Hits hits = searcher.search(query);
			
			log.info("found " + hits.length() + " documents that matched: " + search_terms);
			
			LinkedList<Article> result = new LinkedList<Article>();
			for(int i=offset; i<hits.length() && i<(offset+limit); i++) {
				Document doc = hits.doc(i);
				Article a = main.getStorageBackend().getArticleByNamespaceTitle(doc.get("namespace"),doc.get("title")); 
								
				result.add(a);
			}
			
			log.debug("returning "+result.size()+ " articles");
			
			return result;
		} catch(ParseException ex) {
			log.error("error while searching index",ex);
			throw new DBException("error while searching index",ex);
		} catch(IOException ex) {
			log.error("error while searching index",ex);
			throw new DBException("error while searching index",ex);
		}
	}
	
}
