/**
 *  Copyright (C) 2006
 *  Sebastian Schaffert and Rupert Westenthaler
 *  Salzburg Research Forschungsgesellschaft m.b.H.
 *
 *  IkeWiki is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU Lesser General Public
 * 	License as published by the Free Software Foundation; either
 * 	version 2 of the License, or (at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * 	Lesser General Public License for more details.
 *
 * 	You should have received a copy of the GNU Lesser General Public
 * 	License along with this library; if not, write to the
 * 	Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * 	Boston, MA 02111-1307, USA.
 * 
 *  File: RDFNodeDecoder.java
 */
package at.srfg.ikewiki.db;

import at.srfg.ikewiki.Main;
import at.srfg.ikewiki.data.WikiAnonResource;
import at.srfg.ikewiki.data.WikiLiteral;
import at.srfg.ikewiki.data.WikiNode;

import com.hp.hpl.jena.rdf.model.AnonId;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.RDFNode;

/**
 * Decode a database string representation of an RDF node either into a Jena RDFNode or into a WikiNode
 * 
 * @author wastl
 *
 */
public class RDFNodeDecoder {

	private String encoded, overflow;
	
	/**
	 * 
	 */
	public RDFNodeDecoder(String encoded, String overflow) {
		this.encoded = encoded;
		this.overflow = overflow;
	}

	/**
	 * Decode the encoded string representation into an IkeWiki WikiNode structure.
	 * Takes the "main" instance as parameter to be able to resolve namespace prefixes etc.
	 * 
	 * @param main
	 * @return
	 */
	public WikiNode asWikiNode(Main main) throws DBException {
		// split "encoded" into the prefix and the content
		String[] components = encoded.split("::");
		
		String prefix  = components[0];
		String content = components[1];
		
		// parse the prefix (all fields separated by ":")
		components = prefix.split(":");
		
		String type = components[0];
		
		if(type.equals("A")) {
			return new WikiAnonResource(content.trim());
		} else if(type.equals("U")) {
			return main.getResourceByUri(content.trim());
		} else if(type.equals("L")) {
			assert(components.length == 3);
			String language = components[1].equals("null")?null:components[1];
			String lit_content = components[2].equals("+")?this.overflow:content;
			return new WikiLiteral(lit_content,language);
		} else {
			throw new IllegalArgumentException("unknown type in encoded content:" + type);
		}
		
	}
	
	
	/**
	 * Decode the encoded string representation into an Jena RDFNode structure.
	 * Takes the model to construct appropriate instances.
	 * 
	 * @param main
	 * @return
	 */
	public RDFNode asJenaNode(Model m) {
		// split "encoded" into the prefix and the content
		String[] components = encoded.split("::");
		
		String prefix  = components[0];
		String content = components[1];
		
		// parse the prefix (all fields separated by ":")
		components = prefix.split(":");
		
		String type = components[0];
		
		if(type.equals("A")) {
			return m.createResource(AnonId.create(content.trim()));
		} else if(type.equals("U")) {
			return m.createResource(content.trim());
		} else if(type.equals("L")) {
			assert(components.length == 3);
			String language = components[1].equals("null")?null:components[1];
			String lit_content = components[2].equals("+")?this.overflow:content;
			if(language == null) {
				return m.createLiteral(lit_content);
			} else {
				return m.createLiteral(lit_content,language);
			}
		} else {
			throw new IllegalArgumentException("unknown type in encoded content:" + type);
		}		
	}
}
