/**
 *  Copyright (C) 2005
 *  Sebastian Schaffert and Rupert Westenthaler
 *  Salzburg Research Forschungsgesellschaft m.b.H.
 *
 *  IkeWiki is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU Lesser General Public
 * 	License as published by the Free Software Foundation; either
 * 	version 2 of the License, or (at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * 	Lesser General Public License for more details.
 *
 * 	You should have received a copy of the GNU Lesser General Public
 * 	License along with this library; if not, write to the
 * 	Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * 	Boston, MA 02111-1307, USA.
 * 
 */
package at.srfg.ikewiki.db;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;

import at.srfg.ikewiki.Main;
import at.srfg.ikewiki.data.Article;
import at.srfg.ikewiki.data.Namespace;
import at.srfg.ikewiki.data.Role;
import at.srfg.ikewiki.data.User;
import at.srfg.ikewiki.data.WikiImage;
import at.srfg.ikewiki.data.WikiLink;
import at.srfg.ikewiki.data.WikiLiteral;
import at.srfg.ikewiki.data.WikiNode;
import at.srfg.ikewiki.data.WikiResource;
import at.srfg.ikewiki.misc.CanonicalXMLWriter;
import at.srfg.ikewiki.misc.XMLWriter;
import at.srfg.ikewiki.render.RenderingException;
import at.srfg.ikewiki.transactions.Transaction;
import at.srfg.ikewiki.transactions.TransactionItem;

import com.drew.imaging.jpeg.JpegMetadataReader;
import com.drew.imaging.jpeg.JpegProcessingException;
import com.drew.metadata.Directory;
import com.drew.metadata.Metadata;
import com.drew.metadata.MetadataException;
import com.drew.metadata.Tag;
import com.hp.hpl.jena.db.DBConnection;
import com.hp.hpl.jena.db.IDBConnection;
import com.hp.hpl.jena.db.ModelRDB;
import com.hp.hpl.jena.graph.Node;
import com.hp.hpl.jena.graph.Node_Fluid;
import com.hp.hpl.jena.graph.Node_NULL;
import com.hp.hpl.jena.ontology.OntClass;
import com.hp.hpl.jena.ontology.OntModel;
import com.hp.hpl.jena.ontology.OntModelSpec;
import com.hp.hpl.jena.ontology.OntProperty;
import com.hp.hpl.jena.ontology.OntResource;
import com.hp.hpl.jena.query.Query;
import com.hp.hpl.jena.query.QueryExecution;
import com.hp.hpl.jena.query.QueryExecutionFactory;
import com.hp.hpl.jena.query.QueryFactory;
import com.hp.hpl.jena.query.QuerySolution;
import com.hp.hpl.jena.query.ResultSet;
import com.hp.hpl.jena.query.ResultSetFormatter;
import com.hp.hpl.jena.rdf.model.Literal;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelChangedListener;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.ModelMaker;
import com.hp.hpl.jena.rdf.model.NodeIterator;
import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.RDFWriter;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;
import com.hp.hpl.jena.util.iterator.ExtendedIterator;
import com.hp.hpl.jena.vocabulary.ReasonerVocabulary;

/**
 * @author sschaffe
 *
 */
public class RDFStore extends DBCommon implements ModelChangedListener {

	public enum OntologyLevel {
		RDFS, OWL_RDFS, OWL_DL, OWL_FULL
	}
	
	private OntologyLevel ontology_level;
	
	
	private static Log log = LogFactory.getLog(RDFStore.class);
	
	private Model model;
	private ModelRDB dbmodel;
	private IDBConnection conn;	
	
	private OntModel ontology;
	
	private ModelMaintanance maintanance = null;
	
	private boolean dirty = false;
	
	
	// a mapping Thread  -> Transaction list that stores for each thread the list of updated
	// statements; a transaction is commited when the transaction's activity count becomes 0
	private Hashtable<Thread,Transaction> transaction_log;
		
	// caches:
//	private SortedSet<WikiResource> all_types      = null;
//	private SortedSet<WikiResource> all_properties = null;
//	
//	private HashMap<WikiResource,SortedSet<WikiResource>> type_properties;
	
	/**
	 * 
	 */
	public RDFStore(Main main) throws DBException {
		super(main);
		
		// create a database connection
		conn = new DBConnection(super.getConnectionString(), DB_USER, DB_PASS, "PostgreSQL");
		if(!conn.isFormatOK()) {
			// when DB is initialised, we want to enable URI compression!
			conn.getDriver().setDoCompressURI(true);
		}
		
		// create a model maker with the given connection parameters
		//ModelMaker dbmaker = ModelFactory.createModelRDBMaker(conn);

		// create a default model
		//dbmodel = dbmaker.createDefaultModel();
		if(!conn.containsDefaultModel()) {
			dbmodel = ModelRDB.createModel(conn);
		} else {
			dbmodel = ModelRDB.open(conn);
		}
		//model.register(this);
		
		ModelMaker maker = createModelMaker();
		model   = maker.createDefaultModel();
		model.add(dbmodel);
		
		log.info("loaded "+model.size()+" statements from database");
		
		maintanance = new ModelMaintanance(this);
		
		determineOntologyLevel();
		
		// create a model capable of doing inferencing
		//ontology = ModelFactory.createOntologyModel(OntModelSpec.OWL_MEM_TRANS_INF,model);
		//ontology = ModelFactory.createOntologyModel(OntModelSpec.RDFS_MEM_RDFS_INF,model);
		
		switch(getOntologyLevel()) {
		case OWL_DL:
		case OWL_FULL:
			log.info("initialising OWL-DL reasoner ...");
			//ontology = ModelFactory.createOntologyModel(PelletReasonerFactory.THE_SPEC,model);			
			ontology = ModelFactory.createOntologyModel(OntModelSpec.OWL_MEM_RULE_INF,model);
			break;
		case OWL_RDFS:
			log.info("initialising OWL-RDFS reasoner ...");
			//ontology = ModelFactory.createOntologyModel(PelletReasonerFactory.THE_SPEC,model);			
			ontology = ModelFactory.createOntologyModel(OntModelSpec.OWL_MEM_RDFS_INF,model);			
//			ontology.getReasoner().setParameter(ReasonerVocabulary.PROPruleMode,"hybrid");
//			ontology.getReasoner().setParameter(ReasonerVocabulary.PROPenableTGCCaching,true);
//			ontology.getReasoner().setParameter(ReasonerVocabulary.PROPenableOWLTranslation,true);
			break;
		case RDFS:
			log.info("initialising RDFS reasoner ...");
			//ontology = ModelFactory.createOntologyModel(PelletReasonerFactory.THE_SPEC,model);			
			ontology = ModelFactory.createOntologyModel(OntModelSpec.RDFS_MEM_RDFS_INF,model);			
			ontology.getReasoner().setParameter(ReasonerVocabulary.PROPruleMode,"hybrid");
			ontology.getReasoner().setParameter(ReasonerVocabulary.PROPenableTGCCaching,true);
			break;
		default:
			ontology = ModelFactory.createOntologyModel(OntModelSpec.RDFS_MEM_TRANS_INF,model);			
		break;
		}
		// should call registerNamespaces(), but this can only be done if storage backend is available :-(
		
		transaction_log = new Hashtable<Thread,Transaction>();
		
		// must be last thing, we don't want to listen to events in the initialisation of the model
		ontology.register(this);
		
		
//		type_properties = new HashMap<WikiResource,SortedSet<WikiResource>>();
	}
	
	
	private void determineOntologyLevel() {
		Resource owl_class = model.createResource(Namespace.NS_OWL+"Class");
		if(model.containsResource(owl_class)) {
			ontology_level = OntologyLevel.OWL_RDFS; 
			//ontology_level = OntologyLevel.OWL_DL; 
		} else {
			ontology_level = OntologyLevel.RDFS;
		}
		//ontology_level = OntologyLevel.OWL_DL; 
	}
	
	
	private static ModelMaker createModelMaker() {
		return ModelFactory.createMemModelMaker();
	}
	
	
	public OntologyLevel getOntologyLevel() {
		return ontology_level;
	}
	
	
	public void registerNamespaces(DBStore db) throws DBException {
		Collection<Namespace> namespaces = db.listNamespaces();
		
		for(Namespace ns : namespaces) {
			model.setNsPrefix(ns.getPrefix(),ns.getUri());
		}
	}
	
	
	public void updateNamespace(Namespace n) {
		String old_prefix = model.getNsURIPrefix(n.getUri());
		if(old_prefix != null) {
			model.removeNsPrefix(old_prefix);
			model.setNsPrefix(n.getPrefix(),n.getUri());
		} else {
			log.error("could not update namespace "+n+" because uri does not exist in RDF model, persistent data may be inconsistent");
		}
	}
	
	
	public void deleteNamespace(String uri) {
		String old_prefix = model.getNsURIPrefix(uri);
		if(old_prefix != null) {
			model.removeNsPrefix(old_prefix);
		} else {
			log.error("could not remove namespace "+uri+" because uri does not exist in RDF model, persistent data may be inconsistent");
		}
	}
	
	
	public Model addTripleData(String rdf_data) throws DBException {
		return addTripleData(rdf_data.getBytes());
	}

	public Model addTripleData(byte[] rdf_data) throws DBException {
		
		return addTripleData(new ByteArrayInputStream(rdf_data));
	}
	
	/**
	 * Add the statements in the provided data source (RDF/XML format) to the database;
	 * 
	 * @param is the input stream from which to read the statements
	 * @return a model containing all newly added/updated statements
	 * @throws DBException
	 */
	public Model addTripleData(InputStream is) throws DBException {
		ModelMaker maker = createModelMaker();
		Model tmpmodel   = maker.createDefaultModel();
		
		tmpmodel.read(is,main.getNamespaceBase().getUri());
		
		ontology.add(tmpmodel);
		
		return tmpmodel;
	}
	
	
	public void addStatement(String subject, String object, String predicate) {
		
	}
	
	
	public Model getModel() {
		return ontology;
	}
	
	public Model getMemModel() {
		return model;
	}
	
	
	public OntModel getOntologyModel() {
		return ontology;
	}
	
	public static String resolveURI(Resource r) {
		if(r.getURI() == null) {
			return Namespace.NS_IKEWIKI_SPECIAL+"#Anonymous"+r.getId().toString().replace(':','_'); 
		} else {
			return r.getURI();
		}
	}
	
		
	/**
	 * Retrieve all outgoing properties of a given subject from the triple database;
	 * 
	 * Returns a list of WikiResource objects representing the (distinct) properties
	 * 
	 * @param subject
	 * @return
	 * @throws DBException
	 */
	public List<WikiResource> getPropertiesBySubject(WikiResource subject) throws DBException {
		ArrayList<WikiResource> results = new ArrayList<WikiResource>();

		// used for eliminating duplicates
		Set<String> seen = new HashSet<String>();
		// don't want comments to appear in links:
		seen.add("http://www.w3.org/2000/01/rdf-schema#comment");


		Resource r = ontology.getResource(subject.getUri());
		
		for(Iterator it = ontology.listStatements(r,null,(RDFNode)null); it.hasNext(); ) {
			Statement st = (Statement)it.next();

			Property predicate = st.getPredicate();
            if(!seen.contains(predicate.getURI())) {
            	results.add(main.getResourceByUri(resolveURI(predicate)));
            	seen.add(predicate.getURI());
            }
		}
		
		return results;
	}

	/**
	 * Retrieve all outgoing properties of a given subject from the triple database;
	 * 
	 * Returns a list of WikiResource objects representing the (distinct) properties
	 * 
	 * @param subject
	 * @return
	 * @throws DBException
	 */
	public List<WikiResource> getPropertiesByObject(WikiResource object) throws DBException {
		ArrayList<WikiResource> results = new ArrayList<WikiResource>();

		// used for eliminating duplicates
		Set<String> seen = new HashSet<String>();
		
		Resource r = ontology.getResource(object.getUri());
		
		for(Iterator it = ontology.listStatements(null,null,r); it.hasNext(); ) {
			Statement st = (Statement)it.next();

			Property predicate = st.getPredicate();
            if(!seen.contains(predicate.getURI())) {
            	results.add(main.getResourceByUri(resolveURI(predicate)));
            	seen.add(predicate.getURI());
            }
		}
				
		return results;
	}
	
	/**
	 * Retrieve all outgoing properties of a given subject from the triple database;
	 * 
	 * Returns a list of WikiResource objects representing the (distinct) properties
	 * 
	 * @param subject
	 * @return
	 * @throws DBException
	 */
	public List<WikiResource> getPropertiesBySubjectObject(WikiResource subject, WikiResource object) throws DBException {
		ArrayList<WikiResource> results = new ArrayList<WikiResource>();

		// used for eliminating duplicates
		Set<String> seen = new HashSet<String>();
		
		Resource rs = ontology.getResource(subject.getUri());
		Resource ro = ontology.getResource(object.getUri());
		
		for(Iterator it = ontology.listStatements(rs,null,ro); it.hasNext(); ) {
			Statement st = (Statement)it.next();

			Property predicate = st.getPredicate();
            if(!seen.contains(predicate.getURI())) {
            	results.add(main.getResourceByUri(resolveURI(predicate)));
            	seen.add(predicate.getURI());
            }
		}
		
		
		return results;
	}
	
	/**
	 * Return all objects reachable from the given subject via the given property.
	 * 
	 * Returns a list of WikiResource objects representing the objects.
	 * 
	 * @param subject
	 * @param property
	 * @return
	 * @throws DBException
	 */
	public List<WikiNode> getObjectsBySubjectProperty(WikiResource subject, WikiResource property) throws DBException {
		ArrayList<WikiNode> results = new ArrayList<WikiNode>();

		Resource os = ontology.getResource(subject.getUri());
		Property op = ontology.getProperty(property.getUri());
		
		
		for(Iterator it = ontology.listStatements(os,op,(RDFNode)null); it.hasNext(); ) {
			Statement st = (Statement)it.next();

			RDFNode      ro = st.getObject(); 

			if(ro.isResource()) {
				WikiResource o = main.getResourceByUri(resolveURI((Resource)ro));
		
				results.add(o);
			} else if(ro.isLiteral()){
				Literal _ro = (Literal)ro;
            	WikiNode     o=new WikiLiteral(_ro.getString(),_ro.getLanguage());
            	
            	results.add(o);
            } else {
            	// TODO
            	log.warn("Warning: not adding literal resource!");
            }
		}
		
				
		
		return results;
		
	}
	
	
	public Collection<WikiLink> getTriplesBySubjectProperty(WikiResource subject, WikiResource property) throws DBException {
		Collection<WikiLink> results = new TreeSet<WikiLink>();

		
		Resource os = ontology.getResource(subject.getUri());
		Property op = ontology.getProperty(property.getUri());
		
		
		for(Iterator it = ontology.listStatements(os,op,(RDFNode)null); it.hasNext(); ) {
			Statement st = (Statement)it.next();

			RDFNode      ro = st.getObject(); 

			if(ro.isResource()) {
				WikiResource s = subject;
				WikiResource p = property;
				WikiResource o = main.getResourceByUri(resolveURI((Resource)ro));
		
				WikiLink l = new WikiLink(s,p,o,"");
            
				l.setImplicit(!ontology.isInBaseModel(st));
            
				results.add(l);
			} else if(ro.isLiteral()){
				Literal _ro = (Literal)ro;
				
            	WikiResource s=subject;
            	WikiResource p=property;
            	WikiNode     o=new WikiLiteral(_ro.getString(),_ro.getLanguage());
            	
            	results.add(new WikiLink(s,p,o,""));
            } else {
            	// TODO
            	log.warn("Warning: not adding literal resource!");
            }
		}

		return results;
	}

	
	public Collection<WikiLink> getTriplesByPropertyObject(WikiResource property, WikiResource object) throws DBException {
		Collection<WikiLink> results = new TreeSet<WikiLink>();

		Resource ro = ontology.getResource(object.getUri());
		Property rp = ontology.getProperty(property.getUri());
		
		for(Iterator it = ontology.listStatements(null,rp,ro); it.hasNext(); ) {
			Statement st = (Statement)it.next();

            WikiResource s=main.getResourceByUri(resolveURI(st.getSubject()));
            WikiResource p=property;
            WikiResource o=object;
		
            WikiLink l = new WikiLink(s,p,o,"");
            
            l.setImplicit(!ontology.isInBaseModel(st));
            
            results.add(l);
		}
		return results;
	}
	
	/**
	 * Return all objects reachable from the given subject via the given property.
	 * 
	 * Returns a list of WikiResource objects representing the objects.
	 * 
	 * @param subject
	 * @param property
	 * @return
	 * @throws DBException
	 */
	public List<WikiResource> getSubjectsByPropertyObject(WikiResource property, WikiResource object) throws DBException {
		ArrayList<WikiResource> results = new ArrayList<WikiResource>();

		
		Resource ro = ontology.getResource(object.getUri());
		Property rp = ontology.getProperty(property.getUri());
		
		for(Iterator it = ontology.listStatements(null,rp,ro); it.hasNext(); ) {
			Statement st = (Statement)it.next();

            WikiResource s=main.getResourceByUri(resolveURI(st.getSubject()));
            
            results.add(s);
		}
				
		return results;
	}
	
	
	/**
	 * Return all triples where the given resource appears as object.
	 * 
	 * Returns a list of WikiLink objects representing the triples.
	 * 
	 * @param object
	 * @return
	 * @throws DBException
	 */
	public Collection<WikiLink> getTriplesByObject(WikiResource object) throws DBException {
		Collection<WikiLink> results = new TreeSet<WikiLink>();

		Resource r = ontology.getResource(object.getUri());
		
		for(Iterator it = ontology.listStatements(null,null,r); it.hasNext(); ) {
			Statement st = (Statement)it.next();

            WikiResource s=main.getResourceByUri(resolveURI(st.getSubject()));
            WikiResource p=main.getResourceByUri(resolveURI(st.getPredicate()));
            WikiResource o=object;
		
            WikiLink l = new WikiLink(s,p,o,"");
            
            l.setImplicit(!ontology.isInBaseModel(st));
            
            results.add(l);
		}
		
		log.debug("number of triples: "+results.size());
		
		return results;
	}	
	
	
	
	/**
	 * Return all triples where the given resource appears as subject.
	 * 
	 * Returns a list of WikiLink objects representing the triples.
	 * 
	 * @param subject
	 * @return
	 * @throws DBException
	 */
	public Collection<WikiLink> getTriplesBySubject(WikiResource subject) throws DBException {
		Collection<WikiLink> results = new TreeSet<WikiLink>();

		
		Resource r = ontology.getResource(subject.getUri());
		
		for(Iterator it = ontology.listStatements(r,null,(RDFNode)null); it.hasNext(); ) {
			Statement st = (Statement)it.next();

			RDFNode      ro = st.getObject(); 

			if(ro.isResource()) {
				WikiResource s = subject;
				WikiResource p = main.getResourceByUri(resolveURI(st.getPredicate()));
				WikiResource o = main.getResourceByUri(resolveURI((Resource)ro));
		
				WikiLink l = new WikiLink(s,p,o,"");
            
				l.setImplicit(!ontology.isInBaseModel(st));
            
				results.add(l);
			} else if(ro.isLiteral()){
				Literal _ro = (Literal)ro;
				
            	WikiResource s=subject;
            	WikiResource p=main.getResourceByUri(resolveURI(st.getPredicate()));
            	WikiNode     o=new WikiLiteral(_ro.getString(),_ro.getLanguage());
            	
				WikiLink l = new WikiLink(s,p,o,"");

				l.setImplicit(!ontology.isInBaseModel(st));

            	results.add(l);
            } else {
            	// TODO
            	log.warn("Warning: not adding literal resource!");
            }
		}
		
		
		log.debug("number of triples: "+results.size());
		
		return results;
	}	

	/**
	 * Return all triples where the given resource appears as subject.
	 * 
	 * Returns a list of WikiLink objects representing the triples.
	 * 
	 * @param subject
	 * @return
	 * @throws DBException
	 */
	public Collection<WikiLink> getTriplesByProperty(WikiResource property) throws DBException {
		TreeSet<WikiLink> results = new TreeSet<WikiLink>();
		
		Property rp = ontology.getProperty(property.getUri());
		
		for(Iterator it = ontology.listStatements(null,rp,(RDFNode)null); it.hasNext(); ) {
			Statement st = (Statement)it.next();

			RDFNode      ro = st.getObject(); 

			if(ro.isResource()) {
				WikiResource s = main.getResourceByUri(resolveURI(st.getSubject()));
				WikiResource p = property;
				WikiResource o = main.getResourceByUri(resolveURI((Resource)ro));
		
				WikiLink l = new WikiLink(s,p,o,"");
            
				l.setImplicit(!ontology.isInBaseModel(st));
            
				results.add(l);
			} else if(ro.isLiteral()){
				Literal _ro = (Literal)ro;
				
            	WikiResource s=main.getResourceByUri(resolveURI(st.getSubject()));
            	WikiResource p=property;
            	WikiNode     o=new WikiLiteral(_ro.getString(),_ro.getLanguage());
            	
            	results.add(new WikiLink(s,p,o,""));
            } else {
            	// TODO
            	log.warn("Warning: not adding literal resource!");
            }

		}
		

		
		return results;
	}	
	
	public boolean hasTriple(WikiResource subject, WikiResource property, WikiResource object) {
		Resource rs = ontology.getResource(subject.getUri());
		Resource ro = ontology.getResource(object.getUri());
		Property rp = ontology.getProperty(property.getUri());
		
		Iterator it = ontology.listStatements(rs,rp,ro); 
		return it.hasNext();		
	}
	
	
	
	public byte[] getRDFXML() throws DBException {
		return serializeToFormat("RDF/XML");
		
	}
	
	public byte[] getN3() throws DBException {
		return serializeToFormat("N3");
	}

	public byte[] getTurtle() throws DBException {
		return serializeToFormat("TURTLE");
	}
	
	public byte[] getTriples() throws DBException {
		return serializeToFormat("N-TRIPLE");
	}
	
	private byte[] serializeToFormat(String format) throws DBException {
		ByteArrayOutputStream os = new ByteArrayOutputStream();				
		
		
		// create an RDF writer and allow "malformed" RDF/XML
		RDFWriter writer = ontology.getWriter(format);
		if(!format.equals("N-TRIPLE"))
			writer.setProperty("allowBadURIs","true");
		
		writer.write(ontology,os,main.getNamespaceBase().getUri());
		
		return os.toByteArray();
		
	}

	
	
	public Collection<WikiResource> getAllProperties() throws DBException {
		Collection<WikiResource> all_properties = new TreeSet<WikiResource>();
		
		
		// when inferencing is enabled, retrieve all properties from the ontology
		ExtendedIterator[] props;
		
		switch(getOntologyLevel()) {
		case OWL_RDFS:
		case OWL_DL:
		case OWL_FULL:
			ExtendedIterator[] myprops1 = { ontology.listObjectProperties() };
			props = myprops1;
			break;
		default:
			ExtendedIterator[] myprops2 = { ontology.listOntProperties() };
		props = myprops2;
		break;
		}
		
		for(ExtendedIterator it : props) {
			
			while(it.hasNext()) {
				Resource r = (Resource)it.next();
				try {
					all_properties.add(main.getResourceByUri(resolveURI(r)));
				} catch(DBException ex) {
					log.warn("unknown namespace for URI "+r);
				}
			}
		}
		
		return all_properties;		
	}
	/**
	 * Checks for all properties with all the parsed domain and range restrictions
	 * @param domains The domain constraints
	 * @param ranges the range constraints
	 * @return all valid properteis based on the constraints
	 * @author rwesten
	 */
	public Collection<WikiResource> getAllProperties(Collection<WikiResource> domains, Collection<WikiResource> ranges){
		Set<Namespace> namspaces = new HashSet<Namespace>();
		namspaces.add(new Namespace("rdfs",Namespace.NS_RDFS));
		namspaces.add(new Namespace("rdf",Namespace.NS_RDF));
		namspaces.add(new Namespace("owl",Namespace.NS_OWL));
		//first build the query and collect the differernt Namespaces
		StringBuffer queryString = new StringBuffer();
		queryString.append("SELECT ?prop\n");
		queryString.append("WHERE\n");
		queryString.append("  {\n");
		queryString.append("    ?prop rdf:type owl:ObjectProperty .\n");
		for(WikiResource domain : domains){
			namspaces.add(domain.getNamespace());
			queryString.append("    ?prop rdfs:domain "+domain.getQualifiedTitle()+" .\n");
		}
		for(WikiResource range : ranges){
			namspaces.add(range.getNamespace());
			queryString.append("    ?prop rdfs:range "+range.getQualifiedTitle()+" .\n");
		}
		queryString.append("  }\n");

		//scond add the used PREFIXs to the raw query
		StringBuffer SPARQL_Query = new StringBuffer();
		for(Namespace ns : namspaces){
			SPARQL_Query.append("PREFIX "+ns.getPrefix()+": <"+ns.getUri()+">\n");
		}
		SPARQL_Query.append(queryString);
		log.debug("SPARQL_Query:\n"+SPARQL_Query);
		//now the query is ready to be executed
		Set<WikiResource> props = new  HashSet<WikiResource>(); 
	  Query query = QueryFactory.create(SPARQL_Query.toString()) ;
	  QueryExecution qexec = QueryExecutionFactory.create(query, ontology) ; // Query the Ontology
	  ResultSet results = qexec.execSelect() ;
	  try {
	    while(results.hasNext()){
	      QuerySolution soln = results.nextSolution() ;
	      RDFNode propNode = soln.get("prop");
				try {
	        if(propNode.isResource()){
	        	props.add(main.getResourceByUri(resolveURI((Resource)propNode)));
	        }
				}catch(DBException ex) {
	  			log.warn("unknown namespace for URI "+propNode);
	  		}
	    }
	  } finally { qexec.close() ; }		
		return props;
	}
	
	public Collection<WikiResource> getAllPropertiesByNamespace(String ns_uri) throws DBException {
		Collection<WikiResource> results = new TreeSet<WikiResource>(getAllProperties());
		
		for(Iterator<WikiResource> it = results.iterator(); it.hasNext(); ) {
			WikiResource r = it.next();
			if(!r.getNamespace().getUri().equals(ns_uri)) {
				it.remove();
			}
		}
		
		return results;		
	}
	
	
	public Collection<WikiResource> getAllPropertiesByType(WikiResource type) throws DBException {
		SortedSet<WikiResource> results = new TreeSet<WikiResource>(getAllProperties());
		
		for(Iterator<WikiResource> it = results.iterator(); it.hasNext(); ) {
			WikiResource r = it.next();
			
			Collection<WikiResource> types = getTypes(r);
			
			if(!types.contains(type)) {
				it.remove();
			}
			
		}
		
		
		return results;		
	}

	/**
	 * Retrieve a list of all annotation properties. If the underlying ontology model does not support
	 * this kind of interence, this method only returns dublin core properties (if available).
	 * 
	 * @return
	 * @throws DBException
	 */
	public Collection<WikiResource> getAnnotationProperties() throws DBException {

		Collection<WikiResource> results = new TreeSet<WikiResource>();
		
		switch(getOntologyLevel()) {
		case OWL_RDFS:
		case OWL_DL:
		case OWL_FULL:
			for(Iterator it = ontology.listAnnotationProperties(); it.hasNext(); ) {
				Property r = (Property)it.next();
				results.add(main.getResourceByUri(resolveURI(r)));
			}
			break;
			
		default:
			String[] property_namespaces = { Namespace.NS_DC_CORE };		
		
			for(String ns : property_namespaces) {
				results.addAll(getAllPropertiesByNamespace(ns));
			}
			break;
		}
		
		
		return results;
	}
	
	
	/**
	 * Retrieve a list of all datatype properties for which the given domain resource is acceptable.
	 * If the ontology model does not support this kind of inference, this method returns an empty list.
	 * 
	 * @param domain
	 * @return
	 * @throws DBException
	 */
	public Collection<WikiResource> getDatatypePropertiesByDomain(WikiResource domain) throws DBException {

		switch(getOntologyLevel()) {
		case OWL_RDFS:
		case OWL_DL:
		case OWL_FULL:
			Collection<WikiResource> results = new TreeSet<WikiResource>();
			
			OntResource dom = ontology.getOntResource(domain.getUri());
			
			for(Iterator it = ontology.listDatatypeProperties(); it.hasNext(); ) {
				OntProperty r = (OntProperty)it.next();
				if(r.hasDomain(dom)) {
					results.add(main.getResourceByUri(resolveURI(r)));
				}
				
			}
			
			return results;
			
		default:
			return Collections.emptyList();
		}
	}
	
	
	
	/**
	 * Select the properties that are applicable for the given subject and object.
	 * 
	 * The method retrieves the list based on the ontology model in the following steps:
	 * 1. determine the types of subject and object
	 * 2. select all properties such that the domain matches one of the types of the subject 
	 *    and the range matches one of the types of the object
	 *    // TODO: properties without range/domain specification!
	 *  
	 * @param subject
	 * @param object
	 * @return
	 */	
	public Collection<WikiResource> getApplicableObjectProperties(WikiResource subject, WikiResource object) throws DBException {

		log.debug("subject: "+subject.getQualifiedTitle()+", object: "+object.getQualifiedTitle());
		
		switch(getOntologyLevel()) {
		case OWL_RDFS:
		case OWL_DL:
		case OWL_FULL:
			Collection<WikiResource> results = new TreeSet<WikiResource>();
			
			OntResource subj = ontology.getOntResource(subject.getUri());
			OntResource obj = ontology.getOntResource(object.getUri());
			
			log.debug("subj: "+subj+", obj: "+obj);
			
			for(Iterator it = ontology.listObjectProperties(); it.hasNext(); ) {
				OntProperty r = (OntProperty)it.next();

				
				// check whether the subject is an instance of the valid domains and the
				// object is an instance of the valid ranges
				boolean has_domain = false, has_range = false;
				
				Iterator it2;
				
				// check whether domain of r matches the subject
				it2 = r.listDomain();				
				if(!it2.hasNext()) {
					// every resource is applicable as domain
					has_domain = true;
				}
				
				while( it2.hasNext() && !has_domain ) {
					OntResource dom = (OntResource)it2.next();
					
					if(subj.hasRDFType(dom)) {
						has_domain = true;
					}
				}
				
				
				// check whether range of r matches the object
				it2 = r.listRange();
				if(!it2.hasNext()) {
					// every resource is applicable as range
					has_range = true;
				}
				
				while(it2.hasNext() && !has_range) {
					OntResource ran = (OntResource)it2.next();
					
					if(obj.hasRDFType(ran)) {
						has_range = true;
					}
				}
				
				
				if(has_domain && has_range) {
					results.add(main.getResourceByUri(resolveURI(r)));
				}
				
			}
						
			return results;
			
		default:
			return Collections.emptyList();
		}
	}
	
	
	public Collection<WikiResource> getTypes(WikiResource subject) throws DBException {
		return getTypes(subject,false);
	}
	
	public Collection<WikiResource> getDirectTypes(WikiResource subject) throws DBException {
		return getTypes(subject,true);
	}

	/**
	 * Get the types for a subject.
	 * 
	 * @param subject
	 * @param direct if true, only direct types are considered
	 * @return
	 * @throws DBException
	 */
	public Collection<WikiResource> getTypes(WikiResource subject, boolean direct) throws DBException {
		SortedSet<WikiResource> results  = new TreeSet<WikiResource>();
		

		OntResource r = ontology.getOntResource(subject.getUri());
		
		if( r != null ) {
			for( ExtendedIterator it = r.listRDFTypes(direct) ; it.hasNext(); ) {
				
				Resource type = (Resource)it.next();
				
				if(!type.isAnon()) {
					
					WikiResource w = main.getResourceByUri(type.getURI());
					
					Statement stmt = ontology.createStatement(r,ontology.getProperty(Namespace.NS_RDF+"type"),type);
					
					w.setImplicit(!ontology.isInBaseModel(stmt));
					
					results.add(w);
					
				}
			}
		}
		
	
        return results;
	}	
	
	
	public Collection<WikiResource> getAllTypes() throws DBException {
		Collection<WikiResource> all_types = new TreeSet<WikiResource>();
		
		
		OntResource r = ontology.getOntResource(Namespace.NS_RDFS+"Class");
		
		for(ExtendedIterator it = ontology.listIndividuals(r); it.hasNext(); ) {
			Resource r2 = (Resource)it.next();
			if(!r2.isAnon()) {
				all_types.add(main.getResourceByUri(resolveURI(r2)));
			}
		}	
		
		
		return all_types;		
	}

	
	public Iterable<WikiResource> getInstances(WikiResource type) throws DBException {
		final OntClass cls = ontology.getOntClass(type.getUri());
		if(cls != null) {
			return new Iterable<WikiResource>() {
				public Iterator<WikiResource> iterator() {
					return new Iterator<WikiResource>() {
						ExtendedIterator it = ontology.listIndividuals(cls);
						
						public boolean hasNext() {
							return it.hasNext();
						}

						public WikiResource next() {
							try {
								return main.getResourceByUri(resolveURI((Resource)it.next()));
							} catch(DBException ex) {
								log.error("error while retrieving individual",ex);
								return null;
							}
						}
						
						public void remove() {
							throw new UnsupportedOperationException();
						}
						
					};
				}
				
			};
		} else {
			return Collections.emptyList();
		}
	}
	
	
	@Deprecated
	public Collection<WikiResource> getDomainForProperty(WikiResource p) throws DBException {
		return getDomain(p);
	}
	
	
	/**
	 * Try to determine the language of the specified resource; currently, this method
	 * only evaluates the dc:language property
	 * @param r
	 * @return
	 * @throws DBException
	 */
	public Locale getLanguage(WikiResource r) throws DBException {
		WikiResource p_lang = main.getResourceByUri("http://purl.org/dc/elements/1.1/language");

		Collection<WikiNode> objects = getObjectsBySubjectProperty(r,p_lang);
		for(WikiNode n : objects) {
			if(n.isLiteral()) {
				WikiLiteral l = (WikiLiteral)n;
				return new Locale(l.getContent().trim());
			}
		}
		return Locale.getDefault();
	}
	
	
	/**
	 * Make an internal WikiLink external by storing it explicitly in the RDF repository
	 * @param l
	 */
	public void addExternalWikiLink(WikiLink l) {
		Property p;
		if(l.getProperty() == null) {
			p=ontology.createProperty(Namespace.NS_IKEWIKI_BASE+"WikiLink");
		} else {
			p=ontology.createProperty(l.getProperty().getUri());
		}
		Resource s=ontology.createResource(l.getSource().getUri());
		RDFNode o;
		
		if(l.getTarget().isResource()) {
			o=ontology.createResource(((WikiResource)l.getTarget()).getUri());
		} else {
			o=ontology.createLiteral(((WikiLiteral)l.getTarget()).getContent());
		}

		Statement triple = ontology.createStatement(s,p,o);
		ontology.add(triple);

		log.debug("added triple "+triple+", size now "+ontology.size());
		
		Property inverse = getInverseFor(p);
		if(inverse != null && l.getTarget().isResource()) {
			Statement triple2 = ontology.createStatement((Resource)o,inverse,s);
			ontology.add(triple2);			
			log.debug("added inverse triple "+triple2+", size now "+ontology.size());
		}
		
		if(isSymmetric(p) && o.isResource()) {
			Statement triple2 = ontology.createStatement((Resource)o,p,s);
			ontology.add(triple2);			
			log.debug("added symmetric triple "+triple2+", size now "+ontology.size());			
		}
		
	}
	
	/**
	 * Remove an external WikiLink from the RDF repository. This does not delete corresponding
	 * internal links if they exist (these are handled dynamically by the Wiki rendering engine).
	 * @param l
	 */
	public void removeExternalWikiLink(WikiLink l) {
		Property p;
		if(l.getProperty() == null) {
			p=ontology.createProperty(Namespace.NS_IKEWIKI_BASE+"WikiLink");
		} else {
			p=ontology.createProperty(l.getProperty().getUri());
		}
		Resource s=ontology.createResource(l.getSource().getUri());
		RDFNode o;
		
		if(l.getTarget().isResource()) {
			o=ontology.createResource(((WikiResource)l.getTarget()).getUri());
		} else {
			o=ontology.createLiteral(((WikiLiteral)l.getTarget()).getContent());
		}
			
		Statement triple = ontology.createStatement(s,p,o);
		if(ontology.contains(triple)) {
			if(ontology.isInBaseModel(triple)) {
				ontology.remove(triple);
				log.debug("removed triple "+triple+", size now "+ontology.size());
			} else {
				log.warn("tried to remove triple "+triple+" which is not in the base model");
			}
		} else {
			log.debug("triple "+triple+" not found, not removed");			
		}
		
		Property inverse = getInverseFor(p);
		if(inverse != null && l.getTarget().isResource()) {
			Statement triple2 = ontology.createStatement((Resource)o,inverse,s);
			ontology.remove(triple2);			
			log.debug("removed inverse triple "+triple2+", size now "+ontology.size());
		}
		
		if(isSymmetric(p) && o.isResource()) {
			Statement triple2 = ontology.createStatement((Resource)o,p,s);
			ontology.remove(triple2);			
			log.debug("removed symmetric triple "+triple2+", size now "+ontology.size());			
		}
	}
	
	/**
	 * Find the owl:inverseOf property of a given property
	 * @param r
	 * @return
	 */
	private Property getInverseFor(Property r) {
		
		switch(getOntologyLevel()) {
		case OWL_RDFS:
		case OWL_DL:
		case OWL_FULL:
			OntProperty p = ontology.createOntProperty(r.getURI());
			Property p1 = null; try { p1 = p.getInverseOf(); } catch(Exception ex) {}
			Property p2 = null; try { p2 = p.getInverse();   } catch(Exception ex) {}
			if(p1 != null) {
				return p1;
			} else {
				return p2;
			}
			
		default:
			
			Property p_inverseof = ontology.createProperty(Namespace.NS_OWL+"inverseOf");
			
			for(Iterator it = ontology.listStatements(r,p_inverseof,(RDFNode)null); it.hasNext(); ) {
				Statement stmt = (Statement)it.next();
				
				RDFNode rpn = stmt.getObject();
				if(rpn.isResource()) {
					return ontology.createProperty(((Resource)rpn).getURI());
				}
			}
		
			for(Iterator it = ontology.listStatements(null,p_inverseof,r); it.hasNext(); ) {
				Statement stmt = (Statement)it.next();
				
				Resource rpn = stmt.getSubject();
				return ontology.createProperty(rpn.getURI());
			}

			return null;
		}
	}
	
	private boolean isSymmetric(Property p) {
		
		
		switch(getOntologyLevel()) {
		case OWL_RDFS:
		case OWL_DL:
		case OWL_FULL:
			OntProperty op = ontology.createOntProperty(p.getURI());
			return op.isSymmetricProperty();
			
		default:
			Property _p = ontology.createProperty(Namespace.NS_RDF+"type");
			Resource _o = ontology.createResource(Namespace.NS_OWL+"SymmetricProperty");
			
			Iterator it = ontology.listStatements(p,_p,_o);
			
			return it.hasNext();
		}
	}
	
	private void storeLiteralForResource(Resource subject, Property property, String text) {

		if(text != null) {
			// then add a new statement with the article text as comment
			Literal object = ontology.createLiteral(text);
			storeTriple(subject,property,object);
		}	
	}

	private void storeXMLLiteralForResource(Resource subject, Property property, Document dom) {

		if(dom != null) {
			try {
				// serialise the DOM tree as XML without declaration
				XMLWriter writer = new CanonicalXMLWriter(dom);
				
				/*
				StringWriter out = new StringWriter();
				OutputFormat format = new OutputFormat(dom);
				format.setOmitXMLDeclaration(true);
				format.setPreserveSpace(true);
				format.setEncoding("utf-8");
				format.setOmitDocumentType(true);
				XMLSerializer output = new XMLSerializer(out, format);
				output.setNamespaces(true);
				output.serialize(dom);
				*/

				// then add a new statement with the article text as comment
				Literal object = ontology.createTypedLiteral(writer.writeToString(),"http://www.w3.org/1999/02/22-rdf-syntax-ns#XMLLiteral");
				storeTriple(subject,property,object);
			} catch (Exception e) {
				log.error("could not serialise XML content of article",e);
			}
		}	
	}

	
	private void storeTriple(Resource subject, Property property, RDFNode object) {
		List<RDFNode> remove = new ArrayList<RDFNode>();
		// first remove all comment links referred to by this resource
		for(NodeIterator it = ontology.listObjectsOfProperty(subject,property); it.hasNext(); ) {
			RDFNode node = it.nextNode();
			if(node.isLiteral()) {
				remove.add(node);
			}
		}		
		
		for(RDFNode n : remove) {
			Statement triple = ontology.createStatement(subject,property,n);
			if(inTransaction()) {
				transactionRemove(triple);
			} else {
				ontology.remove(triple);
			}
			log.debug("removed triple "+triple+", size now "+ontology.size());
		}

		//ontology.commit();

		if(object != null) {
			Statement triple = ontology.createStatement(subject,property,object);
			if(inTransaction()) {
				transactionAdd(triple);
			} else {
				ontology.add(triple);		
			}
			log.debug("added triple "+triple);			
		}
		
	}
	
	private void removeTriple(Resource subject, Property property, RDFNode object) {

		if(object != null) {
			Statement triple = ontology.createStatement(subject,property,object);
			if(inTransaction()) {
				transactionRemove(triple);
			} else {
				ontology.remove(triple);		
			}
			log.debug("removed <"+subject+","+property+","+object+"> for resource "+subject);			
		}
		
	}
	
	
	protected boolean inTransaction() {
		Thread current = Thread.currentThread();
		return transaction_log.get(current) != null && transaction_log.get(current).getActivityCount() > 0;
	}
	
	
	protected void transactionAdd(Statement stmt) {
		Thread current = Thread.currentThread();
		if(inTransaction()) {
			transaction_log.get(current).addAddition(stmt);
		}		
	}
	
	protected void transactionRemove(Statement stmt) {
		Thread current = Thread.currentThread();
		if(inTransaction()) {
			transaction_log.get(current).addRemoval(stmt);
		}		
	}

	/**
	 * Begin a transaction for the currently running thread. Subsequent calls of storeTriple will
	 * be stored in a journal instead of directly performed on the model and only commited when
	 * endTransaction is called.
	 * 
	 *
	 */
	public synchronized void beginTransaction() {
		Thread current = Thread.currentThread();
		if(transaction_log.get(current) != null) {
			transaction_log.get(current).increaseActivityCount();
		} else {
			transaction_log.put(current,new Transaction());
			log.debug("begin transaction for thread "+current.getId());
		}
			
	}
	
	/**
	 * End a transaction started with beginTransaction in the current Thread.
	 * 
	 * This will perform all changes made to the model in "batch mode"
	 *
	 */
	public synchronized void endTransaction() {
		
		// perform metadata changes in exactly the same order as they come; by default, this is set
		// to false, meaning that the metadata changes are reordered such that all removals occur first,
		// all additions afterwords, leading to two transactions in the metadata journal
		// change this to true if you experience transaction problems; this may lead to increased
		// splitting up of transactions in the metadata journal, though.
		boolean inOrder = false;
		
		
		Thread current = Thread.currentThread();
		transaction_log.get(current).decreaseActivityCount();
		if(transaction_log.get(current).getActivityCount() < 1) {
			
			Transaction transaction = transaction_log.get(current);
			if(transaction.getTransactionItems().size() > 0) {
				
				if(inOrder) {
					List<Statement> buf = new LinkedList<Statement>();
					List<TransactionItem> transactions = transaction.getTransactionItems();
					
					TransactionItem.KINDS previous_kind = transactions.get(0).getKind();

					for(TransactionItem ti : transactions) {
						if(ti.getKind() == previous_kind) {
							buf.add(ti.getStatement());
						} else {
							// flush and create a new buf
							if(previous_kind == TransactionItem.KINDS.ADD) {
								ontology.add(buf);
							} else if(previous_kind == TransactionItem.KINDS.REMOVE) {
								ontology.remove(buf);
							}
							buf = new LinkedList<Statement>();
							buf.add(ti.getStatement());
							previous_kind = ti.getKind();
						}
					}
					
					// flush
					if(previous_kind == TransactionItem.KINDS.ADD) {
						ontology.add(buf);
					} else if(previous_kind == TransactionItem.KINDS.REMOVE) {
						ontology.remove(buf);
					}
				} else {
					ontology.remove(transaction.getStatements(TransactionItem.KINDS.REMOVE));
					ontology.add(transaction.getStatements(TransactionItem.KINDS.ADD));
				}
				
				transaction_log.remove(current.getId());
				log.debug("end transaction for thread "+current.getId());
			}
		}
		
	}
	
	
	
	public Hashtable<String,String> getDatatypePropertiesByNamespace(WikiResource subject, String property_namespace) throws DBException {
		Hashtable<String,String> result = new Hashtable<String,String>();
		
		Collection<WikiLink> triples = getTriplesBySubject(subject);
	
		for(WikiLink l : triples) {
			WikiResource prop = l.getProperty();
			//log.debug("considering property "+prop.getNamespace()+", "+prop.getTitle());
			if(prop.getNamespace().getUri().equals(property_namespace) && l.getTarget().isLiteral()) {
				result.put(prop.getTitle(),((WikiLiteral)l.getTarget()).getContent());
			}
		}
		return result;
	}
	
	/**
	 * Store an article in the database. For RDFStore, this updates the rdfs:comment label.
	 * 
	 * @param a
	 */
	public void storeArticle(Article a) {

		//ontology.begin();
		beginTransaction();

		Resource subject = ontology.createResource(a.getUri());
		
		// internal links
		try {
			List<WikiLink> internal_references = main.getRenderingPipeline().getReferences(a);		
			
			Property p_navlink = ontology.createProperty(Namespace.NS_IW+"NavigationalLink");
			
			// first delete all NavigationalLinks that are not contained in internal_references
			for(StmtIterator it = ontology.listStatements(subject,p_navlink,(RDFNode)null); it.hasNext(); ) {
				Statement stmt = it.nextStatement();
				
				if(stmt.getObject().isURIResource()) {
					WikiResource r = main.getResourceByUri(((Resource)stmt.getObject()).getURI());
					
					if(!internal_references.contains(new WikiLink(a,null,r,""))) {
						transactionRemove(stmt);
					}
					
				}
				
			}
			
			// then add all new internal links
			for(WikiLink l : internal_references) {
				if(l.getTarget().isResource()) {
					Resource r_target = ontology.createResource(((WikiResource)l.getTarget()).getUri());
					
					if(!ontology.contains(subject,p_navlink,r_target)) {					
						storeTriple(subject,p_navlink,r_target);
					}
				}
			}
		
		
		} catch(DBException ex) {
			log.error("while parsing internal links",ex);
		}
		
		// metadata
		Property p_comment = ontology.createProperty("http://www.w3.org/2000/01/rdf-schema#comment");
//		Property p_description = ontology.createProperty("http://purl.org/dc/elements/1.1/description");
		Property p_title   = ontology.createProperty("http://purl.org/dc/elements/1.1/title");
		Property p_creator = ontology.createProperty("http://purl.org/dc/elements/1.1/creator");
		Property p_date    = ontology.createProperty("http://purl.org/dc/elements/1.1/date");
		Property p_language= ontology.createProperty("http://purl.org/dc/elements/1.1/language");
		Property p_xml     = ontology.createProperty(Namespace.NS_IKEWIKI_BASE+"article-xml");
		Property p_struct  = ontology.createProperty(Namespace.NS_IKEWIKI_BASE+"article-structured");
		Property p_hasauthor = ontology.createProperty(Namespace.NS_IKEWIKI_BASE+"hasAuthor");
					
		// rdfs:comment
		if(a.getComment() != null && !a.getComment().equals("")) 
			storeLiteralForResource(subject,p_comment,a.getComment());
		
		// dc:title
		if(a.getTitle() != null && !a.getTitle().equals("")) 
			storeLiteralForResource(subject,p_title,a.getTitle());
		
		// dc:description
//		if(a.getText() != null && !a.getText().equals("")) 
//			storeLiteralForResource(subject,p_description,a.getText());
		
		// dc:creator
		if(a.getCreator() != null) {
			storeLiteralForResource(subject,p_creator,a.getCreator());		

			OntResource r_user = ontology.createOntResource(Namespace.NS_IKEWIKI_USER + a.getCreator());
			storeTriple(subject,p_hasauthor,r_user);
		}
		
		
		// dc:date
		if(a.getTimestamp() != null) { 
			SimpleDateFormat df = new SimpleDateFormat("yyyy-MM-dd");
			storeLiteralForResource(subject,p_date,df.format(a.getTimestamp()));
		}
			
		// dc:language
		if(a.getLanguage() != null) 
			storeLiteralForResource(subject,p_language,a.getLanguage().getLanguage());

		
		// ikewiki:article-structured
		if(a.getText() != null && !a.getText().equals("")) 
			storeLiteralForResource(subject,p_struct,a.getText());
		
		// ikewiki:article-xml
		try {
			if(a.getBaseDom() == null)
				main.getRenderingPipeline().createDOM(a);
			if(a.getBaseDom() != null) 
				storeXMLLiteralForResource(subject,p_xml,a.getBaseDom());
		} catch(RenderingException ex) {}
		
		endTransaction();
		//ontology.commit();
	}

	
	/**
	 * Store image metadata in the RDF store.
	 * 
	 * @param subject
	 * @param property
	 * @param object
	 */
	public void storeImage(WikiImage img) {
		beginTransaction();
		
		OntResource subject = ontology.createOntResource(img.getUri());
		Resource t_image    = ontology.createResource(Namespace.NS_IKEWIKI_BASE+"Image");
		
		// set type to image so that the renderer can determine the type correctly
		subject.setRDFType(t_image);
		
		storeArticle(img);
		
		//ontology.begin();
		
		// Store EXIF metadata, if the exif ontology is loaded
		Resource exif_desc = model.createResource(Namespace.NS_EXIF+"imageDescription");
		
		if(model.containsResource(exif_desc)) {
		
			Property p_bytes = ontology.createProperty(Namespace.NS_EXIF+"bytes");
			storeTriple(subject,p_bytes,ontology.createLiteral(img.getBytes().length));
			
			try {
				
				ByteArrayInputStream img_in = new ByteArrayInputStream(img.getBytes());
				
				Metadata metadata = JpegMetadataReader.readMetadata(img_in);
				
				// iterate through metadata directories
				Iterator directories = metadata.getDirectoryIterator();
				while (directories.hasNext()) {
					Directory directory = (Directory)directories.next();
					// iterate through tags and print to System.out
					Iterator tags = directory.getTagIterator();
					while (tags.hasNext()) {
						Tag tag = (Tag)tags.next();
						// use Tag.toString()
						
						Property prop = getEXIFProperty(tag.getTagType());
						if(prop != null) {
							try {
								storeTriple(subject,prop,ontology.createLiteral(tag.getDescription()));
								log.debug("stored EXIF tag: "+tag);
							} catch(MetadataException ex) {
								log.error(ex);								
							}
						} else {
							log.warn("no property for tag "+tag.getTagName()+" (id: "+tag.getTagType()+")");
						}
					}
				}
				
			} catch(JpegProcessingException ex) {
				log.error(ex);
			}
		} else {
			log.info("not storing image metadata as EXIF ontology not loaded");
		}
		
		//ontology.commit();
		endTransaction();
		
	}
	
	public void storeUser(User user) throws DBException {
		String uri = Namespace.NS_IKEWIKI_USER + user.getLogin();
		
		beginTransaction();
		
		OntResource r_user = ontology.createOntResource(uri);
		OntResource t_user = ontology.createOntResource(Namespace.NS_IKEWIKI_BASE+"User");
		OntResource t_person = ontology.createOntResource(Namespace.NS_FOAF+"Person");
		OntProperty p_given   = ontology.createOntProperty(Namespace.NS_FOAF+"givenname");
		OntProperty p_surname = ontology.createOntProperty(Namespace.NS_FOAF+"surname");
		OntProperty p_mbox    = ontology.createOntProperty(Namespace.NS_FOAF+"mbox");
		
		
		if(ontology.containsResource(t_user))
			r_user.addRDFType(t_user);
		
		// add FOAF properties
		if(ontology.containsResource(t_person)) {
			r_user.addRDFType(t_person);
			
			
			storeLiteralForResource(r_user,p_given,user.getFirstName());
			storeLiteralForResource(r_user,p_surname,user.getLastName());

			// remove all old mbox entries
			StmtIterator it = ontology.listStatements(r_user,p_mbox,(RDFNode)null);
			while(it.hasNext()) {
				Statement stmt = it.nextStatement(); 
				removeTriple(stmt.getSubject(),stmt.getPredicate(),stmt.getObject());
			}
			
			// add new mbox entry
			OntResource r_mailto = ontology.createOntResource("mailto:"+user.getEmail());
			storeTriple(r_user,p_mbox,r_mailto);
		}
		
		endTransaction();
	}
	
	
	public void deleteUser(User user) throws DBException {
		WikiResource r = main.getResourceByUri(Namespace.NS_IKEWIKI_USER+user.getLogin());
		
		removeResource(r);
	}
	
	
	public void storeRole(Role role) throws DBException {
		String uri = Namespace.NS_IKEWIKI_ROLE + role.getName();
		
		beginTransaction();
		
		OntResource r_role = ontology.createOntResource(uri);
		OntResource t_role = ontology.createOntResource(Namespace.NS_IKEWIKI_BASE+"Role");
		OntResource t_group = ontology.createOntResource(Namespace.NS_FOAF+"Group");
		OntProperty p_name   = ontology.createOntProperty(Namespace.NS_FOAF+"name");
		OntProperty p_comment = ontology.createOntProperty(Namespace.NS_RDFS+"comment");
		
		if(ontology.containsResource(t_role))
			r_role.addRDFType(t_role);
		
		// add FOAF properties
		if(ontology.containsResource(t_group)) {
			r_role.addRDFType(t_group);
			
			storeLiteralForResource(r_role,p_name,role.getName());
			storeLiteralForResource(r_role,p_comment,role.getDescription());
		}
		
		endTransaction();		
	}
	
	public void assignRole(User user, Role role) throws DBException {
		beginTransaction();
		String user_uri = Namespace.NS_IKEWIKI_USER + user.getLogin();
		String role_uri = Namespace.NS_IKEWIKI_ROLE + role.getName();
		
		OntResource r_role = ontology.createOntResource(role_uri);
		OntResource r_user = ontology.createOntResource(user_uri);
		OntProperty p_hasrole = ontology.createOntProperty(Namespace.NS_IKEWIKI_BASE+"hasRole");

		storeTriple(r_user,p_hasrole,r_role);
		endTransaction();
	}
	
	
	public void retractRole(User user, Role role) throws DBException {
		beginTransaction();
		String user_uri = Namespace.NS_IKEWIKI_USER + user.getLogin();
		String role_uri = Namespace.NS_IKEWIKI_ROLE + role.getName();
		
		OntResource r_role = ontology.createOntResource(role_uri);
		OntResource r_user = ontology.createOntResource(user_uri);
		OntProperty p_hasrole = ontology.createOntProperty(Namespace.NS_IKEWIKI_BASE+"hasRole");

		removeTriple(r_user,p_hasrole,r_role);
		endTransaction();
	}
	
	/**
	 * Retrieve the EXIF property associated with the given tag id
	 * @param tag_id
	 * @return
	 */
	private Property getEXIFProperty(int tag_id) {
		Literal  obj  = ontology.createLiteral(tag_id);
		Property prop = ontology.createProperty(Namespace.NS_EXIF+"tagNumber");
		
		Iterator it = ontology.listSubjectsWithProperty(prop,obj);
		if(it.hasNext()) {
			Resource  subj = (Resource)it.next();
			return ontology.createProperty(subj.getURI());
		}
		return null;
	}
	

	/**
	 * Retrieve the tag id associated with an EXIF property
	 * @param tag_id
	 * @return
	 */
	public int getEXIFTagId(WikiResource wr_subj) {
		Resource subj = ontology.createResource(wr_subj.getUri());
		Property prop = ontology.createProperty(Namespace.NS_EXIF+"tagNumber");
		
		Iterator it = ontology.listObjectsOfProperty(subj,prop);
		while(it.hasNext()) {
			RDFNode  obj = (RDFNode)it.next();
			if(obj.isLiteral()) {
				Literal l = (Literal)obj;
				return l.getInt();
			}
		}
		return -1;
	}

	
	public void storeTriple(WikiResource subject, WikiResource property, WikiNode object) {
		Resource rdf_subject  = ontology.createResource(subject.getUri());
		Property rdf_property = ontology.createProperty(property.getUri());
		
		if(object.isLiteral()) {
			storeLiteralForResource(rdf_subject,rdf_property,((WikiLiteral)object).getContent());
		} else {
			Resource rdf_object = ontology.createResource(((WikiResource)object).getUri());
			storeTriple(rdf_subject,rdf_property,rdf_object);
		}
		
	}
	
	
	public void storeType(WikiResource subject, WikiResource object) {
		OntResource rdf_subject  = ontology.createOntResource(subject.getUri());
		//Property rdf_property = ontology.createProperty(Namespace.NS_RDF+"type");
		Resource rdf_object   = ontology.createResource(object.getUri());
		
		//Statement triple = ontology.createStatement(rdf_subject,rdf_property,rdf_object);
		//ontology.add(triple);		
		
		rdf_subject.addRDFType(rdf_object);
		
		log.debug("added new type for resource "+rdf_subject+": "+rdf_object);		
	}
	
	
	public void removeType(WikiResource subject, WikiResource object) {
		OntResource rdf_subject  = ontology.createOntResource(subject.getUri());
				
		Resource rdf_object   = ontology.createResource(object.getUri());
		
		if(rdf_subject.hasRDFType(rdf_object,true)) {
			rdf_subject.removeRDFType(rdf_object);				
			log.debug("removed type for resource "+rdf_subject+": "+rdf_object);
		} else {
			rdf_subject.removeRDFType(rdf_object);				
			log.warn("could not remove type for resource "+rdf_subject+": "+rdf_object+"; it is not in the base model");
		}
	}


	/**
	 * Update the caches for the changed statement stmt.
	 * 
	 * @param stmt
	 */
	private void updateCaches(Statement stmt) {
		// for the moment, no "sensitive" updating, but we can add/remove the statements from
		// the ontology model (see addedStatement and removedStatement below)!
		/*
		if(stmt.getPredicate().getURI().equals(Namespace.NS_RDFS+"subClassOf") ||
		   stmt.getPredicate().getURI().equals(Namespace.NS_RDFS+"subPropertyOf")) {
			all_properties = null;
			all_types = null;
			type_properties.clear();
			log.debug("invalidating type caches because new type was added.");
		}
		*/
		setDirty(true);
	}
	
	/**
	 * Invalidate all caches; this method forces to rebuild every cache and recomputation of the
	 * ontology model, and should thus be used only if strictly necessary.
	 *
	 */
	private void invalidateCaches() {
		/*
		all_types = null;
		all_properties = null;
		type_properties.clear();
		*/
		//ontology.rebind(); // update changes to ontology model
		setDirty(true);
	}
	
	
	// ModelChangedListener:
	
	public void addedStatement(Statement arg0) {
		updateCaches(arg0);
		try {
			main.getStorageBackend().logMetadataAdd(Collections.singletonList(arg0),null);
		} catch(DBException ex) {
			log.error("error storing metadata journal",ex);
		}
		//ontology.add(arg0);
	}


	public void addedStatements(List arg0) {
		for(Object o : arg0) {
			updateCaches((Statement)o);
		}
		try {
			main.getStorageBackend().logMetadataAdd(arg0,null);
		} catch(DBException ex) {
			log.error("error storing metadata journal",ex);
		}
	}


	public void addedStatements(Model arg0) {
		invalidateCaches();
		StmtIterator it = arg0.listStatements();
		List<Statement> l = new LinkedList<Statement>();
		while(it.hasNext()) {
			l.add(it.nextStatement());
		}
		try {
			main.getStorageBackend().logMetadataAdd(l,null);
		} catch(DBException ex) {
			log.error("error storing metadata journal",ex);
		}		
	}


	public void addedStatements(Statement[] arg0) {
		List<Statement> l = new LinkedList<Statement>();
		for(Statement s : arg0) {
			updateCaches(s);
			l.add(s);
		}
		try {
			main.getStorageBackend().logMetadataAdd(l,null);
		} catch(DBException ex) {
			log.error("error storing metadata journal",ex);
		}		
	}


	public void addedStatements(StmtIterator arg0) {
		List<Statement> l = new LinkedList<Statement>();
		while(arg0.hasNext()) {
			Statement s = arg0.nextStatement();
			updateCaches(s);
			l.add(s);
		}
		try {
			main.getStorageBackend().logMetadataAdd(l,null);
		} catch(DBException ex) {
			log.error("error storing metadata journal",ex);
		}		
	}


	public void notifyEvent(Model arg0, Object arg1) {
		// TODO Auto-generated method stub
		
	}


	public void removedStatement(Statement arg0) {
		updateCaches(arg0);
		try {
			main.getStorageBackend().logMetadataDelete(Collections.singletonList(arg0),null);
		} catch(DBException ex) {
			log.error("error storing metadata journal",ex);
		}
	}


	public void removedStatements(List arg0) {
		for(Object o : arg0) {
			updateCaches((Statement)o);
		}
		try {
			main.getStorageBackend().logMetadataDelete(arg0,null);
		} catch(DBException ex) {
			log.error("error storing metadata journal",ex);
		}
	}


	public void removedStatements(Model arg0) {
		invalidateCaches();
		StmtIterator it = arg0.listStatements();
		List<Statement> l = new LinkedList<Statement>();
		while(it.hasNext()) {
			l.add(it.nextStatement());
		}
		try {
			main.getStorageBackend().logMetadataDelete(l,null);
		} catch(DBException ex) {
			log.error("error storing metadata journal",ex);
		}		
	}


	public void removedStatements(Statement[] arg0) {
		List<Statement> l = new LinkedList<Statement>();
		for(Statement s : arg0) {
			updateCaches(s);
			l.add(s);
		}
		try {
			main.getStorageBackend().logMetadataDelete(l,null);
		} catch(DBException ex) {
			log.error("error storing metadata journal",ex);
		}		
	}


	public void removedStatements(StmtIterator arg0) {
		List<Statement> l = new LinkedList<Statement>();
		while(arg0.hasNext()) {
			Statement s = arg0.nextStatement();
			updateCaches(s);
			l.add(s);
		}
		try {
			main.getStorageBackend().logMetadataDelete(l,null);
		} catch(DBException ex) {
			log.error("error storing metadata journal",ex);
		}		
	}
	
	/**
	 * Return true of the resource can be viewed as a class in the ontology model.
	 * @param r
	 * @return
	 */
	public boolean isClass(WikiResource r) {
		return ontology.getOntClass(r.getUri()) != null;
	}
	
	/**
	 * Return true if the resource can be viewed as an object property in the ontology model.
	 * @param r
	 * @return
	 */
	public boolean isObjectProperty(WikiResource r) {
		return ontology.getObjectProperty(r.getUri()) != null;
	}
	
	/**
	 * Return true if the resource can be viewed as a datatype property in the ontology model.
	 * @param r
	 * @return
	 */
	public boolean isDatatypeProperty(WikiResource r) {
		return ontology.getDatatypeProperty(r.getUri()) != null;
	}

	/**
	 * Return true if the resource can be viewed as an annotation property in the ontology model.
	 * @param r
	 * @return
	 */
	public boolean isAnnotationProperty(WikiResource r) {
		return ontology.getAnnotationProperty(r.getUri()) != null;
	}
	
	
	
	public void addSuperClass(WikiResource subject, WikiResource superclass) throws DBException {
		OntClass cls  = ontology.getOntClass(subject.getUri());
		OntClass scls = ontology.getOntClass(superclass.getUri());
		cls.addSuperClass(scls);
	}
	

	public void removeSuperClass(WikiResource subject, WikiResource superclass) throws DBException {
		OntClass cls  = ontology.getOntClass(subject.getUri());
		OntClass scls = ontology.getOntClass(superclass.getUri());
		cls.removeSuperClass(scls);
	}
	
	
	public void addSuperProperty(WikiResource subject, WikiResource superclass) throws DBException {
		OntProperty cls  = ontology.getOntProperty(subject.getUri());
		OntProperty scls = ontology.getOntProperty(superclass.getUri());
		cls.addSuperProperty(scls);
	}


	public void removeSuperProperty(WikiResource subject, WikiResource superclass) throws DBException {
		OntProperty cls  = ontology.getOntProperty(subject.getUri());
		OntProperty scls = ontology.getOntProperty(superclass.getUri());
		cls.removeSuperProperty(scls);
	}

	
	/**
	 * Return a list of superclasses of the given resource.
	 * @param r the resource
	 * @param direct consider only direct superclasses
	 * @return
	 * @throws DBException
	 */
	public Collection<WikiResource> getSuperClasses(WikiResource r, boolean direct) throws DBException {
		OntClass cls = ontology.getOntClass(r.getUri());
		if(cls != null) {
			
			Collection<WikiResource> results = new TreeSet<WikiResource>();
			
			for(Iterator it = cls.listSuperClasses(direct); it.hasNext(); ) {
				OntClass superclass = (OntClass)it.next();
				results.add(main.getResourceByUri(resolveURI(superclass)));
			}
			
			return results;
		} else {
			return Collections.emptyList();
		}
	}
	
	public Collection<WikiResource> getSubClasses(WikiResource r, boolean direct) throws DBException {
		OntClass cls = ontology.getOntClass(r.getUri());
		if(cls != null) {
			
			Collection<WikiResource> results = new TreeSet<WikiResource>();
			
			for(Iterator it = cls.listSubClasses(direct); it.hasNext(); ) {
				OntClass superclass = (OntClass)it.next();
				results.add(main.getResourceByUri(resolveURI(superclass)));
			}
			
			return results;
		} else {
			return Collections.emptyList();
		}
	}
	

	
	/**
	 * Return a list of superproperties of the given resource.
	 * @param r the resource
	 * @param direct consider only direct superclasses
	 * @return
	 * @throws DBException
	 */
	public Collection<WikiResource> getSuperProperties(WikiResource r, boolean direct) throws DBException {
		OntProperty cls = ontology.getOntProperty(r.getUri());
		if(cls != null) {
			
			Collection<WikiResource> results = new TreeSet<WikiResource>();
			
			for(Iterator it = cls.listSuperProperties(direct); it.hasNext(); ) {
				OntProperty superclass = (OntProperty)it.next();
				results.add(main.getResourceByUri(resolveURI(superclass)));
			}
			
			return results;
		} else {
			return Collections.emptyList();
		}
	}

	
	/**
	 * Return a list of subproperties of the given resource.
	 * @param r the resource
	 * @param direct consider only direct superclasses
	 * @return
	 * @throws DBException
	 */
	public Collection<WikiResource> getSubProperties(WikiResource r, boolean direct) throws DBException {
		OntProperty cls = ontology.getOntProperty(r.getUri());
		if(cls != null) {
			
			Collection<WikiResource> results = new TreeSet<WikiResource>();
			
			for(Iterator it = cls.listSubProperties(direct); it.hasNext(); ) {
				OntProperty superclass = (OntProperty)it.next();
				results.add(main.getResourceByUri(resolveURI(superclass)));
			}
			
			return results;
		} else {
			return Collections.emptyList();
		}
	}
	
	
	/**
	 * Return a list of properties that are applicable on their left argument (rdfs:domain) to instances of the given 
	 * class.
	 * 
	 * @param r the resource identifying the class
	 * @param direct if true, consider also properties not directly declared for this class but inferred by the reasoner
	 * @return
	 * @throws DBException
	 */
	public Collection<WikiResource> getPropertiesByDomain(WikiResource r, boolean direct) throws DBException {
		OntClass cls = ontology.getOntClass(r.getUri());

		if(cls != null) {

			// TODO: the reasoner should really do this:
			HashSet<OntClass> classes = new HashSet<OntClass>();
			classes.add(cls);
			for(Iterator it = cls.listSuperClasses(); it.hasNext(); ) {
				classes.add((OntClass)it.next());
			}
			

			Collection<WikiResource> results = new TreeSet<WikiResource>();
			
			for(Iterator it = ontology.listOntProperties(); it.hasNext(); ) {
				OntProperty property = (OntProperty)it.next();
				for(OntClass c : classes) {
					if(property.hasDomain(c)) {
						results.add(main.getResourceByUri(resolveURI(property)));
						break;
					}
				}
			}
			
			return results;
		
		} else {
			return Collections.emptyList();
		}
	}
	
	/**
	 * Return a list of properties that are applicable on their right argument (rdfs:range) to instances of the given 
	 * class.
	 * 
	 * @param r the resource identifying the class
	 * @param direct if true, consider also properties not directly declared for this class but inferred by the reasoner
	 * @return
	 * @throws DBException
	 */
	public Collection<WikiResource> getPropertiesByRange(WikiResource r, boolean direct) throws DBException {
		OntClass cls = ontology.getOntClass(r.getUri());
		if(cls != null) {

			// TODO: the reasoner should really do this:
			HashSet<OntClass> classes = new HashSet<OntClass>();
			classes.add(cls);
			for(Iterator it = cls.listSuperClasses(); it.hasNext(); ) {
				classes.add((OntClass)it.next());
			}
			

			Collection<WikiResource> results = new TreeSet<WikiResource>();
			
			for(Iterator it = ontology.listOntProperties(); it.hasNext(); ) {
				OntProperty property = (OntProperty)it.next();
				for(OntClass c : classes) {
					if(property.hasRange(c)) {
						results.add(main.getResourceByUri(resolveURI(property)));
						break;
					}
				}
			}
			
			return results;
		
		} else {
			return Collections.emptyList();
		}
	}
	
	
	public void addDomain(WikiResource r, WikiResource dom) throws DBException {
		OntProperty prop = ontology.getOntProperty(r.getUri());
		OntClass    cls = ontology.getOntClass(dom.getUri());
		
		prop.addDomain(cls);
	}
	
	public void addRange(WikiResource r, WikiResource range) throws DBException {
		OntProperty prop = ontology.getOntProperty(r.getUri());
		OntClass    cls = ontology.getOntClass(range.getUri());
		
		prop.addRange(cls);
	}
	public void addInverseOf(WikiResource r, WikiResource range) throws DBException {
		OntProperty prop = ontology.getOntProperty(r.getUri());
		OntProperty invProp = ontology.getOntProperty(range.getUri());
		
		prop.addInverseOf(invProp);
	}
	
	public void removeDomain(WikiResource r, WikiResource dom) throws DBException {
		OntProperty prop = ontology.getOntProperty(r.getUri());
		OntClass    cls = ontology.getOntClass(dom.getUri());
		
		prop.removeDomain(cls);
	}
	
	public void removeRange(WikiResource r, WikiResource range) throws DBException {
		OntProperty prop = ontology.getOntProperty(r.getUri());
		OntClass    cls = ontology.getOntClass(range.getUri());
		
		prop.removeRange(cls);
	}
	public void removeInverseOf(WikiResource r, WikiResource range) throws DBException {
		OntProperty prop = ontology.getOntProperty(r.getUri());
		OntProperty invProp = ontology.getOntProperty(range.getUri());
		
		prop.removeInverseProperty(invProp);
	}

	
	public Collection<WikiResource> getRange(WikiResource r) throws DBException {
		OntProperty prop = ontology.getOntProperty(r.getUri());
		if(prop != null) {

			Collection<WikiResource> results = new TreeSet<WikiResource>();
			
			for(Iterator it = prop.listRange(); it.hasNext(); ) {
				OntResource cls = (OntResource)it.next();
				results.add(main.getResourceByUri(resolveURI(cls)));
			}
			
			return results;
		
		} else {
			return Collections.emptyList();
		}
		
	}
	

	public Collection<WikiResource> getDomain(WikiResource r) throws DBException {
		OntProperty prop = ontology.getOntProperty(r.getUri());
		if(prop != null) {

			Collection<WikiResource> results = new TreeSet<WikiResource>();
			
			for(Iterator it = prop.listDomain(); it.hasNext(); ) {
				OntResource cls = (OntResource)it.next();
				results.add(main.getResourceByUri(resolveURI(cls)));
			}
			
			return results;
		
		} else {
			return Collections.emptyList();
		}
		
	}
	
	/**
	 * For a property, return a collection of classes whose instances can be used as subjects 
	 * of statements with this property. This is related to getDomain(), but has a slightly 
	 * different semantics, because it lists for each class in getDomain also all subclasses.
	 * @param prop
	 * @return
	 * @throws DBException
	 */
	public Collection<WikiResource> getSubjectTypes(WikiResource prop) throws DBException {
		OntProperty p = ontology.getOntProperty(prop.getUri());
		if(p != null) {

			Collection<WikiResource> results = new TreeSet<WikiResource>();
			
			for(Iterator it = p.listDomain(); it.hasNext(); ) {
				OntResource cls = (OntResource)it.next();
				
				WikiResource c = main.getResourceByUri(resolveURI(cls)); c.setImplicit(false); 
				results.add(c);
				
				// if cls is a class, add all subclasses
				if(cls.isClass()) {
					for(Iterator it2 = ((OntClass)cls).listSubClasses(); it2.hasNext(); ) {
						OntClass cls2 = (OntClass)it2.next();
						WikiResource d = main.getResourceByUri(resolveURI(cls2)); 
						d.setImplicit(true); 
						results.add(d);
						
					}
				}
			}
			
			return results;
		
		} else {
			return Collections.emptyList();
		}		
	}
	
	/**
	 * For a property, return a collection of classes whose instances can be used as objects 
	 * of statements with this property. This is related to getRange(), but has a slightly 
	 * different semantics, because it lists for each class in getRange also all subclasses.
	 * @param prop
	 * @return
	 * @throws DBException
	 */
	public Collection<WikiResource> getObjectTypes(WikiResource prop) throws DBException {
		OntProperty p = ontology.getOntProperty(prop.getUri());
		if(p != null) {

			Collection<WikiResource> results = new TreeSet<WikiResource>();
			
			for(Iterator it = p.listRange(); it.hasNext(); ) {
				OntResource cls = (OntResource)it.next();
				
				WikiResource c = main.getResourceByUri(resolveURI(cls)); c.setImplicit(false); 
				results.add(c);
				
				// if cls is a class, add all subclasses
				if(cls.isClass()) {
					for(Iterator it2 = ((OntClass)cls).listSubClasses(); it2.hasNext(); ) {
						OntClass cls2 = (OntClass)it2.next();
						WikiResource d = main.getResourceByUri(resolveURI(cls2)); 
						d.setImplicit(true); 
						results.add(d);
						
					}
				}
			}
			
			return results;
		
		} else {
			return Collections.emptyList();
		}		
	}

	
	/**
	 * Execute a SPARQL query on the model and return an XML document containing the results in
	 * SPARQL result format.
	 * 
	 * All existing prefixes are added automatically if necessary.
	 * 
	 * @param query
	 * @return
	 * @throws DBException
	 */
	public Document executeSparqlQuery(String queryString) throws DBException {
		
		StringBuilder myQueryString = new StringBuilder();
		
		// add all prefixes to the query string
		for( Namespace ns : main.getStorageBackend().listNamespaces() ) {
			if(ns.getPrefix().equals("")) {
				myQueryString.append("PREFIX default: <"+ns.getUri()+">\n");
			} else {
				myQueryString.append("PREFIX "+ns.getPrefix()+": <"+ns.getUri()+">\n");
			}
		}
		myQueryString.append(queryString);
			
		// execute query 
		Query query = QueryFactory.create(myQueryString.toString()) ;
		QueryExecution qexec = QueryExecutionFactory.create(query, ontology) ;
		ResultSet results = qexec.execSelect() ;

		ByteArrayOutputStream bOut = new ByteArrayOutputStream();
		ResultSetFormatter.outputAsXML(bOut,results) ;
		
		try {
			// create document
			DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
			factory.setNamespaceAware(true);
			
			DocumentBuilder parser = factory.newDocumentBuilder();
			
			ByteArrayInputStream bIn = new ByteArrayInputStream(bOut.toByteArray());
			Document d = parser.parse(bIn);
			
			return d;
			
		} catch(ParserConfigurationException ex) {
			throw new DBException(ex);
		} catch(SAXException ex) {
			throw new DBException(ex);
		} catch(IOException ex) {
			throw new DBException(ex);
		}
			
	}
	
	
	
	/**
	 * Completely remove a resource from the repository.
	 * 
	 * This method will remove all statements from the repository that contain the given
	 * resource as subject, property, or object.
	 * 
	 * @param r
	 */
	public void removeResource(WikiResource r) {
		
		//Resource res = model.getResource(r.getUri());
		
		
		for(StmtIterator it = model.listStatements(); it.hasNext(); ) {
			Statement stmt = it.nextStatement();
			try {
				Resource subject   = stmt.getSubject();
				Resource predicate = stmt.getPredicate();
				RDFNode  object    = stmt.getObject();
				
				if(subject.isURIResource() && subject.getURI().equals(r.getUri())) {
					log.info("deleted statement "+stmt.toString());
					it.remove();
				} else if(predicate.isURIResource() && predicate.getURI().equals(r.getUri())) { 
					log.info("deleted statement "+stmt.toString());
					it.remove();
				} else if(object.isURIResource() && ((Resource)object).getURI().equals((r.getUri()))) {
					log.info("deleted statement "+stmt.toString());
					it.remove();
				}
			} catch(Exception ex) {
				log.warn("exception while deleting",ex);
			}
			
		}
		ontology.rebind();
		
	}
	
	
	
	private void setDirty(boolean v) {
		dirty = v;
	}
	
	
	private synchronized void save() {
		if(dirty) {
			try {
				log.info("saving model back to database ...");
				long start = System.currentTimeMillis();
				synchronized(model) {
					
//					// mark all non-persistent triples in the journal as persistent
					main.getStorageBackend().flushMetadataJournal(); 
					dirty = false;

					// unfortunately, Jena models can contain information that leads to failures
					// when saving; we therefore first test whether we can create a memory copy of
					// the model and only delete the persistent storage afterwards
					//ModelMaker maker = ModelFactory.createMemModelMaker();
					//Model _tmpmodel   = maker.createDefaultModel();
					//_tmpmodel.add(model);
					checkSanity(model);
					
					
					// remove database model from database
					dbmodel.remove();
					dbmodel.close();
					
					// create it anew and fill with the contents of the memory model 
					dbmodel = ModelRDB.createModel(conn);
					dbmodel.add(model);
				}
				long end = System.currentTimeMillis();
				log.info("saving finished after "+(end-start)+" milliseconds");
			} catch(UnsupportedOperationException ex) {
				log.error("inconsistent Jena model, could not save (kept old state of database)");
			} catch(Exception ex) {
				// for any exception, we want to ensure that the save process terminates so that
				// further processing can be done; we therefore simply log exceptions here
				log.error(ex);
			}
		} else {
			log.debug("not saving model; nothing has changed.");
		}
	}

	/**
	 * Check whether the model is "sane", i.e. does not contain nodes that cannot be written out.
	 * 
	 * Contributed by Chris Dollin, jena-dev mailinglist.
	 * @param m
	 */
	public void checkSanity( Model m )
	{
		StmtIterator s = m.listStatements();
		while (s.hasNext()) 
			if (bad( s.nextStatement() ) ) s.remove();
	}
	
	public boolean bad( Statement s )
	{
		if (
				bad( s.getSubject() ) 
				|| bad( s.getPredicate() )
				|| bad( s.getObject() )
		)
		{
			// do something, and return true to delete the
			// statement, false to preserve it
			try {
				log.error("found inconsistent node, deleting statement" + s.toString());
			} catch(Exception ex) {
				log.error("found inconsistent node, deleting statement (could not print statement)");				
			}
			return true;
		}
		else
			return false;
	}
	
	public boolean bad( RDFNode n )
	{
		Node x = n.asNode();
		return x instanceof Node_NULL || x instanceof Node_Fluid;
	}


	
	
	@Override
	protected void finalize() throws Throwable {
		close();
		super.finalize();
	}


	public void close() {
		maintanance.halt();
		// force saving of model		
		save();
		try {
			conn.close();
		} catch(Exception ex) {
			log.error("error while closing database connection",ex);
		}
	}
	
	
	
	
	/**
	 * A thread for saving the in-memory model back to the database from time to time 
	 * (currently every 5 minutes or upon trigger).
	 * 
	 * @author sschaffe
	 *
	 */
	public static class ModelMaintanance extends Thread {
		
		private static Log log = LogFactory.getLog(ModelMaintanance.class);

		private RDFStore store;
		
		private static final int DEFAULTINTERVAL = 5*60*1; 
		
		private int INTERVAL;
		
		private boolean do_halt = false;
		
		private ModelMaintanance(RDFStore store) {
			super();
			this.store = store;
			INTERVAL = DEFAULTINTERVAL;
			this.start();
		}
		
		public void setInterval(int new_interval) {
			INTERVAL = new_interval;
			this.interrupt();
		}
		
		public void halt() {
			this.do_halt = true;
			this.interrupt();
			log.info("model maintanance thread terminated normally.");
		}
		
		public void run() {
			try {
				while(true) {
					try {
						Thread.sleep(INTERVAL * 1000);
					} catch(InterruptedException ex) {
						if(do_halt) {
							break;
						}						
					}
					store.save();
				}
			} finally {
				// save when thread is destroyed
				store.save();
			}
		}
		
	}


	public ModelMaintanance getModelMaintanance() {
		return maintanance;
	}
	
	
}
