/**
 *  Copyright (C) 2005
 *  Sebastian Schaffert and Rupert Westenthaler
 *  Salzburg Research Forschungsgesellschaft m.b.H.
 *
 *  IkeWiki is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU Lesser General Public
 * 	License as published by the Free Software Foundation; either
 * 	version 2 of the License, or (at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * 	Lesser General Public License for more details.
 *
 * 	You should have received a copy of the GNU Lesser General Public
 * 	License along with this library; if not, write to the
 * 	Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * 	Boston, MA 02111-1307, USA.
 * 
 */
package at.srfg.ikewiki.db;

import java.lang.ref.SoftReference;
import java.util.Collection;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Document;

import at.srfg.ikewiki.Main;
import at.srfg.ikewiki.data.Article;
import at.srfg.ikewiki.data.Namespace;
import at.srfg.ikewiki.data.Permission;
import at.srfg.ikewiki.data.Role;
import at.srfg.ikewiki.data.User;
import at.srfg.ikewiki.data.Variable;
import at.srfg.ikewiki.data.WikiImage;
import at.srfg.ikewiki.data.WikiLink;
import at.srfg.ikewiki.data.WikiNode;
import at.srfg.ikewiki.data.WikiResource;

import com.hp.hpl.jena.ontology.OntModel;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.Statement;

/**
 * A unified storage backend for accessing the RDF and DB store.
 * 
 * @author sschaffe
 *
 */
public class StorageBackend {
	
	private DBStore dbstore;
	private RDFStore rdfstore;
	private LuceneStore lucenestore;
	
	private Hashtable<String,SoftReference<Article>> article_cache;
	private Hashtable<String,SoftReference<Namespace>> ns_prefix_cache;
	private Hashtable<String,SoftReference<Namespace>> ns_uri_cache;
	private Hashtable<String,SoftReference<Boolean>> exists_article_cache;
	private Hashtable<String,SoftReference<Boolean>> exists_image_cache;
	
	private Main main;
	
	private static Log log = LogFactory.getLog(StorageBackend.class);
	
	
	private List<ResourceChangeListener> listeners;
	
	public StorageBackend(Main main) throws DBException {
		this.dbstore = new DBStore(main);
		this.rdfstore = new RDFStore(main);
		this.rdfstore.registerNamespaces(dbstore);
		this.lucenestore = new LuceneStore(main);
		
		this.article_cache = new Hashtable<String,SoftReference<Article>>();
		this.ns_prefix_cache = new Hashtable<String,SoftReference<Namespace>>();
		this.ns_uri_cache = new Hashtable<String,SoftReference<Namespace>>();
		
		this.exists_article_cache = new Hashtable<String,SoftReference<Boolean>>();
		this.exists_image_cache = new Hashtable<String,SoftReference<Boolean>>();
		
		listeners = new LinkedList<ResourceChangeListener>();
		
		this.main = main;
	}

	
	
	public void init() {
		lucenestore.initIndexer();
	}
	
	
	
	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.LuceneStore#reIndex()
	 */
	public void reIndex() {
		lucenestore.reIndex();
	}



	public void addResourceChangeListener(ResourceChangeListener l) {
		listeners.add(l);
	}
	
	public void removeResourceChangeListener(ResourceChangeListener l) {
		listeners.remove(l);
	}
	
	
	private void notifyResourceChange(WikiResource r) {
		// deal with our own changes first!
		this.article_cache.remove(r.getNamespace().getPrefix()+r.getTitle());
		this.exists_article_cache.remove(r.getQualifiedTitle());
		this.exists_image_cache.remove(r.getQualifiedTitle());
		
		for(ResourceChangeListener l : listeners) {
			l.changeEvent(r);
		}
	}
	
	private void notifyUserChange(User u) {
		try {
			WikiResource r = main.getResourceByUri(Namespace.NS_IKEWIKI_USER+u.getLogin());

			notifyResourceChange(r);
		} catch(DBException ex) {
			log.error(ex);
		}
	}
	
	private void notifyRoleChange(Role role) {
		try {
			WikiResource r = main.getResourceByUri(Namespace.NS_IKEWIKI_ROLE+role.getName());

			notifyResourceChange(r);
		} catch(DBException ex) {
			log.error(ex);
		}
		
	}
	
	
	private void notifyAllChange() {
		// deal with our own changes first!
		this.article_cache.clear();
		this.exists_article_cache.clear();
		this.exists_image_cache.clear();
		
		for(ResourceChangeListener l : listeners) {
			l.allChangeEvent();
		}
	}
	
	
	/**
	 * @return Returns the dbstore.
	 */
	public DBStore getDBStore() {
		return dbstore;
	}




	/**
	 * @return Returns the rdfstore.
	 */
	public RDFStore getRDFStore() {
		return rdfstore;
	}




	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.RDFStore#addTripleData(java.lang.String)
	 */
	public void addTripleData(String rdf_data) throws DBException {
		rdfstore.addTripleData(rdf_data);
	}





	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#ensureNamespace(at.srfg.ikewiki.data.Namespace)
	 */
	public String ensureNamespace(Namespace n) throws DBException {
		ns_prefix_cache.clear();
		ns_uri_cache.clear();
		return dbstore.ensureNamespace(n);
	}


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#existsArticleByNamespaceTitle(java.lang.String, java.lang.String)
	 */
	public boolean existsArticleByNamespaceTitle(String ns_prefix, String title) throws DBException {
		SoftReference<Boolean> ref = exists_article_cache.get(ns_prefix+title);
		if(ref != null && ref.get() != null) {
			return ref.get();
		} else {
			boolean result = dbstore.existsArticleByNamespaceTitle(ns_prefix, title);
			exists_article_cache.put(ns_prefix+":"+title,new SoftReference<Boolean>(result));
			return result;
		}
	}

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#getArticleByNamespaceTitle(java.lang.String, java.lang.String)
	 */
	public Article getArticleByNamespaceTitle(String ns_prefix, String title) throws DBException {
		SoftReference<Article> ref = article_cache.get(ns_prefix+title);
		if(ref != null && ref.get() != null) {
			//log.debug("cache hit for article "+ns_prefix+":"+title);
			return ref.get();
		} else {
			Article a = dbstore.getArticleByNamespaceTitle(ns_prefix, title);
			article_cache.put(ns_prefix+title,new SoftReference<Article>(a));
			//log.debug("cache miss for article "+ns_prefix+":"+title);
			return a; 
		}		
	}

	public void cacheArticle(Article a) {
		article_cache.put(a.getNamespace().getPrefix()+a.getTitle(),new SoftReference<Article>(a));
	}
	
	
	public Article getArticleByRevisionId(String ns_prefix, String title, long revision_id) throws DBException {
		return dbstore.getArticleByRevisionId(ns_prefix, title, revision_id);
	}




	public List<Article> getRevisionsByNamespaceTitle(String ns_prefix, String title, int limit, int offset) throws DBException {
		return dbstore.getRevisionsByNamespaceTitle(ns_prefix, title, limit, offset);
	}




	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#getNamespaceByPrefix(java.lang.String)
	 */
	public Namespace getNamespaceByPrefix(String ns_prefix) throws DBException {
		SoftReference<Namespace> ref = ns_prefix_cache.get(ns_prefix);
		if(ref != null && ref.get() != null) {
			//log.debug("cache hit for namespace "+ns_prefix);
			return ref.get();
		} else {
			Namespace ns = dbstore.getNamespaceByPrefix(ns_prefix);
			ns_prefix_cache.put(ns_prefix,new SoftReference<Namespace>(ns));
			//log.debug("cache miss for namespace "+ns_prefix);
			return ns; 
		}		
	}

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#getNamespaceByUri(java.lang.String)
	 */
	public Namespace getNamespaceByUri(String uri) throws DBException {
		SoftReference<Namespace> ref = ns_uri_cache.get(uri);
		if(ref != null && ref.get() != null) {
			//log.debug("cache hit for namespace "+uri);
			return ref.get();
		} else {
			Namespace ns = dbstore.getNamespaceByUri(uri);
			ns_uri_cache.put(uri,new SoftReference<Namespace>(ns));
			//log.debug("cache miss for namespace "+uri);
			return ns; 
		}		
	}


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#storeArticle(at.srfg.ikewiki.data.Article)
	 */
	public void storeArticle(Article a) throws DBException {
		// update database (with versioning of content) and RDF triple store	
		//article_cache.clear();
		dbstore.storeArticle(a);
		rdfstore.storeArticle(a);
		lucenestore.storeArticle(a);
		//notifyResourceChange(a);
		notifyAllChange();
	}
	
	

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#storeTriples(com.hp.hpl.jena.rdf.model.Model)
	 */
	public void storeTriples(Model m, User u) throws DBException {
		dbstore.storeTriples(m, u);
		notifyAllChange();
	}




	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.RDFStore#getObjectsBySubjectProperty(at.srfg.ikewiki.data.WikiResource, at.srfg.ikewiki.data.WikiResource)
	 */
	public List<WikiNode> getObjectsBySubjectProperty(WikiResource subject, WikiResource property) throws DBException {
		return rdfstore.getObjectsBySubjectProperty(subject, property);
	}

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.RDFStore#getPropertiesByObject(at.srfg.ikewiki.data.WikiResource)
	 */
	public List<WikiResource> getPropertiesByObject(WikiResource object) throws DBException {
		return rdfstore.getPropertiesByObject(object);
	}

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.RDFStore#getPropertiesBySubject(at.srfg.ikewiki.data.WikiResource)
	 */
	public List<WikiResource> getPropertiesBySubject(WikiResource subject) throws DBException {
		return rdfstore.getPropertiesBySubject(subject);
	}

	
	
	public List<WikiResource> getPropertiesBySubjectObject(WikiResource subject, WikiResource object) throws DBException {
		return rdfstore.getPropertiesBySubjectObject(subject, object);
	}




	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.RDFStore#getAllProperties()
	 */
	public Collection<WikiResource> getAllProperties() throws DBException {
		return rdfstore.getAllProperties();
	}
  
	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.RDFStore#getAllProperties(java.util.Collection,java.util.Collection)
	 */
	public Collection<WikiResource> getAllProperties(Collection<WikiResource> domains, Collection<WikiResource> ranges){
		return rdfstore.getAllProperties(domains, ranges);
	}



	public boolean hasTriple(WikiResource subject, WikiResource property, WikiResource object) {
		return rdfstore.hasTriple(subject, property, object);
	}


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.RDFStore#getRDFXML()
	 */
	public byte[] getRDFXML() throws DBException {
		return rdfstore.getRDFXML();
	}

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.RDFStore#getSubjectsByPropertyObject(at.srfg.ikewiki.data.WikiResource, at.srfg.ikewiki.data.WikiResource)
	 */
	public List<WikiResource> getSubjectsByPropertyObject(WikiResource property, WikiResource object) throws DBException {
		return rdfstore.getSubjectsByPropertyObject(property, object);
	}

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.RDFStore#getTriplesByObject(at.srfg.ikewiki.data.WikiResource)
	 */
	public Collection<WikiLink> getTriplesByObject(WikiResource object) throws DBException {
		return rdfstore.getTriplesByObject(object);
	}
	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.RDFStore#getTriplesByProperty(at.srfg.ikewiki.data.WikiResource)
	 */
	public Collection<WikiLink> getTriplesByProperty(WikiResource property) throws DBException {
		return rdfstore.getTriplesByProperty(property);
	}

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.RDFStore#getTriplesBySubject(at.srfg.ikewiki.data.WikiResource)
	 */
	public Collection<WikiLink> getTriplesBySubject(WikiResource subject) throws DBException {
		return rdfstore.getTriplesBySubject(subject);
	}

	
	public Collection<WikiLink> getTriplesByPropertyObject(WikiResource property, WikiResource object) throws DBException {
		return rdfstore.getTriplesByPropertyObject(property, object);
	}


	public Collection<WikiLink> getTriplesBySubjectProperty(WikiResource subject, WikiResource property) throws DBException {
		return rdfstore.getTriplesBySubjectProperty(subject, property);
	}


	/**
	 * Make an internal WikiLink external by storing it explicitly in the RDF repository
	 * @param l
	 */
	public void addExternalWikiLink(WikiLink l) {
		// TODO
		rdfstore.addExternalWikiLink(l);
		notifyResourceChange(l.getSource());
		if(l.getTarget().isResource()) {
			notifyResourceChange((WikiResource)l.getTarget());
		}
	}
	
	/**
	 * Remove an external WikiLink from the RDF repository. This does not delete corresponding
	 * internal links if they exist (these are handled dynamically by the Wiki rendering engine).
	 * @param l
	 */
	public void removeExternalWikiLink(WikiLink l) {
		rdfstore.removeExternalWikiLink(l);
		notifyResourceChange(l.getSource());
		if(l.getTarget().isResource()) {
			notifyResourceChange((WikiResource)l.getTarget());
		}
	}




	public User getUser(String login, String password) throws DBException {
		return dbstore.getUser(login, password);
	}




	public void storeUser(User u, String password) throws DBException {
		dbstore.storeUser(u, password);
		rdfstore.storeUser(u);
		notifyUserChange(u);
	}




	public void updateUser(User u) throws DBException {
		dbstore.updateUser(u);
		rdfstore.storeUser(u);
		notifyUserChange(u);
	}

	

	


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#deleteUser(at.srfg.ikewiki.data.User)
	 */
	public void deleteUser(User user) throws DBException {
		dbstore.deleteUser(user);
		rdfstore.deleteUser(user);

		WikiResource r = main.getResourceByUri(Namespace.NS_IKEWIKI_USER+user.getLogin());
		dbstore.deleteArticle(r);
		
		notifyUserChange(user);
	}


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#getRole(java.lang.String)
	 */
	public Role getRole(String name) throws DBException {
		return dbstore.getRole(name);
	}


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#storeRole(at.srfg.ikewiki.data.Role)
	 */
	public void storeRole(Role role) throws DBException {
		dbstore.storeRole(role);
		rdfstore.storeRole(role);
		notifyRoleChange(role);
	}
	
	
	


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#deleteRole(at.srfg.ikewiki.data.Role)
	 */
	public void deleteRole(Role role) throws DBException {
		dbstore.deleteRole(role);
		notifyRoleChange(role);
	}


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#updateRole(at.srfg.ikewiki.data.Role)
	 */
	public void updateRole(Role role) throws DBException {
		dbstore.updateRole(role);
		notifyRoleChange(role);
	}


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#assignRole(at.srfg.ikewiki.data.User, at.srfg.ikewiki.data.Role)
	 */
	public void assignRole(User user, Role role) throws DBException {
		dbstore.assignRole(user, role);
		rdfstore.assignRole(user,role);
		notifyUserChange(user);
		notifyRoleChange(role);
	}


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#listRoles()
	 */
	public Collection<Role> listRoles() throws DBException {
		return dbstore.listRoles();
	}


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#listRoles(at.srfg.ikewiki.data.User)
	 */
	public Collection<Role> listRoles(User user) throws DBException {
		return dbstore.listRoles(user);
	}


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#retractRole(at.srfg.ikewiki.data.User, at.srfg.ikewiki.data.Role)
	 */
	public void retractRole(User user, Role role) throws DBException {
		dbstore.retractRole(user, role);
		rdfstore.retractRole(user,role);
		notifyUserChange(user);
		notifyRoleChange(role);
	}


	public Collection<Namespace> listNamespaces() throws DBException {
		return dbstore.listNamespaces();
	}




	public void deleteNamespace(String uri) throws DBException {
		ns_prefix_cache.clear();
		ns_uri_cache.clear();
		dbstore.deleteNamespace(uri);
		rdfstore.deleteNamespace(uri);
	}




	public void updateNamespace(Namespace n) throws DBException {
		ns_prefix_cache.clear();
		ns_uri_cache.clear();
		dbstore.updateNamespace(n);
		rdfstore.updateNamespace(n);
	}




	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#listArticles()
	 */
	public Iterable<Article> listArticles() throws DBException {
		return dbstore.listArticles();
	}


	public List<Article> searchArticles(String keyword,int limit, int offset) throws DBException {
		return lucenestore.searchArticles(keyword,limit,offset);
	}


	public Collection<WikiResource> getAllPropertiesByNamespace(String ns_uri) throws DBException {
		return rdfstore.getAllPropertiesByNamespace(ns_uri);
	}




	public Collection<WikiResource> getAllPropertiesByType(WikiResource type) throws DBException {
		return rdfstore.getAllPropertiesByType(type);
	}




	public Collection<WikiResource> getAnnotationProperties() throws DBException {
		return rdfstore.getAnnotationProperties();
	}


	public Collection<WikiResource> getDatatypePropertiesByDomain(WikiResource domain) throws DBException {
		return rdfstore.getDatatypePropertiesByDomain(domain);
	}


	public Collection<WikiResource> getTypes(WikiResource subject) throws DBException {
		return rdfstore.getTypes(subject);
	}


	public Collection<WikiResource> getDirectTypes(WikiResource subject) throws DBException {
		return rdfstore.getDirectTypes(subject);
	}


	public void storeTriple(WikiResource subject, WikiResource property, WikiNode object) {
		rdfstore.storeTriple(subject, property, object);
		notifyResourceChange(subject);
		notifyResourceChange(property);
		if(object.isResource()) {
			notifyResourceChange((WikiResource)object);
		}
	}




	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.RDFStore#getAllTypes()
	 */
	public Collection<WikiResource> getAllTypes() throws DBException {
		return rdfstore.getAllTypes();
	}




	public void removeType(WikiResource subject, WikiResource object) {
		rdfstore.removeType(subject, object);
		notifyResourceChange(subject);
		notifyResourceChange(object);
	}




	public void storeType(WikiResource subject, WikiResource object) {
		rdfstore.storeType(subject, object);
		notifyResourceChange(subject);
		notifyResourceChange(object);
	}




	public Collection<WikiResource> getApplicableObjectProperties(WikiResource subject, WikiResource object) throws DBException {
		return rdfstore.getApplicableObjectProperties(subject, object);
	}


	public void deleteImage(WikiImage img) throws DBException {
		dbstore.deleteImage(img);
	}


	public WikiImage getImageByNamespaceTitle(String ns_prefix, String title) throws DBException {
		return dbstore.getImageByNamespaceTitle(ns_prefix, title);
	}


	public void storeImage(WikiImage img) throws DBException {
		dbstore.storeArticle(img);
		dbstore.storeImage(img);
		rdfstore.storeImage(img);
		notifyAllChange();
	}


	public boolean existsImageByNamespaceTitle(String ns_prefix, String title) throws DBException {
		SoftReference<Boolean> ref = exists_image_cache.get(ns_prefix+title);
		if(ref != null && ref.get() != null) {
			return ref.get();
		} else {
			boolean result = dbstore.existsImageByNamespaceTitle(ns_prefix, title);
			exists_image_cache.put(ns_prefix+":"+title,new SoftReference<Boolean>(result));
			return result;
		}
		
	}


	public void deleteVariable(String name) throws DBException {
		dbstore.deleteVariable(name);
	}


	public String getVariable(String name) throws DBException {
		return dbstore.getVariable(name);
	}


	public void storeVariable(String name, String value) throws DBException {
		dbstore.storeVariable(name, value);
	}


	public List<Variable> listVariables() throws DBException {
		return dbstore.listVariables();
	}
	
	public void flushCaches() {
		notifyAllChange();
	}


	public Locale getLanguage(WikiResource r) throws DBException {
		return rdfstore.getLanguage(r);
	}


	public Model getModel() {
		return rdfstore.getModel();
	}


	public OntModel getOntologyModel() {
		return rdfstore.getOntologyModel();
	}


	
	
	public boolean isAnnotationProperty(WikiResource r) {
		return rdfstore.isAnnotationProperty(r);
	}


	public boolean isClass(WikiResource r) {
		return rdfstore.isClass(r);
	}


	public boolean isDatatypeProperty(WikiResource r) {
		return rdfstore.isDatatypeProperty(r);
	}


	public boolean isObjectProperty(WikiResource r) {
		return rdfstore.isObjectProperty(r);
	}


	
	
	public void addSuperClass(WikiResource subject, WikiResource superclass) throws DBException {
		rdfstore.addSuperClass(subject, superclass);
		notifyResourceChange(subject);
		notifyResourceChange(superclass);
	}

	

	public void addSuperProperty(WikiResource subject, WikiResource superclass) throws DBException {
		rdfstore.addSuperProperty(subject, superclass);
		notifyResourceChange(subject);
		notifyResourceChange(superclass);
	}


	
	
	public void removeSuperClass(WikiResource subject, WikiResource superclass) throws DBException {
		rdfstore.removeSuperClass(subject, superclass);
		notifyResourceChange(subject);
		notifyResourceChange(superclass);
	}


	public void removeSuperProperty(WikiResource subject, WikiResource superclass) throws DBException {
		rdfstore.removeSuperProperty(subject, superclass);
		notifyResourceChange(subject);
		notifyResourceChange(superclass);
	}


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.RDFStore#getSubClasses(at.srfg.ikewiki.data.WikiResource, boolean)
	 */
	public Collection<WikiResource> getSubClasses(WikiResource r, boolean direct) throws DBException {
		return rdfstore.getSubClasses(r, direct);
	}


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.RDFStore#getSuperClasses(at.srfg.ikewiki.data.WikiResource, boolean)
	 */
	public Collection<WikiResource> getSuperClasses(WikiResource r, boolean direct) throws DBException {
		return rdfstore.getSuperClasses(r, direct);
	}
	
	
	
	public Collection<WikiResource> getSubProperties(WikiResource r, boolean direct) throws DBException {
		return rdfstore.getSubProperties(r, direct);
	}


	public Collection<WikiResource> getSuperProperties(WikiResource r, boolean direct) throws DBException {
		return rdfstore.getSuperProperties(r, direct);
	}


	public Collection<WikiResource> getPropertiesByDomain(WikiResource r, boolean direct) throws DBException {
		return rdfstore.getPropertiesByDomain(r, direct);
	}


	public Collection<WikiResource> getPropertiesByRange(WikiResource r, boolean direct) throws DBException {
		return rdfstore.getPropertiesByRange(r, direct);
	}


	
	
	
	public void addDomain(WikiResource r, WikiResource dom) throws DBException {
		rdfstore.addDomain(r, dom);
		notifyResourceChange(r);
	}


	public void addRange(WikiResource r, WikiResource range) throws DBException {
		rdfstore.addRange(r, range);
		notifyResourceChange(r);
	}
	public void addInverseOf(WikiResource r, WikiResource range) throws DBException {
		rdfstore.addInverseOf(r, range);
		notifyResourceChange(r);
	}


	public void removeDomain(WikiResource r, WikiResource dom) throws DBException {
		rdfstore.removeDomain(r, dom);
		notifyResourceChange(r);
	}


	public void removeRange(WikiResource r, WikiResource range) throws DBException {
		rdfstore.removeRange(r, range);
		notifyResourceChange(r);
	}

	public void removeInverseOf(WikiResource r, WikiResource range) throws DBException {
		rdfstore.removeInverseOf(r, range);
		notifyResourceChange(r);
	}


	public Collection<WikiResource> getDomain(WikiResource r) throws DBException {
		return rdfstore.getDomain(r);
	}


	public Collection<WikiResource> getRange(WikiResource r) throws DBException {
		return rdfstore.getRange(r);
	}

	
	

	public Collection<WikiResource> getObjectTypes(WikiResource prop) throws DBException {
		return rdfstore.getObjectTypes(prop);
	}


	public Collection<WikiResource> getSubjectTypes(WikiResource prop) throws DBException {
		return rdfstore.getSubjectTypes(prop);
	}


	public Iterable<WikiResource> getInstances(WikiResource type) throws DBException {
		return rdfstore.getInstances(type);
	}


	public Document executeSparqlQuery(String queryString) throws DBException {
		return rdfstore.executeSparqlQuery(queryString);
	}


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.RDFStore#removeResource(at.srfg.ikewiki.data.WikiResource)
	 */
	public void deleteArticle(WikiResource r) {
		rdfstore.removeResource(r);
		try {
			dbstore.deleteArticle(r);
		} catch(DBException ex) {
			log.warn("problem while deleting article",ex);
		}
		notifyAllChange();
	}

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#undeleteArticle(at.srfg.ikewiki.data.WikiResource)
	 */
	public void undeleteArticle(WikiResource a) throws DBException {
		dbstore.undeleteArticle(a);
	}

	
	


	public void logMetadataAdd(Collection<Statement> stmts, User user) throws DBException {
		dbstore.logMetadataAdd(stmts, user);
	}


	public void logMetadataDelete(Collection<Statement> stmts, User user) throws DBException {
		dbstore.logMetadataDelete(stmts, user);
	}


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#flushMetadataJournal()
	 */
	public void flushMetadataJournal() throws DBException {
		dbstore.flushMetadataJournal();
	}

	

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#getUser(java.lang.String)
	 */
	public User getUser(String login) throws DBException {
		return dbstore.getUser(login);
	}


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#listUsers()
	 */
	public Collection<User> listUsers() throws DBException {
		return dbstore.listUsers();
	}


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#listUsers(at.srfg.ikewiki.data.Role)
	 */
	public Collection<User> listUsers(Role role) throws DBException {
		return dbstore.listUsers(role);
	}

	
	

	public boolean hasPermission(User user, String resource, String action) throws DBException {
		return dbstore.hasPermission(user, resource, action);
	}




	public Collection<Permission> listPermissions(Role role) throws DBException {
		return dbstore.listPermissions(role);
	}


	public Collection<Permission> listPermissions(User user) throws DBException {
		return dbstore.listPermissions(user);
	}


	
	

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#addPermission(at.srfg.ikewiki.data.Role, at.srfg.ikewiki.data.Permission)
	 */
	public void addPermission(Role role, Permission p) throws DBException {
		dbstore.addPermission(role, p);
	}


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#addPermission(at.srfg.ikewiki.data.User, at.srfg.ikewiki.data.Permission)
	 */
	public void addPermission(User user, Permission p) throws DBException {
		dbstore.addPermission(user, p);
	}



	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#removePermission(at.srfg.ikewiki.data.Role, at.srfg.ikewiki.data.Permission)
	 */
	public void removePermission(Role role, Permission p) throws DBException {
		dbstore.removePermission(role, p);
	}


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#removePermission(at.srfg.ikewiki.data.User, at.srfg.ikewiki.data.Permission)
	 */
	public void removePermission(User user, Permission p) throws DBException {
		dbstore.removePermission(user, p);
	}



	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.db.DBStore#listRecentChanges(int, int)
	 */
	public List<Article> listRecentChanges(int limit, int offset) throws DBException {
		return dbstore.listRecentChanges(limit, offset);
	}


	/**
	 * Clean up and save the model.
	 *
	 */
	public void close() {
		rdfstore.close();
		dbstore.close();
	}
	
}
