/**
 *  Copyright (C) 2006
 *  Sebastian Schaffert and Rupert Westenthaler
 *  Salzburg Research Forschungsgesellschaft m.b.H.
 *
 *  IkeWiki is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU Lesser General Public
 * 	License as published by the Free Software Foundation; either
 * 	version 2 of the License, or (at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * 	Lesser General Public License for more details.
 *
 * 	You should have received a copy of the GNU Lesser General Public
 * 	License along with this library; if not, write to the
 * 	Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * 	Boston, MA 02111-1307, USA.
 * 
 */
package at.srfg.ikewiki.installer;

import java.util.LinkedList;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import at.srfg.ikewiki.Main;
import at.srfg.ikewiki.data.Namespace;
import at.srfg.ikewiki.data.Permission;
import at.srfg.ikewiki.data.Role;
import at.srfg.ikewiki.db.DBException;

/**
 * @author sschaffe
 *
 */
public class InitData {

	private static Log log = LogFactory.getLog(InitOntologies.class);

	private Main main;
	
	private static List<Role> roles = new LinkedList<Role>();
	
	static {
		roles.add(new Role("Administrator", "role for system administrators; full access to all aspects of the system"));
		roles.add(new Role("Manager","role for managagers; create and modify all content in the system"));
		roles.add(new Role("Authenticated","default role for authenticated users; read access to all pages, write access to base namespace"));
		roles.add(new Role("Reader","default role for unauthenticated users; read access to all pages, but no write access"));		
	};
	
	private static List<Permission> permissions = new LinkedList<Permission>();
	
//	static {
//		permissions.add(new Permission("view","permission to view the object"));
//		permissions.add(new Permission("modify","permission to modify the object"));
//		permissions.add(new Permission("execute","permission to execute the object"));
//	}
	
	
	/**
	 * 
	 */
	public InitData() throws DBException {
		super();
		main = Main.getInstance();
		
		// effectively disable maintanance thread
		main.getStorageBackend().getRDFStore().getModelMaintanance().setInterval(10000);
		
		
		
	}

	public void initData(String admin_password,String permission_set) throws DBException {
		
		// add initial roles
		populateRoles();
		// add initial permissions
//		populatePermissions();
		
		// add initial user data
		populateUsers(admin_password);
		
		// add initial role->permission associations
		if(permission_set.equals("simple")) {
			populatePermissionAssociationsSimple();
		} else if(permission_set.equals("authenticated")) {
			populatePermissionAssociationsAuthenticated();
		} else if(permission_set.equals("restricted")) {
			populatePermissionAssociationsRestricted();
		}
		
	}
	
	protected void populateRoles() throws DBException {
		for(Role r : roles) {
			main.getStorageBackend().storeRole(r);
		}
		log.info("stored initial role data");
	}
	
	
//	protected void populatePermissions() throws DBException {
//		for(Permission p : permissions) {
//			main.getStorageBackend().addPermission(p);
//		}
//		log.info("stored initial permissions");
//	}
	
	
	protected void populatePermissionAssociationsSimple() throws DBException {
		String[] disallowed_actions = {  "Flush", "EditUsers", "EditRoles", "EditNamespaces" }; 
		String[] modify_actions = { "Delete", "Undelete", "Edit", "New", "Upload", "History" }; 

		
		// Administrator role
		List<Permission> admin_permissions = new LinkedList<Permission>();
		admin_permissions.add(new Permission(Permission.Type.PERMISSION,5,"%","%","administrators are given full access"));
		
		for(Permission p : admin_permissions) {
			main.getStorageBackend().addPermission(main.getStorageBackend().getRole("Administrator"),p);
		}
		
		
		// Manager role
		List<Permission> manager_permissions = new LinkedList<Permission>();
		manager_permissions.add(new Permission(Permission.Type.PERMISSION,3,"%","%","managers are given full access except for internal resources"));

		for(String action : modify_actions) {
			manager_permissions.add(new Permission(Permission.Type.RESTRICTION,4,"http://ikewiki.srfg.at/%",Namespace.NS_IKEWIKI_ACTION+action+"%","managers are given full access except for internal resources"));
		}
		for(String action : disallowed_actions) {
			manager_permissions.add(new Permission(Permission.Type.RESTRICTION,4,"%",Namespace.NS_IKEWIKI_ACTION+action+"%","managers are given full access except for internal resources"));
		}
		for(Permission p : manager_permissions) {
			main.getStorageBackend().addPermission(main.getStorageBackend().getRole("Manager"),p);
		}
		
		// Authenticated role
		List<Permission> authenticated_permissions = new LinkedList<Permission>();
		authenticated_permissions.add(new Permission(Permission.Type.PERMISSION,3,"%","%","authenticated users are given full access except for internal resources"));

		for(String action : disallowed_actions) {
			authenticated_permissions.add(new Permission(Permission.Type.RESTRICTION,4,"%",Namespace.NS_IKEWIKI_ACTION+action+"%","authenticated users are given full access except for internal resources"));
		}
		for(Permission p : authenticated_permissions) {
			main.getStorageBackend().addPermission(main.getStorageBackend().getRole("Authenticated"),p);
		}

		// Reader Role
		List<Permission> reader_permissions = new LinkedList<Permission>();
		reader_permissions.add(new Permission(Permission.Type.PERMISSION,3,"%","%","anonymous users are given full access except for internal resources"));

		for(String action : disallowed_actions) {
			reader_permissions.add(new Permission(Permission.Type.RESTRICTION,4,"%",Namespace.NS_IKEWIKI_ACTION+action+"%","anonymous users are given full access except for internal resources"));
		}
		
		for(Permission p : reader_permissions) {
			main.getStorageBackend().addPermission(main.getStorageBackend().getRole("Reader"),p);
		}

		log.info("stored initial role permissions (simple/unrestricted permission set)");
		
	}

	protected void populatePermissionAssociationsAuthenticated() throws DBException {
		String[] disallowed_actions = {  "Flush", "EditUsers", "EditRoles", "EditNamespaces" }; 
		String[] modify_actions = { "Delete", "Undelete", "Edit", "New", "Upload", "History" }; 
		String[] view_actions = { "View", "Login", "Logout", "Search", "Related", "Sparql", "Context" }; 

		
		// Administrator role
		List<Permission> admin_permissions = new LinkedList<Permission>();
		admin_permissions.add(new Permission(Permission.Type.PERMISSION,5,"%","%","administrators are given full access"));
		
		for(Permission p : admin_permissions) {
			main.getStorageBackend().addPermission(main.getStorageBackend().getRole("Administrator"),p);
		}
		
		
		// Manager role
		List<Permission> manager_permissions = new LinkedList<Permission>();
		manager_permissions.add(new Permission(Permission.Type.PERMISSION,3,"%","%","managers are given full access except for internal resources"));

		for(String action : modify_actions) {
			manager_permissions.add(new Permission(Permission.Type.RESTRICTION,4,"http://ikewiki.srfg.at/%",Namespace.NS_IKEWIKI_ACTION+action+"%","managers are given full access except for internal resources"));
		}
		for(String action : disallowed_actions) {
			manager_permissions.add(new Permission(Permission.Type.RESTRICTION,4,"%",Namespace.NS_IKEWIKI_ACTION+action+"%","managers are given full access except for internal resources"));
		}
		for(Permission p : manager_permissions) {
			main.getStorageBackend().addPermission(main.getStorageBackend().getRole("Manager"),p);
		}
		
		// Authenticated role
		List<Permission> authenticated_permissions = new LinkedList<Permission>();
		authenticated_permissions.add(new Permission(Permission.Type.PERMISSION,3,"%","%","authenticated users are given full access except for internal resources"));

		for(String action : disallowed_actions) {
			authenticated_permissions.add(new Permission(Permission.Type.RESTRICTION,4,"%",Namespace.NS_IKEWIKI_ACTION+action+"%","authenticated users are given full access except for internal resources"));
		}
		for(Permission p : authenticated_permissions) {
			main.getStorageBackend().addPermission(main.getStorageBackend().getRole("Authenticated"),p);
		}

		// Reader Role
		List<Permission> reader_permissions = new LinkedList<Permission>();
		for(String action : view_actions) {
			reader_permissions.add(new Permission(Permission.Type.PERMISSION,0,"%",Namespace.NS_IKEWIKI_ACTION+action+"%","authenticated users are denied modification of everything besides the base namespace"));
		}

		String[] read_disallowed_actions = { "Edit", "New", "Upload", "Delete", "Undelete", "Flush", "History" }; 
		for(String action : read_disallowed_actions) {
			reader_permissions.add(new Permission(Permission.Type.RESTRICTION,1,"%",Namespace.NS_IKEWIKI_ACTION+action+"%","authenticated users are denied access to system actions"));
		}
		
		for(Permission p : reader_permissions) {
			main.getStorageBackend().addPermission(main.getStorageBackend().getRole("Reader"),p);
		}

		log.info("stored initial role permissions (authenticated permission set)");
		
	}
	
	
	protected void populatePermissionAssociationsRestricted() throws DBException {
		String[] modify_actions = { "Delete", "Undelete", "Edit", "New", "Upload", "History" }; 
		String[] view_actions = { "View", "Login", "Logout", "Search", "Related", "Sparql", "Context" }; 

		
		// Administrator role
		List<Permission> admin_permissions = new LinkedList<Permission>();
		admin_permissions.add(new Permission(Permission.Type.PERMISSION,5,"%","%","administrators are given full access"));
		
		for(Permission p : admin_permissions) {
			main.getStorageBackend().addPermission(main.getStorageBackend().getRole("Administrator"),p);
		}
		
		
		// Manager role
		List<Permission> manager_permissions = new LinkedList<Permission>();
		manager_permissions.add(new Permission(Permission.Type.PERMISSION,3,"%","%","managers are given full access except for internal resources"));

		for(String action : modify_actions) {
			manager_permissions.add(new Permission(Permission.Type.RESTRICTION,4,"http://ikewiki.srfg.at/%",Namespace.NS_IKEWIKI_ACTION+action+"%","managers are given full access except for internal resources"));
		}
		for(Permission p : manager_permissions) {
			main.getStorageBackend().addPermission(main.getStorageBackend().getRole("Manager"),p);
		}
		
		// Authenticated role
		List<Permission> authenticated_permissions = new LinkedList<Permission>();
		authenticated_permissions.add(new Permission(Permission.Type.PERMISSION,3,main.getNamespaceByPrefix("").getUri()+"%","%","authenticated users are given full access to the base namespace"));
		authenticated_permissions.add(new Permission(Permission.Type.PERMISSION,1,"%","%","authenticated users are given low-priority access to all resources"));

		for(String action : modify_actions) {
			authenticated_permissions.add(new Permission(Permission.Type.RESTRICTION,2,"%",Namespace.NS_IKEWIKI_ACTION+action+"%","authenticated users are denied modification of everything besides the base namespace"));
		}
		String[] auth_disallowed_actions = {  "Flush", "EditUsers", "EditRoles", "EditNamespaces" }; 
		for(String action : auth_disallowed_actions) {
			authenticated_permissions.add(new Permission(Permission.Type.RESTRICTION,4,"%",Namespace.NS_IKEWIKI_ACTION+action+"%","authenticated users are denied access to system actions"));
		}
		
		
		for(Permission p : authenticated_permissions) {
			main.getStorageBackend().addPermission(main.getStorageBackend().getRole("Authenticated"),p);
		}

		// Reader Role
		List<Permission> reader_permissions = new LinkedList<Permission>();
		for(String action : view_actions) {
			reader_permissions.add(new Permission(Permission.Type.PERMISSION,0,"%",Namespace.NS_IKEWIKI_ACTION+action+"%","authenticated users are denied modification of everything besides the base namespace"));
		}

		String[] read_disallowed_actions = { "Edit", "New", "Upload", "Delete", "Undelete", "Flush", "History" }; 
		for(String action : read_disallowed_actions) {
			reader_permissions.add(new Permission(Permission.Type.RESTRICTION,1,"%",Namespace.NS_IKEWIKI_ACTION+action+"%","authenticated users are denied access to system actions"));
		}
		
		for(Permission p : reader_permissions) {
			main.getStorageBackend().addPermission(main.getStorageBackend().getRole("Reader"),p);
		}

		log.info("stored initial role permissions (restricted permission set)");
	}
	
	
	protected void populateUsers(String admin_password) throws DBException {
		// add initial user data
		
		main.getPermissionManager().createUser("Anonymous","Anonymous","Coward","anonymous@example.com",null,"Reader");
		main.getPermissionManager().createUser("Administrator","Joe","Admin","admin@example.com",admin_password,"Administrator");

		log.info("stored initial user data");
}
	

}
