/**
 *  Copyright (C) 2005
 *  Sebastian Schaffert and Rupert Westenthaler
 *  Salzburg Research Forschungsgesellschaft m.b.H.
 *
 *  IkeWiki is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU Lesser General Public
 * 	License as published by the Free Software Foundation; either
 * 	version 2 of the License, or (at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * 	Lesser General Public License for more details.
 *
 * 	You should have received a copy of the GNU Lesser General Public
 * 	License along with this library; if not, write to the
 * 	Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * 	Boston, MA 02111-1307, USA.
 * 
 */
package at.srfg.ikewiki.layout;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import at.srfg.ikewiki.Main;
import at.srfg.ikewiki.data.WikiLink;
import at.srfg.ikewiki.data.WikiResource;
import at.srfg.ikewiki.db.DBException;

public abstract class AbstractViewItemFactory implements ViewItemFactory {

	private static Log log = LogFactory.getLog(AbstractViewItemFactory.class);
	
	public static final String SHOW_AS_Property = "http://ikewiki.srfg.at/view#show-as";
	private WikiResource showAs = null;
	private Map<String, WikiResource> showAsMap = new HashMap<String, WikiResource>(); 
	protected boolean cacheKey2WikiResourceMapping = true;
	private Main main;
	protected Set<String> keys = null;
	protected Map<String,WikiResource> key2resource = null;
	private static Iterator<String> noIterator = new Iterator<String>() {
		public void remove() {}
		public String next() {return null;}
		public boolean hasNext() {return false;} 
	};
	
	public AbstractViewItemFactory() throws DBException {
		main = Main.getInstance();
		Collection<WikiLink> mappingList = null;
		//log.info("Load view mapping ... ");
		try {
			showAs = main.getResourceByUri(SHOW_AS_Property);
			//log.info("Show as property = "+showAs); 
		} catch (DBException e) {
			log.warn("Unable to load show-as property (view-mapping deactivated)",e);
		}
		try {
			mappingList = main.getStorageBackend().getTriplesByProperty(showAs);
			//log.info("Build Mapping Map ("+mappingList.size()+")");
			for(WikiLink l : mappingList){
				if(l.getTarget() != null &&	l.getSource() != null && l.getTarget().isResource()){
					//log.info("  > "+l.getSource().getQualifiedTitle()+" -> "+((WikiResource)l.getTarget()).getQualifiedTitle());
					showAsMap.put(l.getSource().getQualifiedTitle(), (WikiResource)l.getTarget());
				}
			//log.info("   ... "+showAsMap.size()+" mappings are present");
			}
		} catch (DBException e) {
			log.warn("Unable to load view mapping configuration (view-mapping deactivated)",e);
		}
	}
	public boolean isViewMapping(String qName){
		return showAsMap.containsKey(qName);
	}
	public WikiResource getViewMapping(String qName){
		return showAsMap.get(qName);
	}
	public List<WikiResource> getMappedProps(WikiResource object){
		List<WikiResource> mappedProps;
		try {
			mappedProps = Main.getInstance().getStorageBackend().getSubjectsByPropertyObject(showAs, object);
		} catch (DBException e) {
			log.warn("Unable to get Properties mapped to "+object.getQualifiedTitle()+" (return null -> no mappings)", e);
			mappedProps = null;
		}
		return mappedProps.isEmpty()?null:mappedProps;
	}
	
	public abstract ViewItem getViewItem(WikiResource resource);
	
	protected void setItemKeys(Set<String> keys){
		this.keys = keys;
		if(keys == null){
			key2resource = null;
		} else {
			key2resource = new HashMap<String, WikiResource>(keys.size());
		}
	}

	public final Iterator<String> getLiteralKeys() {
		if(keys == null){
			return noIterator;
		}else{
			return keys.iterator();
		}
	}

	public final boolean isLiteralKey(String key) {
		return keys == null?false:keys.contains(key);
	}

	public WikiResource getLiteralProperty(String key) {
		if(isLiteralKey(key)){
			WikiResource resource = null;
			if(cacheKey2WikiResourceMapping){
				key2resource.get(key);
			}
			if(resource == null){
				try {
					resource = Main.getInstance().getResourceByQTitle(key);
				} catch (DBException e) {
					log.warn("Unable to get WikiResource for literal key", e);
				}
				// if init was successful ... add to map
				if(resource != null && cacheKey2WikiResourceMapping){
					key2resource.put(key, resource);
				}
			}
			return resource;
		} else {
			return null;
		}
	}
	/**
	 * Getter for the state of the caching functionality for key to WikiResource
	 * Mapping.
	 * @return The state
	 */
	public boolean isCacheKey2WikiResourceMapping() {
		return cacheKey2WikiResourceMapping;
	}
	/**
	 * Setter for the state of the caching functionality for key to WikiResource
	 * Mapping.<p>
	 * Setting the state to false, clears all cached information
	 * @param cacheKey2WikiResourceMapping the new state
	 */
	public void setCacheKey2WikiResourceMapping(boolean cacheKey2WikiResourceMapping) {
		this.cacheKey2WikiResourceMapping = cacheKey2WikiResourceMapping;
		if(!cacheKey2WikiResourceMapping){
			key2resource.clear();
		}
	}
	public List<WikiResource> getMappedProps(String qName) {
		try {
			return getMappedProps(main.getResourceByQTitle(qName));
		} catch (DBException e) {
			log.warn("Unable to get WikiResource for qName"+qName+" (return null -> no mappings)", e);
			return null;
		}
	}
	public WikiResource getViewMapping(WikiResource property) {
		return getViewMapping(property.getQualifiedTitle());
	}
	public boolean isViewMapping(WikiResource property) {
		return isViewMapping(property.getQualifiedTitle());
	}

}
