/**
 *  Copyright (C) 2005
 *  Sebastian Schaffert and Rupert Westenthaler
 *  Salzburg Research Forschungsgesellschaft m.b.H.
 *
 *  IkeWiki is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU Lesser General Public
 * 	License as published by the Free Software Foundation; either
 * 	version 2 of the License, or (at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * 	Lesser General Public License for more details.
 *
 * 	You should have received a copy of the GNU Lesser General Public
 * 	License along with this library; if not, write to the
 * 	Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * 	Boston, MA 02111-1307, USA.
 * 
 */
package at.srfg.ikewiki.layout.perspectives;

import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import at.srfg.ikewiki.Main;
import at.srfg.ikewiki.actions.Action;
import at.srfg.ikewiki.data.WikiResource;
import at.srfg.ikewiki.db.DBException;

import com.hp.hpl.jena.rdf.model.NodeIterator;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.ResIterator;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;

/**
 * Produce a site map of the whole Wiki. Uses an adapted form of the Sugiyama algorithm for graph layout.
 * 
 * 
 * @author sschaffe
 *
 */
public class SiteMapPerspective extends BasePerspective {

	private static Log log = LogFactory.getLog(SiteMapPerspective.class);

	private Graphics2D myGraphics;
	
	/**
	 * @param m
	 */
	public SiteMapPerspective(Main m) {
		super(m);

		// used for calculating string sizes with font metrics
		BufferedImage img = new BufferedImage(100,100,BufferedImage.TYPE_INT_RGB);
		myGraphics = img.createGraphics();
		myGraphics.setFont(new Font("Arial",Font.PLAIN,9));
	
	}

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.layout.perspectives.BasePerspective#getTitle()
	 */
	@Override
	public String getTitle() {
		return "Site Map";
	}

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.layout.perspectives.BasePerspective#getLocalTitle()
	 */
	@Override
	protected String getLocalTitle() {
		return "SiteMap";
	}

	/**
	 * Sugiyama algorithm Step 0: find the root nodes in the pseudo-hierarchy. This method selects all resources that have no incoming edges, i.e.
	 * only occur as subjects but never as objects.
	 * 
	 * TODO: if the resulting set is empty (there are no real root nodes), we should select an arbitrary resource
	 *  
	 * @return
	 * @throws DBException
	 */
	private Collection<Resource> findRoots() throws DBException {
		HashSet<Resource> roots = new HashSet<Resource>();
		
		// add all existing subjects to the set
		for(ResIterator it = main.getStorageBackend().getModel().listSubjects(); it.hasNext(); ) {
			roots.add(it.nextResource());
		}

		log.debug("number of subjects in hierarchy: "+roots.size());
		
		// remove all subjects that are also objects from the set
		for(NodeIterator it = main.getStorageBackend().getModel().listObjects(); it.hasNext(); ) {
			RDFNode n = it.nextNode();
			if(n.isResource()) {
				roots.remove(n);
			}
		}
		
		log.debug("number of root nodes in hierarchy: "+roots.size());
		
		return roots;
	}
	
	/**
	 * Sugiyama algorithm Step 1: create a hierarchy from the graph. We use a simplified form that does not inject nodes; instead, we count only the edge that is
	 * closest to the root nodes.
	 * 
	 * @param root_nodes
	 * @return
	 * @throws DBException
	 */
	private List<List<ResourceContainer>> createHierarchy(Collection<Resource> root_nodes) throws DBException {
		LinkedList<List<ResourceContainer>> levels = new LinkedList<List<ResourceContainer>>();
		
		// the two currently considered levels
		ArrayList<ResourceContainer> prevLevel, nextLevel = new ArrayList<ResourceContainer>();
		
		for(Resource r : root_nodes) {
			nextLevel.add(new ResourceContainer(r));
		}
		
		// mark resources as seen to avoid cycles 
		HashSet<Resource> seen = new HashSet<Resource>();

		// add root nodes; this is not necessary if the root nodes are "real" roots, but if other resources point to them
		// this might be useful (future)
		seen.addAll(root_nodes);
		
		// iterate over levels and compute the successor level
		while(!nextLevel.isEmpty()) {
			levels.add(nextLevel);
			
			prevLevel = nextLevel;
			nextLevel = new ArrayList<ResourceContainer>();

			HashMap<Resource,ResourceContainer> lookup = new HashMap<Resource,ResourceContainer>(); 
			
			for(ResourceContainer s : prevLevel) {
				for(StmtIterator it = main.getStorageBackend().getModel().listStatements(s.getResource(),null,(RDFNode)null); it.hasNext(); ) {
					Statement stmt = it.nextStatement();
					RDFNode   o    = stmt.getObject();
					if(o.isResource() && ((Resource)o).getURI() != null) {						
						if(!seen.contains(o)) {
							ResourceContainer rc = new ResourceContainer((Resource)o);
							nextLevel.add(rc);
							lookup.put((Resource)o,rc);
							seen.add((Resource)o);
							
							// for computing the BaryCentre values, we store the reference to the parent
							rc.parents.add(s);
							s.children.add(rc);
						} else {
							// if the parent is in the previous level, add a link to it for computing the BaryCentre values later
							ResourceContainer rc = lookup.get((Resource)o);
							if(rc != null) {
								rc.parents.add(s);
								s.children.add(rc);
							}
						}
					}
				}
			}
			
		}
		
		return levels;
	}
	
	/**
	 * Sugiyama algorithm Step II: sort the levels according to the Barycenter value of the resources. We use a fixed maximum number of iterations
	 * in this step.
	 * 
	 * @param levels
	 */
	private void sortLevels(List<List<ResourceContainer>> levels) {
		int MAX_ITERATIONS = 5;
		boolean changed = true;
		
			
		for(int lvl = 0; lvl<levels.size()-1; lvl++) {
			List<ResourceContainer> prevLevel = levels.get(lvl), nextLevel = levels.get(lvl+1);

			for(int i=0; i<MAX_ITERATIONS && changed; i++) {
				changed = false;
				
				// phase I: reorder nodes with different baryCenters
				
				// step 1: compute downBaryCenters and sort by them
				
				// compute BaryCentres of next level wrt previous level
				for(ResourceContainer r : nextLevel) {
					r.downBaryCenter = computeDownBaryCenter(r,prevLevel);
				}
				
				// sort nextLevel according to downBaryCenter values
				ArrayList<ResourceContainer> nextLevelS = new ArrayList<ResourceContainer>(nextLevel);
				Collections.sort(nextLevelS,new DownBaryCenterComparator());
				
				// if new sorting has fewer edge crossings, replace nextLevel by nextLevelS
				if(computeCrossings(prevLevel,nextLevelS) < computeCrossings(prevLevel,nextLevel)) {
					levels.set(lvl+1,nextLevelS);
					nextLevel = nextLevelS;
					changed = true;
				}
												

				// step 2: compute upBaryCenters and sort by them				
			
				// compute BaryCentres of previous level wrt next level
				for(ResourceContainer r : prevLevel) {
					r.upBaryCenter = computeUpBaryCenter(r,nextLevel);
				}
						
				// sort prevLevel according to upBaryCenter values
				ArrayList<ResourceContainer> prevLevelS = new ArrayList<ResourceContainer>(prevLevel);
				Collections.sort(prevLevelS,new UpBaryCenterComparator());
				
				// if new sorting has fewer edge crossings, replace nextLevel by nextLevelS
				if(computeCrossings(prevLevel,nextLevelS) < computeCrossings(prevLevel,nextLevel)) {
					levels.set(lvl,prevLevelS);
					prevLevel = prevLevelS;
					changed = true;
				}
				
				
				// phase II: reverse nodes with equal up/down-baryCenter
			}
		}
	}
	
	
	private int computeCrossings(List<ResourceContainer> prevLevel, List<ResourceContainer> nextLevel) {
		int crossings = 0;
		
		// first, we create an ajacency matrix for the bipartit graph
		int[][] ajm = new int[prevLevel.size()][nextLevel.size()];
		
		for(int i = 0; i<prevLevel.size(); i++) {
			for(int j = 0; j<nextLevel.size(); j++) {
				if(prevLevel.get(i).children.contains(nextLevel.get(j))) {
					ajm[i][j] = 1;
				} else {
					ajm[i][j] = 0;
				}
			}
		}
		
		// we now iterate over each row of the matrix; for each cell that is "true", we count the number of cells
		// in previous rows that are in a higher column and add them to the number of crossings
		for(int i = 0; i<ajm.length; i++) {
			for(int j=0; j<ajm[i].length; j++) {
				if(ajm[i][j] > 0) {
					// count all 1's in previous rows
					for(int k=0; k<i; k++) {
						for(int l=j+1; l<ajm[k].length; l++) {
							crossings += ajm[k][l];
						}
					}
				}
			}
		}
		
		
		return crossings;
	}
	
	private double computeDownBaryCenter(ResourceContainer x, List<ResourceContainer> level) {
		int sum = 0, count = 0;
		
		for(int i=0; i<level.size(); i++) {
			if(level.get(i).parents.contains(x)) {
				sum += i;
				count++;
			}
		}
		
		return count > 0 ? sum/count : 0;
	}
	
	private double computeUpBaryCenter(ResourceContainer x, List<ResourceContainer> level) {
		int sum = 0, count = 0;
		
		for(int i=0; i<level.size(); i++) {
			if(level.get(i).children.contains(x)) {
				sum += i;
				count++;
			}
		}
		
		return count > 0 ? sum/count : 0;
	}
	
	
	
	@Override
	public String getContent() {
		return paintSiteMap();
	}

	private String paintSiteMap() {
		StringBuilder result = new StringBuilder();
		
		
		int rectHeight = 20, rectWidth = 80, minSpacing = 20, width = 800;
		
		String fontFamily = "Arial";
		int fontSize = 9;
		
		try {
			List<List<ResourceContainer>> levels = createHierarchy(findRoots());
		
			sortLevels(levels);
			
			int height = maxSize(levels) * (minSpacing + rectHeight);
			
			
			result.append("<svg xmlns:xlink=\"http://www.w3.org/1999/xlink\" width=\""+width+"\" height=\""+height+"\" xmlns=\"http://www.w3.org/2000/svg\" font-family=\""+fontFamily+"\" font-size=\""+fontSize+"\">");
			result.append("<style type=\"text/css\">rect:hover { fill-opacity:0.9;} </style>\n");
			
			int offsetX = 0;
			
			for(int i=0; i<levels.size(); i++) {
				paintColumn(result,levels.get(i),offsetX, width, rectWidth, rectHeight, minSpacing);
				
				offsetX += rectWidth + minSpacing;
			}
			
			result.append("</svg>\n");
			
		} catch(DBException ex) {
			ex.printStackTrace();
		}
		
		return result.toString();
	}
	
	
	private int maxSize(List<List<ResourceContainer>> lists) {
		int maxsize = 0;
		for(List l : lists) {
			if(l.size() > maxsize)
				maxsize = l.size();
		}
		return maxsize;
	}
	
	
	private void paintColumn(StringBuilder result, List<ResourceContainer> row, int offsetX, int width, int rectWidth, int rectHeight, int minSpacing) throws DBException {
		// compute the actual spacing
		int spacing = minSpacing;
		
		// create a graphics element for the row
		result.append("<g>\n");
		
		int offsetY = spacing;
		
		// append a rectangle for each resource
		for(ResourceContainer r : row) {
			
			WikiResource wr = main.getResourceByUri(r.resource.getURI());
			
			paintRect(result,offsetX,offsetY,rectWidth,rectHeight,wr.getTitle(),"index.jsp?action=view&amp;title="+Main.encodeURL(wr.getQualifiedTitle()));
			
			offsetY += rectHeight + spacing;
		}
		
		
		result.append("</g>\n");
	}
	
	
	private void paintRect(StringBuilder result, int curX, int curY, int rectWidth, int rectHeight, String label, String uri) {
		result.append("<a xlink:href=\""+uri+"\"><g>\n");
		// draw rectangle						
		paintRect(result,curX,curY,rectWidth,rectHeight,"stroke=\"black\" fill=\"green\" opacity=\"0.5\"");
		
		String mylabel = label;

		
		int strWidth = myGraphics.getFontMetrics().stringWidth(label);
		int offset = (rectWidth - strWidth) / 2; // center text;		
		
		while(offset < 2 && mylabel.length() > 6) {
			mylabel = mylabel.substring(0,mylabel.length()-4) + "...";
			strWidth = myGraphics.getFontMetrics().stringWidth(mylabel);
			offset = (rectWidth - strWidth) / 2;
		}
		
		
		//g2d.drawString(mylabel,curX+offset,curY+rectHeight-3);

		
		result.append("<text font-stretch=\"condensed\" fill-opacity=\"1\" fill=\"black\" xml:space=\"preserve\" stroke=\"none\" text-anchor=\"middle\" x=\"" + (curX+rectWidth/2) + "\" y=\""+(curY+rectHeight-3)+"\">");
		result.append(mylabel);
		result.append("</text>\n");
		
		
		result.append("</g></a>");
	}
	
	private void paintRect(StringBuilder result, int x, int y, int width, int height, String attributes) {
		result.append("<rect x=\""+x+"\" y=\""+y+"\" width=\""+width+"\" height=\""+height+"\" "+attributes+" />");		
		
	}
	
	
	private class DownBaryCenterComparator implements Comparator<ResourceContainer> {

		public int compare(ResourceContainer o1, ResourceContainer o2) {
			if(o1.downBaryCenter < o2.downBaryCenter) {
				return -1;
			} else if(o1.downBaryCenter > o2.downBaryCenter) {
				return 1;
			} else {
				return 0;
			}
		}
		
	}
	
	private class UpBaryCenterComparator implements Comparator<ResourceContainer> {

		public int compare(ResourceContainer o1, ResourceContainer o2) {
			if(o1.upBaryCenter < o2.upBaryCenter) {
				return -1;
			} else if(o1.upBaryCenter > o2.upBaryCenter) {
				return 1;
			} else {
				return 0;
			}
		}
		
	}
	
	/**
	 * A helper data structure serving as a container for resources.
	 * @author sschaffe
	 *
	 */
	private class ResourceContainer {
		double downBaryCenter, upBaryCenter;
		
		HashSet<ResourceContainer> parents,children;
		
		Resource resource;
		
		ResourceContainer(Resource r) {
			this.resource = r;
			this.parents = new HashSet<ResourceContainer>();
			this.children = new HashSet<ResourceContainer>();
		}

		public double getDownBaryCenter() {
			return downBaryCenter;
		}

		public void setDownBaryCenter(double downBaryCenter) {
			this.downBaryCenter = downBaryCenter;
		}


		public Resource getResource() {
			return resource;
		}

		public void setResource(Resource resource) {
			this.resource = resource;
		}

		public double getUpBaryCenter() {
			return upBaryCenter;
		}

		public void setUpBaryCenter(double upBaryCenter) {
			this.upBaryCenter = upBaryCenter;
		}

		@Override
		public boolean equals(Object obj) {
			if(obj instanceof ResourceContainer) {
				return resource.equals(((ResourceContainer)obj).resource);
			} else {
				return false;
			}
		}

		@Override
		public int hashCode() {
			return resource.hashCode();
		}
		
		
		
	}
	
	public Iterable<Action> getActions() {
		return new Iterable<Action> () {

			public Iterator<Action> iterator() {
				return new Iterator<Action>() {
					
					boolean retrieved = false;
					
					public boolean hasNext() {
						return !retrieved;
					}

					public Action next() {
						retrieved = true;
						return main.getAction("sitemap");
					}

					public void remove() {
						throw new UnsupportedOperationException();						
					}
					
				};
			}
		};
	}

}
