/**
 *  Copyright (C) 2005
 *  Sebastian Schaffert and Rupert Westenthaler
 *  Salzburg Research Forschungsgesellschaft m.b.H.
 *
 *  IkeWiki is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU Lesser General Public
 * 	License as published by the Free Software Foundation; either
 * 	version 2 of the License, or (at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * 	Lesser General Public License for more details.
 *
 * 	You should have received a copy of the GNU Lesser General Public
 * 	License along with this library; if not, write to the
 * 	Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * 	Boston, MA 02111-1307, USA.
 * 
 */
package at.srfg.ikewiki.layout.perspectives;

import java.net.URI;
import java.net.URISyntaxException;
import java.text.Collator;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.Date;
import java.util.Hashtable;
import java.util.List;
import java.util.Locale;
import java.util.TreeSet;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import at.srfg.ikewiki.Main;
import at.srfg.ikewiki.actions.Action;
import at.srfg.ikewiki.data.Article;
import at.srfg.ikewiki.data.Namespace;
import at.srfg.ikewiki.data.WikiLink;
import at.srfg.ikewiki.data.WikiLiteral;
import at.srfg.ikewiki.data.WikiNode;
import at.srfg.ikewiki.data.WikiResource;
import at.srfg.ikewiki.db.DBException;
import at.srfg.ikewiki.render.RenderingException;
import at.srfg.ikewiki.render.RenderingPipeline;

/**
 * The class ViewArticlePerspective performs the rendering for "normal" display of the article.
 * 
 * It makes use of the radeox wiki rendering engine to format the structured text as
 * HTML. Currently, the default rendering of Radeox is used, but future implementations
 * might refine this.
 * 
 * @author sschaffe
 *
 */
public class ViewArticlePerspective implements Perspective {

	protected Article article;
	protected Main main;
	protected static RenderingPipeline engine;
	//private RenderContext context;
	//private String renderedContent;
	//private List<WikiLink> internal_references;
	private List<WikiLink> references;
	private Collection<WikiLink> incoming_references;
	private Collection<WikiLink> outgoing_references;
	
	private String title_postfix = "";
	
	private static Log log = LogFactory.getLog(ViewArticlePerspective.class);
	
	/**
	 * Construct an article view for the given article.
	 * 
	 * The constructor immediately renders the structured text into HTML, collects all links and
	 * stores them in references.
	 * 
	 * @param a
	 */	
	public ViewArticlePerspective(Main main,Article a) {
		article = a;			
		this.main = main;
		
		// load rendering engine that belongs to this article
		engine = main.getRenderingPipeline();
		
		references = new ArrayList<WikiLink>();
				
		
		try {
			incoming_references = main.getStorageBackend().getTriplesByObject(this.article);
			outgoing_references = main.getStorageBackend().getTriplesBySubject(this.article);
		
			for(WikiLink l : incoming_references) {
				references.add(l);
			}
			
			for(WikiLink l : outgoing_references) {
				references.add(l);
			}			
            
		} catch(DBException ex) {
			ex.printStackTrace();
		}

	}

	
	public ViewArticlePerspective(Main main, Namespace namespace, String title) throws URISyntaxException {
		this(main,new Article(namespace,title,"","Sample Text","en","http://ikewiki.srfg.at/Special:"+title,new Date())); 
	}
	
	public ViewArticlePerspective(Main main, Namespace namespace, String title, String text) throws URISyntaxException {
		this(main,new Article(namespace,title,"",text,"en","http://ikewiki.srfg.at/Special:"+title,new Date())); 		
	}

	
	
	public String getTitlePrefix() {
		return "";
	}

	
	public String getTitlePostfix() {
		return title_postfix;
	}
	
	public void setTitlePostfix(String s) {
		title_postfix = s;
	}

	public String getContentType() {
		return "application/xhtml+xml";
	}


	/**
	 * Return the context article of this ViewArticlePerspective.
	 * @return
	 */
	public Article getArticle() {
		return article;
	}
	
	
	/**
	 * Return the namespace of the article
	 */
	public Namespace getNamespace() {
		return article.getNamespace();
	}

	
	public List<WikiResource> getOutgoingReferenceTypes() throws DBException {
		return main.getStorageBackend().getPropertiesBySubject(article);
	}
	
	public List<WikiLink> getOutgoingReferences(WikiResource property) throws DBException {
		return new ArrayList<WikiLink>(main.getStorageBackend().getTriplesBySubjectProperty(article,property));
	}

	public List<WikiResource> getIncomingReferenceTypes() throws DBException {
		return main.getStorageBackend().getPropertiesByObject(article);
	}
	
	public List<WikiLink> getIncomingReferences(WikiResource property) throws DBException {
		return new ArrayList<WikiLink>(main.getStorageBackend().getTriplesByPropertyObject(property,article));
	}

	

	public Collection<WikiLink> getIncomingReferences() {
		return incoming_references;
	}


	public Collection<WikiLink> getOutgoingReferences() {
		return outgoing_references;
	}

	
	public Collection<WikiNode> getOutgoingReferenceTargets() {
		TreeSet<WikiNode> result = new TreeSet<WikiNode>(new Comparator<WikiNode>() {
			Collator myCollator = Collator.getInstance();
			public int compare(WikiNode o1, WikiNode o2) {
				
				return myCollator.compare(o1.toString(),o2.toString());
			}
			
		});
		
		for(WikiLink l : outgoing_references) {
			result.add(l.getTarget());
		}
		
		return result;
	}
	
	public Collection<WikiLink> getReferenceTypes(WikiResource source, WikiNode target) {
		TreeSet<WikiLink> result = new TreeSet<WikiLink>(new Comparator<WikiLink>() {
			Collator myCollator = Collator.getInstance();
			public int compare(WikiLink o1, WikiLink o2) {
				
				return myCollator.compare(o1.getProperty().toString(),o2.getProperty().toString());
			}
			
		});
		
		for(WikiLink l : references) {
			if(l.getTarget() == target && l.getSource() == source) {
				result.add(l);
			}
		}
		
		return result;		
	}
	
	
	public Collection<WikiResource> getIncomingReferenceSources() {
		TreeSet<WikiResource> result = new TreeSet<WikiResource>(new Comparator<WikiResource>() {
			Collator myCollator = Collator.getInstance();
			public int compare(WikiResource o1, WikiResource o2) {
				
				return myCollator.compare(o1.toString(),o2.toString());
			}
			
		});
		
		for(WikiLink l : incoming_references) {
			result.add(l.getSource());
		}
		
		return result;		
	}
	

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.render.Perspective#getTitle()
	 */
	public String getTitle() {
		// TODO Auto-generated method stub
		return article.getTitle();
	}

	public String getQualifiedTitle() {
		return article.getQualifiedTitle();
	}
	
	public Locale getLanguage() {
		return article.getLanguage();
	}
	
	
	
	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.render.Perspective#toHTML()
	 */
	public String getContent() {
		try {
			return engine.transform(article);
		} catch(RenderingException ex) {
			log.error("rendering error",ex);
			return "<div id=\"content\"><h1 class=\"firstHeading\">"+article.getTitle()+" (rendering error)</h1><div id=\"bodyContent\"><div class=\"error\">An error occurred while rendering this page. Please change the content <a href=\"index.jsp?action=edit_content&amp;title="+Main.encodeURL(article.getQualifiedTitle())+"\">here</a>.<br /><br /><strong>Error Message:</strong> "+Main.encodeXML(ex.getMessage())+"</div></div></div>";
		}
	}

	
	public Date getLastModified() {
		return article.getTimestamp();
	}
	
	
	public int getViewCount() {
		// TODO: dummy
		return 0;
	}


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.render.Perspective#getURI()
	 */
	public URI getURI() {
		try {
			return new URI(article.getUri());
		} catch(URISyntaxException ex) {
			ex.printStackTrace();
			return null;
		}
	}


	public List<WikiLink> getExternalReferences() {
		return this.references;
	}
	
	
	public Hashtable<String,String> getDublinCore() throws DBException {
		return main.getStorageBackend().getRDFStore().getDatatypePropertiesByNamespace(this.article,Namespace.NS_DC_CORE);
	}
	
	/*
	public Hashtable<String,String> getDublinCoreTerms() throws DBException {
		return main.getStorageBackend().getRDFStore().getDatatypePropertiesByNamespace(this.article,"http://purl.org/dc/terms/");
	}
	*/

	/**
	 * Get a list of all relevant metadata fields for the "edit metadata" tab.
	 * 
	 * Currently, this method only returns dublin core properties derieved from the
	 * ontology model. Future implementations might respect domains/ranges of properties
	 * and consider user-configurable namespaces.
	 * 
	 * @return
	 * @throws DBException
	 */
	public Collection<WikiResource> getMetadataFieldProperties() throws DBException {
		// add more if appropriate
		// TODO: should depend on configuration/loaded ontologies and on type
		
		// strategy: 
		// 1. if OWL is loaded, take all properties that are of type owl:AnnotationProperty
		//    otherwise, take all properties from DublinCore
		// 2. if there are datatype properties where rdfs:domain matches with the type of the 
		//    current resource, we add them as well
		
		
		/*
		String[] property_namespaces = { RDFStore.NS_DC_CORE };		
		
		ArrayList<WikiResource> result = new ArrayList<WikiResource>();
		
		for(String ns : property_namespaces) {
			result.addAll(main.getStorageBackend().getAllPropertiesByNamespace(ns));
		}*/
		
		TreeSet<WikiResource> results = new TreeSet<WikiResource>();
		
		//WikiResource ann_prop = main.getResourceByUri(RDFStore.NS_OWL+"AnnotationProperty");
		
		// retrieve all annotation properties		
		Collection<WikiResource> annotations = main.getStorageBackend().getAnnotationProperties();
		results.addAll(annotations);
		
		// retrieve all relevant datatype properties
		for(WikiResource type : main.getStorageBackend().getTypes(this.article)) {
			Collection<WikiResource> dataprop = main.getStorageBackend().getDatatypePropertiesByDomain(type); 
			results.addAll(dataprop);
			
			log.debug("added "+dataprop.size()+" datatype properties for type "+type.getUri());
		}
		return results;
	}


	/**
	 * Retrieve the metadata field value for the current article. Returns null if no value is found, 
	 * otherwise the first relevant literal node.
	 * @param property
	 * @return
	 * @throws DBException
	 */
	public WikiLiteral getMetadataFieldValue(WikiResource property) throws DBException {
		List<WikiNode> results = main.getStorageBackend().getObjectsBySubjectProperty(this.article,property);
	
		// TODO: maybe merge multiple literal nodes?
		for(WikiNode n : results) {
			if(n.isLiteral()) {
				return (WikiLiteral)n;
			}
		}
		
		return null;
	}


	public ViewArticlePerspective getMetadataFieldDescription(WikiResource property)  {
		try {
			return main.getViewArticlePerspective(property.getNamespace().getPrefix(),property.getTitle());
		} catch(DBException ex) {
			try {
				return new ViewArticlePerspective(main,main.getNamespaceSpecial(),"Warning","no description available");
			} catch(Exception ex2) {
				log.error(ex2);
				return null;
			}
		}
	}


	public void setMetadataFieldValue(String property, String value) throws DBException {
		WikiResource p_resource = main.getResourceByQTitle(property);
		WikiLiteral  v_resource = new WikiLiteral(value,article.getLanguage().getLanguage());
		
		main.getStorageBackend().storeTriple(this.article,p_resource,v_resource);
	}

	
	public void addType(String typename) throws DBException {
		WikiResource o_resource = main.getResourceByQTitle(typename);
		main.getStorageBackend().storeType(this.article,o_resource);
	}
	
	public void removeType(String typename) throws DBException {
		WikiResource o_resource = main.getResourceByQTitle(typename);
		main.getStorageBackend().removeType(this.article,o_resource);		
	}
	
	
	public Collection<WikiResource> getAvailableTypes() throws DBException {
		// currently, just return all types from the RDF store
		return main.getStorageBackend().getAllTypes();
	}
	
	public List<WikiResource> getAssociatedTypes() throws DBException {
		// from the RDF store, list all objects o such that there
		// exists a triple (article,rdf:type,o)
		
		/*
		WikiResource subject = this.article;
		WikiResource property = main.getResourceByUri(RDFStore.NS_RDF+"type");
		
		ArrayList<WikiResource> results = new ArrayList<WikiResource>();
		
		for(WikiNode n : main.getStorageBackend().getObjectsBySubjectProperty(subject,property)) {
			if(n.isResource()) {
				results.add((WikiResource)n);
			} else {
				log.warn("literal as type: "+n.toString());
			}
		}
		return results;
		*/
		return new ArrayList<WikiResource>(main.getStorageBackend().getTypes(this.article));
	}


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.layout.perspectives.Perspective#getActions()
	 */
	public Iterable<Action> getActions() {
		return main.getAllActions();
	}
	
	String error_msg;
	
	public String getErrorMessage() {
		return error_msg;
	}

	public void setErrorMessage(String message) {
		error_msg = message;
	}


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.layout.perspectives.Perspective#hasContext()
	 */
	public boolean hasContext() {
		return true;
	}


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.layout.perspectives.Perspective#getContext()
	 */
	public WikiResource getContext() {
		return getArticle();
	}
	
	
	
}
