/**
 *  Copyright (C) 2005
 *  Sebastian Schaffert and Rupert Westenthaler
 *  Salzburg Research Forschungsgesellschaft m.b.H.
 *
 *  IkeWiki is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU Lesser General Public
 * 	License as published by the Free Software Foundation; either
 * 	version 2 of the License, or (at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * 	Lesser General Public License for more details.
 *
 * 	You should have received a copy of the GNU Lesser General Public
 * 	License along with this library; if not, write to the
 * 	Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * 	Boston, MA 02111-1307, USA.
 * 
 */
package at.srfg.ikewiki.layout.perspectives;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.Paint;
import java.awt.RenderingHints;
import java.awt.image.BufferedImage;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.Collection;
import java.util.Hashtable;

import javax.imageio.ImageIO;

import org.apache.commons.collections.functors.NotPredicate;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import at.srfg.ikewiki.Main;
import at.srfg.ikewiki.data.Article;
import at.srfg.ikewiki.data.WikiLink;
import at.srfg.ikewiki.data.WikiResource;
import at.srfg.ikewiki.db.DBException;
import edu.uci.ics.jung.exceptions.ConstraintViolationException;
import edu.uci.ics.jung.graph.ArchetypeEdge;
import edu.uci.ics.jung.graph.ArchetypeVertex;
import edu.uci.ics.jung.graph.Edge;
import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.graph.Vertex;
import edu.uci.ics.jung.graph.decorators.EdgeStringer;
import edu.uci.ics.jung.graph.decorators.PickableEdgePaintFunction;
import edu.uci.ics.jung.graph.decorators.VertexPaintFunction;
import edu.uci.ics.jung.graph.decorators.VertexStringer;
import edu.uci.ics.jung.graph.impl.DirectedSparseEdge;
import edu.uci.ics.jung.graph.impl.SparseGraph;
import edu.uci.ics.jung.graph.impl.SparseVertex;
import edu.uci.ics.jung.visualization.FRLayout;
import edu.uci.ics.jung.visualization.LayoutMutable;
import edu.uci.ics.jung.visualization.PluggableRenderer;
import edu.uci.ics.jung.visualization.ShapePickSupport;
import edu.uci.ics.jung.visualization.VisualizationViewer;
import edu.uci.ics.jung.visualization.control.DefaultModalGraphMouse;

/**
 * The ViewContextPerspective Perspective creates a graph image for displaying the
 * "context" of a resource, i.e. its incoming and outgoing connections to other
 * resources. At the moment, only RDF links are visualised, WikiLinks not.
 * <p>
 * The class ViewContextPerspective is derived from ViewArticlePerspective, which provides the "basic 
 * functionality".
 * <p>
 * ViewContextPerspective currently uses the JUNG library to generate a graph view. Future 
 * implementations could be based on a nicer layout.
 * 
 * @author sschaffe
 *
 */
public class ViewContextPerspective extends ViewArticlePerspective implements Perspective {

	private static Log log = LogFactory.getLog(ViewContextPerspective.class);

	private static final Dimension DEFAULT_SIZE = new Dimension( 600, 600 );

    private Graph graph;
    private VisualizationViewer vv;
    private LayoutMutable layout;
    private PluggableRenderer pr;
    
    private BufferedImage image;
    
    // for labelling vertices and edges
    private Hashtable<Vertex,WikiResource> vertex_resource;
    private Hashtable<Edge,WikiResource> edge_resource;
    
    // for looking up vertices by URI that have already been produced
    private Hashtable<String,Vertex> uri_vertex;
    
	public ViewContextPerspective(Main main, Article resource) {
		super(main,resource);
		
		pr = new PluggableRenderer();
		
		this.graph  = new SparseGraph();
		this.layout = new FRLayout(this.graph);
		
		this.vertex_resource = new Hashtable<Vertex,WikiResource>();
		this.edge_resource   = new Hashtable<Edge,WikiResource>();
		
		this.uri_vertex = new Hashtable<String,Vertex>();
		
		
		this.vv = new VisualizationViewer(this.layout,pr);
        //this.vv.getModel().setRelaxerThreadSleepTime(500);
        vv.setPickSupport(new ShapePickSupport());
        vv.setGraphMouse(new DefaultModalGraphMouse());
        vv.setSize(DEFAULT_SIZE);
        
        EdgeStringer estringer = new EdgeStringer(){
            public String getLabel(ArchetypeEdge e) {
            	WikiResource r = edge_resource.get(e);
            	if(r != null)
            		return r.getQualifiedTitle();
            	else
            		return "";
            }
        };
        pr.setEdgeStringer(estringer);
        pr.setEdgePaintFunction(new PickableEdgePaintFunction(pr, Color.black, Color.cyan));

        
        VertexStringer vstringer = new VertexStringer() {
            public String getLabel(ArchetypeVertex v) {
            	WikiResource r = vertex_resource.get(v);
            	if(r != null)
            		return r.getQualifiedTitle();
            	else
            		return "";
            }        	
        };
        pr.setVertexStringer(vstringer);

        
        VertexPaintFunction vpf = new VertexPaintFunction() {
        	public Paint getDrawPaint(Vertex v) {
        		return Color.BLACK;
        	}
        	
        	public Paint getFillPaint(Vertex v) {
        		if(vertex_resource.get(v) == article) {
        			return Color.RED;
        		} else {
        			return Color.GREEN;
        		}
        	}
        };
        pr.setVertexPaintFunction(vpf);

        vv.setBackground(Color.WHITE);
        
		buildContext();
		
		
	}
	
	private void buildContext() {
		try {
			Collection<WikiLink> incoming = main.getStorageBackend().getTriplesByObject(this.article);
			Collection<WikiLink> outgoing = main.getStorageBackend().getTriplesBySubject(this.article);
		
			Vertex ctx = graph.addVertex(new SparseVertex());
			vertex_resource.put(ctx,article);
			uri_vertex.put(article.getUri(),ctx);
			
			for(WikiLink l : incoming) {
				Vertex v;
				if(uri_vertex.get(l.getSource().getUri()) == null) {
					v = graph.addVertex(new SparseVertex());
					vertex_resource.put(v,l.getSource());
					uri_vertex.put(l.getSource().getUri(),v);
				} else {
					v = uri_vertex.get(l.getSource().getUri()); 
				}
				Edge   e = graph.addEdge(new DirectedSparseEdge(v,ctx));
				edge_resource.put(e,l.getProperty());
			}
			
			for(WikiLink l : outgoing) {
				if(l.getTarget().isResource()) {
					Vertex v;
					if(uri_vertex.get(((WikiResource)l.getTarget()).getUri()) == null) {
						v = graph.addVertex(new SparseVertex());
						vertex_resource.put(v,(WikiResource)l.getTarget());
						uri_vertex.put(((WikiResource)l.getTarget()).getUri(),v);
					} else {
						v = uri_vertex.get(((WikiResource)l.getTarget()).getUri()); 
					}
					
					Edge   e = graph.addEdge(new DirectedSparseEdge(ctx,v));
					edge_resource.put(e,l.getProperty());
				}
			}
			
			
			
            if (!vv.isVisRunnerRunning())
                vv.init();
            //vv.repaint();

            vv.setDoubleBuffered( false );

            synchronized (vv) {
        		layout.update();
            }
            
            // update layout, at most 100 times
            int count = 0;
        	while(!layout.incrementsAreDone() && count < 100) {
        		layout.advancePositions();
        		count++;
        	}
            

            this.image = new BufferedImage(vv.getWidth(),vv.getHeight(),BufferedImage.TYPE_INT_RGB);
            
            Graphics2D g2 = image.createGraphics();
            g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                    			RenderingHints.VALUE_ANTIALIAS_ON);
            vv.setVisible(true);
            vv.validate();
            vv.paint( g2 );
            g2.dispose();
            
		} catch(ConstraintViolationException ex) {
			if(ex.getViolatedConstraint() instanceof NotPredicate) {
				NotPredicate p = (NotPredicate)ex.getViolatedConstraint();			
				log.error("error while evaluating predicate: "+p.getPredicates()[0],ex);
			} else {
				throw ex;
			}
		} catch(DBException ex) {
			// dump stack for debugging, leave graph empty
			ex.printStackTrace();
		}
	}
	
 	
	public byte[] getImage() {
		try {
			ByteArrayOutputStream os = new ByteArrayOutputStream();
		
			ImageIO.write(this.image,"png",os);
			
			return os.toByteArray();
			
		} catch(IOException ex) {
			ex.printStackTrace();
			return new byte[1];
		}
	}
}
