/**
 *  Copyright (C) 2005
 *  Sebastian Schaffert and Rupert Westenthaler
 *  Salzburg Research Forschungsgesellschaft m.b.H.
 *
 *  IkeWiki is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU Lesser General Public
 * 	License as published by the Free Software Foundation; either
 * 	version 2 of the License, or (at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * 	Lesser General Public License for more details.
 *
 * 	You should have received a copy of the GNU Lesser General Public
 * 	License along with this library; if not, write to the
 * 	Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * 	Boston, MA 02111-1307, USA.
 * 
 */
package at.srfg.ikewiki.layout.perspectives;

import java.util.List;

import org.apache.commons.jrcs.diff.AddDelta;
import org.apache.commons.jrcs.diff.ChangeDelta;
import org.apache.commons.jrcs.diff.Chunk;
import org.apache.commons.jrcs.diff.DeleteDelta;
import org.apache.commons.jrcs.diff.Delta;
import org.apache.commons.jrcs.diff.DifferentiationFailedException;
import org.apache.commons.jrcs.diff.Revision;

import at.srfg.ikewiki.Main;
import at.srfg.ikewiki.data.Article;
import at.srfg.ikewiki.db.DBException;
import at.srfg.ikewiki.diff.DiffEngine;

/**
 * ViewHistoryPerspective provides methods for showing the revisions of a specific article and differences between 
 * them. It extends ViewArticlePerspective.
 * 
 * @author wastl
 */
public class ViewHistoryPerspective extends ViewArticlePerspective implements Perspective {


	public ViewHistoryPerspective (Main m, Article a) {
		super(m,a);
	}
	
	
	public List<Article> getRevisions(int limit, int offset) throws DBException {		
		return main.getStorageBackend().getRevisionsByNamespaceTitle(article.getNamespace().getPrefix(),article.getTitle(),limit,offset);
	}


	/**
	 * From an array of Strings, select a subsequence (identified by start and length) and join them
	 * to a single String. Helper function for getContents()
	 * 
	 * @param wordlist the array containing all strings
	 * @param start the index of the first substring to use
	 * @param len the number of substrings to use
	 * @return the joined subsequence beginning with first
	 */
	public static String joinWords(String[] wordlist, int start, int len) {
		StringBuffer result = new StringBuffer();
		for(int i=start; i<start+len; i++) {
			result.append(wordlist[i]);
		}
		return result.toString();
	}
	
	/**
	 * Create a comparison view between two revisions of the article this ViewHistoryPerspective refers to and highlight
	 * differences by specific markup.
	 * 
	 * @param old_revno number of the old revision
	 * @param new_revno number of the new revision
	 * @return a HTML-formatted string showing the changes between the revisions
	 * @throws DBException in case one of the revisions cannot be retrieved
	 * @throws DifferentiationFailedException in case computation of the difference fails
	 */
	public String compareRevisions(long old_revno, long new_revno) throws DBException, DifferentiationFailedException {
		
		// retrieve articles from database
		Article old_art = main.getStorageBackend().getArticleByRevisionId(this.getNamespace().getPrefix(),this.getTitle(),old_revno);
		Article new_art = main.getStorageBackend().getArticleByRevisionId(this.getNamespace().getPrefix(),this.getTitle(),new_revno);
		
		// evaluate diff algorithm on content
		DiffEngine diff = new DiffEngine();
		Revision rev = diff.diff(old_art.getText(),new_art.getText());
		
		// create output
		String old_rev = "<div class=\"delta-old-revision\">";
		String new_rev = "<div class=\"delta-new-revision\">";;
		
		// two character arrays that we will use to copy from
		String[] old_content = DiffEngine.tokenize(old_art.getText());
		String[] new_content = DiffEngine.tokenize(new_art.getText());
		
		// pointers in the content arrays, used for determining where we are
		int old_ptr = 0, new_ptr = 0;
		
		for(int i=0; i<rev.size(); i++) {
			Delta d = rev.getDelta(i);
			
			Chunk c_from = d.getOriginal();
			Chunk c_to   = d.getRevised();
			
			// copy everything from the old position up to now to the result strings, it is skipped
			old_rev += joinWords(old_content,old_ptr,c_from.first()-old_ptr);
			new_rev += joinWords(new_content,new_ptr,c_to.first()-new_ptr);
			
			
			// depending on the kind of delta, append span elements of different classes
			if(d instanceof DeleteDelta) {
				old_rev += "<span class=\"delta-delete\">";
			} else if(d instanceof AddDelta) {
				new_rev += "<span class=\"delta-insert\">";
			} else if(d instanceof ChangeDelta) {
				old_rev += "<span class=\"delta-change\">";
				new_rev += "<span class=\"delta-change\">";
			}
			
			// copy content of chunk to result strings
			old_rev += joinWords(old_content,c_from.first(),c_from.size());
			new_rev += joinWords(new_content,c_to.first(),c_to.size());
			
			// depending on the kind of delta, append span elements of different classes
			if(d instanceof DeleteDelta) {
				old_rev += "</span>";
			} else if(d instanceof AddDelta) {
				new_rev += "</span>";
			} else if(d instanceof ChangeDelta) {
				old_rev += "</span>";
				new_rev += "</span>";
			}
			
			
			// set pointers to end of chunk
			old_ptr = c_from.last()+1;
			new_ptr = c_to.last()+1;
		}

		// add remaining content to result strings and terminate divs
		old_rev += joinWords(old_content,old_ptr,old_content.length-1-old_ptr);
		new_rev += joinWords(new_content,new_ptr,new_content.length-1-new_ptr);
		
		old_rev +="</div>";
		new_rev +="</div>";
		
		
		
		return old_rev.replace("\n","<br/>") + new_rev.replace("\n","<br/>");
	}


	@Override
	public String getTitlePrefix() {
		return "History: ";
	}
	
	
	
}
