/**
 *  Copyright (C) 2005
 *  Sebastian Schaffert and Rupert Westenthaler
 *  Salzburg Research Forschungsgesellschaft m.b.H.
 *
 *  IkeWiki is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU Lesser General Public
 * 	License as published by the Free Software Foundation; either
 * 	version 2 of the License, or (at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * 	Lesser General Public License for more details.
 *
 * 	You should have received a copy of the GNU Lesser General Public
 * 	License along with this library; if not, write to the
 * 	Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * 	Boston, MA 02111-1307, USA.
 * 
 */
package at.srfg.ikewiki.layout.views;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Hashtable;
import java.util.ResourceBundle;

import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import at.srfg.ikewiki.Main;
import at.srfg.ikewiki.data.Namespace;
import at.srfg.ikewiki.data.WikiLiteral;
import at.srfg.ikewiki.data.WikiNode;
import at.srfg.ikewiki.data.WikiResource;
import at.srfg.ikewiki.db.DBException;
import at.srfg.ikewiki.layout.perspectives.Perspective;

/**
 * @author sschaffe
 *
 */
public class BioTaxonomyView extends View {

	private String ns_uri = "http://ikewiki.srfg.at/biology#";

	private String NS_XHTML = "http://www.w3.org/1999/xhtml";
	
	private Collection<WikiResource> properties;
	
	
	private WikiResource r_scientific_name, r_discovered_by;
	
	/**
	 * @param main
	 */
	public BioTaxonomyView(Main main) throws DBException {
		super(main);

		properties = new ArrayList<WikiResource>();
		
		Namespace myns = main.getStorageBackend().getNamespaceByUri(ns_uri);
		
		// the property names recognised by the BioTaxonomy view
		String[] prop_names = { "hasDomain", "hasKingdom", 
								"hasDivision", "hasSubdivision", 
								"hasClass", "hasSubclass",
								"hasOrder", "hasSuborder",
								"hasFamily", "hasSubfamily",
								"hasGenus", "hasSpecies",
		};
		
		for(String prop_name : prop_names) {
			properties.add(new WikiResource(myns,prop_name));
		}

		r_scientific_name = new WikiResource(myns,"hasScientificName");
		r_discovered_by   = new WikiResource(myns,"discoveredBy");
		
	}


	
	
	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.layout.views.View#getType()
	 */
	@Override
	public TYPES getType() {
		return TYPES.CONTENT_START;
	}




	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.layout.views.View#responsibleFor(at.srfg.ikewiki.data.WikiResource)
	 */
	@Override
	public boolean responsibleFor(WikiResource r) {
		try {
			Namespace ns = main.getStorageBackend().getNamespaceByUri(ns_uri);
			return r.getNamespace().equals(ns);
		} catch(DBException ex) {
			return false;
		}
	}

	/**
	 * the taxonomy will be created by not only considering directly specified
	 *  relationships but also the classifications of "parent" resources in the taxonomy
	 *  for this purpose, we build a Hashtable containing all the relevant properties as keys
	 *  and the associated resources as values
	 *  
	 *  This function recursively calls itself. If the data contains cycles (which are not allowed
	 *  by the ontology but might be specified by the user), it currently does not terminate.
	 *  
	 * @param context
	 * @return
	 */
	private Hashtable<WikiResource,Collection<WikiNode>> buildTaxonomy(WikiResource context) throws DBException {
		Hashtable<WikiResource,Collection<WikiNode>> taxonomy = new Hashtable<WikiResource,Collection<WikiNode>>();

		for(WikiResource property : properties) {

			Collection<WikiNode> objects = main.getStorageBackend().getObjectsBySubjectProperty(context,property);
			
			if(objects.size() > 0) {
				taxonomy.put(property,objects);
				
				for(WikiNode n : objects) {
					if(n.isResource()) {
						// call buildTaxonomy recursively
						Hashtable<WikiResource,Collection<WikiNode>> upper = buildTaxonomy((WikiResource)n);
						
						// copy all values from upper to taxonomy; if a value already exists, the result of the call to
						// upper will be appended
						for(WikiResource key : upper.keySet()) {
							if(taxonomy.get(key) != null) {
								if(!taxonomy.get(key).containsAll(upper.get(key))) {
									taxonomy.get(key).addAll(upper.get(key));
								}
							} else {
								taxonomy.put(key,upper.get(key));
							}
						}
						
					}
				}
			}
		}
		
		return taxonomy;
	}
	
	
	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.layout.views.View#getHTMLContent(at.srfg.ikewiki.data.WikiResource)
	 */
	@Override
	public String getHTMLContent(WikiResource context, Perspective perspective) {
		return "<div class=\"error\">BioTaxonomyView handled via Wiklets</div>";
	}

	
	
	private Node formatWikiNode(WikiNode n, Document root) {
		if(n.isLiteral()) {
			WikiLiteral l = (WikiLiteral)n;
			return root.createTextNode(l.getContent()+" ");
		} else { // isResource()
			WikiResource r = (WikiResource)n;

			Element a = root.createElementNS(NS_XHTML,"a");
			a.setAttribute("name",        r.getQualifiedTitle());
			a.setAttribute("href",        "index.jsp?action=view&title="+r.getQualifiedTitle());
			a.setAttribute("onmouseover", "showWMTT('tt-"+r.getQualifiedTitle()+"')");
			a.setAttribute("onmouseout",  "hideWMTT()");
			
			a.appendChild(root.createTextNode(r.getTitle()));
			
			return a;
		}
	}
	
	
	
	/**
	 * Create the view as HTML DOM for the given WikiResource
	 * 
	 * @param context     the WikiResource to be shown in this view
	 */
	public DocumentFragment createView(WikiResource context, Document root) {
		DocumentFragment result = root.createDocumentFragment();

		Element row, heading, col, e, f;
		
		try {
			ResourceBundle bundle = ResourceBundle.getBundle("at.srfg.ikewiki.layout.views.BioTaxonomyResources",main.getStorageBackend().getLanguage(context));
			
			boolean bild = false;
			
			result.appendChild(root.createComment("content generated from context by BioTaxonomyView"));
			
			// table header
			Element table = root.createElementNS(NS_XHTML,"table");
			result.appendChild(table);
			table.setAttribute("class","taxobox");
			table.appendChild(root.createElementNS(NS_XHTML,"tr"));
			
			heading = root.createElementNS(NS_XHTML,"th");
			table.getLastChild().appendChild(heading);
			heading.appendChild(root.createTextNode(context.getTitle()));
			
			
			// image, we need to check whether it exists (currently substituted by the boolean "bild")
			if(bild) {
				row = root.createElementNS(NS_XHTML,"tr");
				table.appendChild(row);
				
				col = root.createElementNS(NS_XHTML,"td");
				row.appendChild(col);
				col.setAttribute("class","taxo-bild");
				
				Element div = root.createElementNS(NS_XHTML,"div");
				col.appendChild(div);
				div.setAttribute("class","thumb");

				// image still missing
			}
			
			// the real taxonomy
			row = root.createElementNS(NS_XHTML,"tr");
			table.appendChild(row);
			
			heading = root.createElementNS(NS_XHTML,"th");
			row.appendChild(heading);
			
			e = root.createElementNS(NS_XHTML,"b");
			heading.appendChild(e);
			e.appendChild(root.createTextNode(bundle.getString("Taxonomy")));
			

			row = root.createElementNS(NS_XHTML,"tr");
			table.appendChild(row);

			col = root.createElementNS(NS_XHTML,"td");
			row.appendChild(col);

			e = root.createElementNS(NS_XHTML,"table");
			col.appendChild(e);
			
			
			Hashtable<WikiResource,Collection<WikiNode>> taxonomy = buildTaxonomy(context);
			
			for(WikiResource property : properties) {

				if(taxonomy.get(property) != null) {
					row = root.createElementNS(NS_XHTML,"tr");
					e.appendChild(row);
					
					col = root.createElementNS(NS_XHTML,"td");
					row.appendChild(col);

					f = root.createElementNS(NS_XHTML,"i");
					col.appendChild(f);
					
					f.appendChild(root.createTextNode(bundle.getString(property.getTitle())+":"));
					
					
					col = root.createElementNS(NS_XHTML,"td");
					row.appendChild(col);
					
					
					
										
					for(WikiNode n : taxonomy.get(property)) {
						col.appendChild(formatWikiNode(n,root));
					}
					
				}
			}
			
			
			Collection<WikiNode> scnames = main.getStorageBackend().getObjectsBySubjectProperty(context,r_scientific_name);
			Collection<WikiNode> dby     = main.getStorageBackend().getObjectsBySubjectProperty(context,r_discovered_by);
			
			if(scnames.size() > 0 || dby.size() > 0) {
				row = root.createElementNS(NS_XHTML,"tr");
				table.appendChild(row);
				
				heading = root.createElementNS(NS_XHTML,"th");
				row.appendChild(heading);
				
				heading.appendChild(root.createTextNode(bundle.getString("Nomenclature")));
				
				
				
				if(scnames.size() > 0) {
					row = root.createElementNS(NS_XHTML,"tr");
					table.appendChild(row);
					
					col = root.createElementNS(NS_XHTML,"td");
					row.appendChild(col);
					col.setAttribute("class","Person");
					
				
					for(WikiNode n : scnames) {
						col.appendChild(formatWikiNode(n,root));
					}
					
				}

				if(dby.size() > 0) {
					row = root.createElementNS(NS_XHTML,"tr");
					table.appendChild(row);
					
					col = root.createElementNS(NS_XHTML,"td");
					row.appendChild(col);
					col.setAttribute("class","taxo-name");
				
					for(WikiNode n : dby) {
						col.appendChild(formatWikiNode(n,root));
					}
				}
			}
			
		} catch(DBException ex) {
		}
		
		return result;
		
	}

}
