/**
 *  Copyright (C) 2005
 *  Sebastian Schaffert and Rupert Westenthaler
 *  Salzburg Research Forschungsgesellschaft m.b.H.
 *
 *  IkeWiki is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU Lesser General Public
 * 	License as published by the Free Software Foundation; either
 * 	version 2 of the License, or (at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * 	Lesser General Public License for more details.
 *
 * 	You should have received a copy of the GNU Lesser General Public
 * 	License along with this library; if not, write to the
 * 	Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * 	Boston, MA 02111-1307, USA.
 * 
 */
package at.srfg.ikewiki.layout.views;

import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.image.BufferedImage;
import java.util.Collection;
import java.util.Iterator;

import at.srfg.ikewiki.Main;
import at.srfg.ikewiki.data.Namespace;
import at.srfg.ikewiki.data.WikiNode;
import at.srfg.ikewiki.data.WikiResource;
import at.srfg.ikewiki.db.DBException;
import at.srfg.ikewiki.layout.perspectives.Perspective;

/**
 * A simple view that displays the RDFS/OWL class hierarchy with a context of 1
 * 
 * @author sschaffe
 *
 */
public class RDFSClassHierarchyView extends View {

	private String ns_uri = "http://www.w3.org/2000/01/rdf-schema#";
	
	WikiResource p_subclassof;

	
	Graphics2D myGraphics;
	
	/**
	 * @param main
	 */
	public RDFSClassHierarchyView(Main main) throws DBException {
		super(main);

		Namespace myns = main.getStorageBackend().getNamespaceByUri(ns_uri);
		
		p_subclassof = new WikiResource(myns,"subClassOf"); 

		// used for calculating string sizes with font metrics
		BufferedImage img = new BufferedImage(100,100,BufferedImage.TYPE_INT_RGB);
		myGraphics = img.createGraphics();
		myGraphics.setFont(new Font(getFontName(),Font.PLAIN,getFontSize()));
	}


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.layout.views.View#getType()
	 */
	@Override
	public TYPES getType() {
		return TYPES.PORTLET_RIGHT;
	}


	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.layout.views.View#responsibleFor(at.srfg.ikewiki.data.WikiResource)
	 */
	@Override
	public boolean responsibleFor(WikiResource r) {
		/*try {
			Namespace ns = main.getStorageBackend().getNamespaceByUri(ns_uri);
			return r.getNamespace().equals(ns);
		} catch(DBException ex) {
			return false;
		}*/
		return p_subclassof.equals(r);
	}

	
	public int getMaximumWidth() {
		return 170;
	}
	
	
	public int getMinimumVerticalSpacing() {
		return 15;
	}
	
	public int getMinimumHorizontalSpacing() {
		return 10;
	}
	
	public int getMaximumClasses() {
		return 5;
	}
	
	public int getFontSize() {
		return 9;
	}
	
	public String getFontName() {
		return "Arial";
	}
	
	/**
	 * Trim the number of elements in collection to at most getMaximumClasses()
	 * @param collection
	 * @return
	 */
	private int trimClasses(Collection collection) {
		if(collection.size() > getMaximumClasses()) {
			int notshown = collection.size() - getMaximumClasses();
			int i = 0;
			for(Iterator it=collection.iterator(); it.hasNext(); ) {
				i++;it.next();
				if(i > 5)
					it.remove();
			}
			return notshown;
		} else {
			return 0;
		}
		
	}
	
	

	
	
	private void paintClass(StringBuilder result,int curX, int curY, int rectWidth, int rectHeight, String label, String qtitle, String col, String fill) {
		
		result.append("<a xlink:href=\"index.jsp?title="+Main.encodeURL(qtitle)+"&amp;action=view\"><g>\n");
		// draw rectangle						
		paintRect(result,curX,curY,rectWidth,rectHeight,"stroke=\""+col+"\" fill=\""+fill+"\" opacity=\"0.5\"");
		
		String mylabel = label;

		
		int strWidth = myGraphics.getFontMetrics().stringWidth(label);
		int offset = (rectWidth - strWidth) / 2; // center text;		
		
		while(offset < 2 && mylabel.length() > 6) {
			mylabel = mylabel.substring(0,mylabel.length()-4) + "...";
			strWidth = myGraphics.getFontMetrics().stringWidth(mylabel);
			offset = (rectWidth - strWidth) / 2;
		}
		
		
		//g2d.drawString(mylabel,curX+offset,curY+rectHeight-3);

		
		result.append("<text font-stretch=\"condensed\" fill-opacity=\"1\" fill=\"black\" xml:space=\"preserve\" stroke=\"none\" text-anchor=\"middle\" x=\"" + (curX+rectWidth/2) + "\" y=\""+(curY+rectHeight-3)+"\">");
		result.append(mylabel);
		result.append("</text>\n");
		
		
		result.append("</g></a>");
	}
	
	private void paintRect(StringBuilder result, int x, int y, int width, int height, String attributes) {
		result.append("<rect x=\""+x+"\" y=\""+y+"\" width=\""+width+"\" height=\""+height+"\" "+attributes+" />");		
	}
	
	private void paintLine(StringBuilder result, int x1, int y1, int x2, int y2) {
		result.append("<line x1=\""+x1+"\" y1=\""+y1+"\" x2=\""+x2+"\" y2=\""+y2+"\" />\n");		
	}
	
	
	private SVGContent paint(WikiResource context) throws DBException {
//		Collection<WikiNode> superclasses   = main.getStorageBackend().getObjectsBySubjectProperty(context,p_subclassof);
//		Collection<WikiResource> subclasses = main.getStorageBackend().getSubjectsByPropertyObject(p_subclassof,context);
		Collection<WikiResource> superclasses   = main.getStorageBackend().getSuperClasses(context,true);
		Collection<WikiResource> subclasses = main.getStorageBackend().getSubClasses(context,true);
		
		SVGContent result = new SVGContent();
		
		
		int notshown = trimClasses(superclasses) + trimClasses(subclasses);
		
		
		// vertical spacing
		int vertSpacing = getMinimumVerticalSpacing();
		
		// maximum width of the canvas
		int maxWidth = getMaximumWidth();
		int minSpacing  = getMinimumHorizontalSpacing();
		
		// width of each rectangle for a class, calculated as (maxWidth-spacing)/number of canvases - spacing 
		int rectWidth = (maxWidth - minSpacing)/(Math.max(superclasses.size(),subclasses.size())) - minSpacing;
		if(rectWidth < 10)
			rectWidth = 10;
		int rectHeight = 15;
		
		int curX = 0, curY = vertSpacing, i = 0;


		// store anchor points for arrows
		Point[] anchorsSuper = new Point[superclasses.size()];
		
		// *** superclasses ***
		if(superclasses.size() > 0) {			
			
			result.content.append("<g id=\"superclasses\" style=\"fill-opacity:0.7;\">\n");
			
			// calculate spacing for first row
			int spacingSuper = (maxWidth - (superclasses.size() * rectWidth)) / (superclasses.size() + 1);
			
			curX = spacingSuper; i = 0;
			for(WikiResource sup : superclasses) {
				
				paintClass(result.content, curX,curY,rectWidth,rectHeight,sup.getTitle(),sup.getQualifiedTitle(),"black","blue");
				
				anchorsSuper[i] = new Point(curX + rectWidth/2,curY+rectHeight);
				
				
				curX += rectWidth + spacingSuper;
				i++;
			}
			curY += rectHeight + vertSpacing;
			curX = 0;
			
			result.content.append("</g>\n");
		}
		
		// *** current class ***
		result.content.append("<g id=\"current-class\" style=\"fill-opacity:0.7;\">\n");
		
		// calculate spacing for second row
		int spacingCurrent = (maxWidth - rectWidth) / 2;
		
		// paint current class
        paintClass(result.content,curX+spacingCurrent,curY,rectWidth,rectHeight,context.getTitle(),context.getQualifiedTitle(),"black","red");

		result.content.append("</g>\n");
        
        
        if(superclasses.size() > 0) {
        	// paint lines from superclasses to current class
        	result.content.append("<g stroke=\"black\" stroke-width=\"1\">\n");

        	for(Point p : anchorsSuper) {
        		paintLine(result.content,p.x,p.y,maxWidth/2,curY);
        	}
        	result.content.append("</g>\n");
        }
        	
		curY += rectHeight + vertSpacing;
		curX = 0;
        
		// *** subclasses ***
		if(subclasses.size() > 0) {
			// calculate spacing for last row
			int spacingSub = (maxWidth - (subclasses.size() * rectWidth)) / (subclasses.size() + 1);
			
			// store anchor points for arrows
			Point[] anchorsSub = new Point[subclasses.size()];

			result.content.append("<g id=\"subclasses\" style=\"fill-opacity:0.7;\">\n");

			curX = spacingSub; i = 0;
			for(WikiNode n_sub : subclasses) {
				WikiResource sub = (WikiResource)n_sub;
				
				paintClass(result.content,curX,curY,rectWidth,rectHeight,sub.getTitle(),sub.getQualifiedTitle(),"black","green");
				
				anchorsSub[i] = new Point(curX + rectWidth/2,curY);
				
				
				curX += rectWidth + spacingSub;
				i++;
			}
			
			result.content.append("</g>\n");
			
			// paint lines
        	result.content.append("<g stroke=\"black\" stroke-width=\"1\">\n");
			for(Point p : anchorsSub) {
				paintLine(result.content,p.x,p.y,maxWidth/2,curY-vertSpacing);			
			}
        	result.content.append("</g>\n");
			
			curY += rectHeight + vertSpacing;
			curX = 0;
		}
		
		int extraSpace = 0;
		if(notshown > 0) {
			result.content.append("<g id=\"warning\">\n");
			
			
			String message = notshown + " classes not shown";
			int offset = maxWidth / 2; // center text;		

			result.content.append("<text fill=\"red\" xml:space=\"preserve\" stroke=\"none\" text-anchor=\"middle\" x=\"" + (curX+offset) + "\" y=\""+(curY+rectHeight-3)+"\">");
			result.content.append(message);
			result.content.append("</text>\n");

			result.content.append("</g>\n");
			
			extraSpace = rectHeight + vertSpacing;
		}
		
		
        result.setSVGCanvasSize(maxWidth,4*vertSpacing + 3 * rectHeight + extraSpace);
		
        return result;
	}
	
	
	private String createSVG(WikiResource context) throws DBException {
		
		StringBuilder result = new StringBuilder();
		
		// create the content of the graphics
		SVGContent content = paint(context);
		
		// format text as Arial, 9pt
		String fontFamily = getFontName();
		int fontSize = getFontSize();
		
		result.append("<svg xmlns:xlink=\"http://www.w3.org/1999/xlink\" width=\""+content.width+"\" height=\""+content.height+"\" xmlns=\"http://www.w3.org/2000/svg\" font-family=\""+fontFamily+"\" font-size=\""+fontSize+"\">");
		result.append("<style type=\"text/css\">rect:hover { fill-opacity:0.9;} </style>\n");
		result.append(content.content);
		result.append("</svg>\n");
		
		/*
		// Get a DOMImplementation
        DOMImplementation domImpl =
            GenericDOMImplementation.getDOMImplementation();

        // Create an instance of org.w3c.dom.Document
        Document document = domImpl.createDocument(null, "svg", null);

        // Create an instance of the SVG Generator
        SVGGraphics2D svgGenerator = new SVGGraphics2D(document);
        
        paint(svgGenerator,context);
        
        boolean useCSS = false;
        
        StringWriter out = new StringWriter();
        
        Element root = svgGenerator.getRoot();
        try {
        	svgGenerator.stream(root,out, useCSS);
        } catch(IOException ex) { 
        	ex.printStackTrace();
        }
         
        String result = out.toString();
        // a hack to drop the XML declaration
        int start = result.indexOf("<svg");
        
        return result.substring(start);
        */
		return result.toString();
	}
	
	
	private class SVGContent {
		StringBuilder content;
		int height, width;
		
		SVGContent() {
			content = new StringBuilder();
		}
		
		void setSVGCanvasSize(int width, int height) {
			this.width = width;
			this.height = height;
		}
	}
	
	
	
	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.layout.views.View#getHTMLContent(at.srfg.ikewiki.data.WikiResource)
	 */
	@Override
	public String getHTMLContent(WikiResource context, Perspective perspective) {
		try {
			
			
			
			StringBuilder result = new StringBuilder();
			
			result.append("<!-- generated by RDFSClassHierarchyView view -->\n");
			result.append("<div id=\"p-class-hierarchy\" class=\"portlet\">\n");
			result.append("<h5>class hierarchy</h5>\n");
			result.append("<div class=\"pBody\">\n");
					
			
			result.append(createSVG(context));
			
			//result.append("<svg xmlns=\"http://www.w3.org/2000/svg\" width=\"300\" height=\"200\" version=\"1.1\">");
            //result.append("<circle cx=\"150\" cy=\"100\" r=\"50\" />");
            //result.append("</svg>\n");
			
			result.append("</div>\n");
			result.append("</div>\n");
			
			return result.toString();
			
		} catch(DBException ex) {
			return "";
		}
	}

}
