/**
 *  Copyright (C) 2006
 *  Sebastian Schaffert and Rupert Westenthaler
 *  Salzburg Research Forschungsgesellschaft m.b.H.
 *
 *  IkeWiki is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU Lesser General Public
 * 	License as published by the Free Software Foundation; either
 * 	version 2 of the License, or (at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * 	Lesser General Public License for more details.
 *
 * 	You should have received a copy of the GNU Lesser General Public
 * 	License along with this library; if not, write to the
 * 	Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * 	Boston, MA 02111-1307, USA.
 * 
 */
package at.srfg.ikewiki.permission;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import at.srfg.ikewiki.Main;
import at.srfg.ikewiki.actions.Action;
import at.srfg.ikewiki.data.Permission;
import at.srfg.ikewiki.data.Namespace;
import at.srfg.ikewiki.data.Role;
import at.srfg.ikewiki.data.User;
import at.srfg.ikewiki.data.WikiLink;
import at.srfg.ikewiki.data.WikiResource;
import at.srfg.ikewiki.db.DBException;

/**
 * User- and permission management.
 * 
 * @author wastl
 *
 */
public class PermissionManager {

	private static Log log = LogFactory.getLog(PermissionManager.class);

	private Main main;
	
	/**
	 * 
	 */
	public PermissionManager(Main main) {
		super();
		
		this.main = main;
	}

	/**
	 * Authenticate a user identified by login and password against the credentials
	 * stored in the database and return the User object associated with the user.
	 * 
	 * @param login
	 * @param password
	 * @return
	 */
	public User authenticateUser(String login, String password) {
		try {
			User u = main.getStorageBackend().getUser(login,password);
			
			log.info("user "+login+" logged in successfully.");
			
			return u;
		} catch(DBException ex) {
			
			log.warn("user "+login+": authentication failure");
			
			return null;
		}
	}
		
	
	/**
	 * Create a user with the given login, first/last name, email, and password and return the
	 * User object.
	 * 
	 * @param login
	 * @param firstName
	 * @param lastName
	 * @param email
	 * @param password
	 * @return
	 * @throws DBException
	 */
	public User createUser(String login, String firstName, String lastName, String email, String password) throws DBException {
		return createUser(login,firstName,lastName,email,password,"Authenticated");
	}
	
	
	/**
	 * Create a user with the given login, first/last name, email, and password and return the
	 * User object.
	 * 
	 * @param login     the login name of the user
	 * @param firstName the first name of the user
	 * @param lastName  the last name of the user
	 * @param email     the email address of the user; stored as foaf:mbox in the knowledge base
	 * @param password  the initial password of the user
	 * @param base_role the initial role of the user
	 * @return
	 * @throws DBException
	 */
	public User createUser(String login, String firstName, String lastName, String email, String password, String base_role) throws DBException {
		User u = new User(login,firstName,lastName,email);
		main.getStorageBackend().storeUser(u,password);
		
		Role r = main.getStorageBackend().getRole(base_role);
		main.getStorageBackend().assignRole(u,r);
		
		Permission p = new Permission(Permission.Type.PERMISSION,3,Namespace.NS_IKEWIKI_USER+login,"%","grant user all permissions on his own page");
		main.getStorageBackend().addPermission(u,p);
		
		return u;
	}
	
	
	
	/**
	 * Save a possibly modified user in the database.
	 * 
	 * @param u
	 * @throws DBException
	 */
	public void saveUser(User u) throws DBException {
		main.getStorageBackend().updateUser(u);		
	}

	
	public User getAnonymousUser() throws DBException {
		return main.getStorageBackend().getUser("Anonymous");
	}
	
	
	/**
	 * Checks whether the user "u" has the permission for "action" on object "o".
	 * 
	 * Permissions are checked against the database (see DBStore for concrete implementation). 
	 * 
	 * @param u
	 * @param o
	 * @param action
	 * @return
	 */
	public boolean hasPermission(User u, PermissionObject o, Action action) throws DBException {
		
		boolean result = main.getStorageBackend().hasPermission(u,o.getObjectIdentifier(),action.getObjectIdentifier());

		// in case of WikiResources, check the namespace as well
		if(result && o instanceof WikiResource) {
			// disabled for the moment
//			WikiResource r = (WikiResource)o;
//			result = hasPermission(u,r.getNamespace(),action);
		}
		
		log.debug( (result?"granted":"denied") + " user \"" + u.getLogin() + "\" action \""+action.getObjectIdentifier()+"\" on object identified by \""+o.getObjectIdentifier()+"\"");
		
		return result;
	}

	/**
	 * Check permission of a WikiLink. This method tests all three resources of the triple.
	 * 
	 * @param u
	 * @param l
	 * @param action
	 * @return
	 */
	public boolean hasPermission(User u, WikiLink l, Action action) throws DBException {
		if(l.getTarget().isResource()) {
			return hasPermission(u,l.getSource(),action) && hasPermission(u,l.getProperty(),action) && hasPermission(u,(WikiResource)l.getTarget(),action);
		} else {
			return hasPermission(u,l.getSource(),action) && hasPermission(u,l.getProperty(),action);
		}
	}
}
