/**
 *  Copyright (C) 2005
 *  Sebastian Schaffert and Rupert Westenthaler
 *  Salzburg Research Forschungsgesellschaft m.b.H.
 *
 *  IkeWiki is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU Lesser General Public
 * 	License as published by the Free Software Foundation; either
 * 	version 2 of the License, or (at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * 	Lesser General Public License for more details.
 *
 * 	You should have received a copy of the GNU Lesser General Public
 * 	License along with this library; if not, write to the
 * 	Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * 	Boston, MA 02111-1307, USA.
 * 
 */
package at.srfg.ikewiki.render;


import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.URIResolver;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import at.srfg.ikewiki.Main;
import at.srfg.ikewiki.data.Article;
import at.srfg.ikewiki.data.Namespace;
import at.srfg.ikewiki.data.WikiException;
import at.srfg.ikewiki.data.WikiLink;
import at.srfg.ikewiki.data.WikiResource;
import at.srfg.ikewiki.db.DBException;
import at.srfg.ikewiki.misc.CanonicalXMLWriter;
import at.srfg.ikewiki.misc.XMLWriter;
import at.srfg.ikewiki.render.engine.wikipedia.TokenMgrError;
import at.srfg.ikewiki.render.engine.wikipedia.WikipediaSerializer;
import at.srfg.ikewiki.render.engine.wikipedia.WikipediaXMLRenderEngine;
import at.srfg.ikewiki.render.wiklet.Wiklet;

/**
 * The rendering pipeline transforms a page that is originally stored in structured Wiki text in the
 * following steps:
 * 1. transform structured text into the XML wiki exchange syntax (with IkeWiki extensions)
 * 2. add wiklets to the XML data (could be XSLT)
 * 3. transform XML to (X)HTML using an XSLT stylesheet
 * 
 * 
 * @author sschaffe
 *
 */
public class RenderingPipeline implements Renderer {

	public static final String NS_IKEWIKI_CORE = "http://ikewiki.srfg.at/syntax/1.0/core";
	public static final String NS_IKEWIKI_EXT  = "http://ikewiki.srfg.at/syntax/1.0/ext";

	private static Log log = LogFactory.getLog(RenderingPipeline.class);

	private static final String XSLT_HTML       = "transform-html.xsl";
	private static final String XSLT_MOZILE     = "transform-mozile.xsl";
	private static final String XSLT_TOOLTIP    = "transform-tooltip.xsl";
	private static final String XSLT_SEARCH     = "transform-search.xsl";
	private static final String XSLT_RECENT     = "transform-recent.xsl";
	private static final String XSLT_RECENT_RSS = "transform-recent-rss.xsl";
	private static final String XSLT_FORM       = "transform-form.xsl";
	private static final String XSLT_SPARQL     = "transform-sparql.xsl";
	private static final String XSLT_REFERENCES = "transform-references.xsl";
	private static final String XSLT_HTML_TO_WIF= "transform-html-wif.xsl";
	
	
	// XSLT transformers used by this renderer
	private String[] TRANSFORMERS = {
			XSLT_HTML, XSLT_MOZILE, XSLT_TOOLTIP, XSLT_FORM, XSLT_SPARQL, XSLT_SEARCH, XSLT_REFERENCES,
			XSLT_HTML_TO_WIF, XSLT_RECENT, XSLT_RECENT_RSS
	};
	
	private HashMap<String,Transformer> transformer_cache;
	
	private URIResolver resolver;
	
	private DateFormat df;
	
	private Main main;
	
	// wiklets used by this renderer
	private String[] WIKLETS = {
			"at.srfg.ikewiki.render.wiklet.ExistsLinkWiklet",
			"at.srfg.ikewiki.render.wiklet.AddLinkTypesWiklet",
			"at.srfg.ikewiki.render.wiklet.AddPageTypeWiklet",
			"at.srfg.ikewiki.render.wiklet.AddDomainRangeWiklet",
			"at.srfg.ikewiki.render.wiklet.AddSuperClassesWiklet",
			"at.srfg.ikewiki.render.wiklet.AddSuperPropertiesWiklet",
			"at.srfg.ikewiki.render.wiklet.OWLClassWiklet",
			"at.srfg.ikewiki.render.wiklet.OWLPropertyWiklet",
			"at.srfg.ikewiki.render.wiklet.AddContentViewsWiklet",
			"at.srfg.ikewiki.render.wiklet.AddInversePropertyWiklet",
			"at.srfg.ikewiki.render.wiklet.AddContextWiklet",
			"at.srfg.ikewiki.render.wiklet.CreativeCommonsWiklet",
			"at.srfg.ikewiki.render.wiklet.SparqlWiklet"
	};
	
	// wiklets applicable to all documents, even document fragments
	private String[] GENERAL_WIKLETS = {
			"at.srfg.ikewiki.render.wiklet.ExistsLinkWiklet",
			"at.srfg.ikewiki.render.wiklet.AddLinkTypesWiklet",
			"at.srfg.ikewiki.render.wiklet.SparqlWiklet"
	};
	
	// wiklets applicable to whole pages
	private String[] PAGE_WIKLETS = WIKLETS;
	
	
	private List<Wiklet> wiklets;
	private HashSet<String> general_wiklets,page_wiklets;
	
	// needed for compatibility methods
	Document lastdoc;

	DocumentBuilderFactory factory;
	DocumentBuilder builder;
	
	
	/**
	 * 
	 */
	public RenderingPipeline(Main main) {
		super();
		
		long start = System.currentTimeMillis();
		
		this.main    = main;

		// initialise RFC822 Date Format
		df = new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss Z",Locale.ENGLISH);

		// use Saxon as XSLT transformer
		System.setProperty("java.xml.transform.TransformerFactory","net.sf.saxon.TransformerFactoryImpl");			
		

		
		// initialise Wiklets
		initWiklets();
		
		// initialise transformers
		initTransformers();

		initParser();
		
		long end = System.currentTimeMillis();
		
		log.info("RenderingPipeline initialised ("+(end-start)+" ms).");
	}
	
	private void initWiklets() {
		wiklets = new LinkedList<Wiklet>();
		
		for(String name : WIKLETS) {
			registerWiklet(name);
		}
		
		general_wiklets = new HashSet<String>(Arrays.asList(GENERAL_WIKLETS));
		page_wiklets    = new HashSet<String>(Arrays.asList(PAGE_WIKLETS));
	}
	

	
	private void initTransformers() {
		resolver = new URIResolver() {

			public Source resolve(String href, String base) throws TransformerException {
				// we do not support external references, just simple ones relative to the current
				// directory
				StreamSource source = new StreamSource(this.getClass().getResourceAsStream(href));
				return source;
			}
			
		};
		
		
		transformer_cache = new HashMap<String,Transformer>();
		
		TransformerFactory factory = TransformerFactory.newInstance();
		factory.setURIResolver(resolver);
		
		for(String tname : TRANSFORMERS) {
			StreamSource stylesheet = new StreamSource(this.getClass().getResourceAsStream(tname));

			try {
				Transformer t = factory.newTransformer(stylesheet);
				
				transformer_cache.put(tname,t);
				
			} catch(TransformerConfigurationException ex) {
				log.error("error while creating XSLT transformer "+tname,ex);
			}
		}
	}
	
	private void initParser() {
			
		
		//factory.setValidating(false);
		
		try {
			factory = DocumentBuilderFactory.newInstance();
			
			// disable loading of external entities over the net
			//factory.setFeature("http://apache.org/xml/features/nonvalidating/load-external-dtd",false);
			
			// enable namespace support
			factory.setNamespaceAware(true);

			builder = factory.newDocumentBuilder();
			
			/*
			 * we use our own entity resolver to resolve the most common DTDs (which are included in the JAR file).
			 * Currently, this includes the XHTML DTDs only. 
			 */
			builder.setEntityResolver(new EntityResolver() {

				/* (non-Javadoc)
				 * @see org.xml.sax.EntityResolver#resolveEntity(java.lang.String, java.lang.String)
				 */
				public InputSource resolveEntity(String publicId, String systemId) throws SAXException, IOException {
					if(publicId.equals("-//W3C//DTD XHTML 1.0 Transitional//EN")) {
						InputStream is = this.getClass().getResourceAsStream("schemas/xhtml/xhtml1-transitional.dtd");
						return new InputSource(is);
					} else if(publicId.equals("-//W3C//DTD XHTML 1.0 Strict//EN")) {
						InputStream is = this.getClass().getResourceAsStream("schemas/xhtml/xhtml1-strict.dtd");
						return new InputSource(is);		
					} else if(publicId.equals("-//W3C//ENTITIES Latin 1 for XHTML//EN")) {
						InputStream is = this.getClass().getResourceAsStream("schemas/xhtml/xhtml-lat1.ent");
						return new InputSource(is);								
					} else if(publicId.equals("-//W3C//ENTITIES Symbols for XHTML//EN")) {
						InputStream is = this.getClass().getResourceAsStream("schemas/xhtml/xhtml-symbol.ent");
						return new InputSource(is);								
					} else if(publicId.equals("-//W3C//ENTITIES Special for XHTML//EN")) {
						InputStream is = this.getClass().getResourceAsStream("schemas/xhtml/xhtml-special.ent");
						return new InputSource(is);								
					} else {
						log.info("could not resolve entity with public id "+publicId+" (tried system id "+systemId+")");
						return null;
					}
				}
				
			});
			
		} catch(ParserConfigurationException ex) {
			log.error("error while initialising XML parser",ex);
		}

	}
	

	private Transformer getTransformer(String tname) throws RenderingException {
		Transformer t = transformer_cache.get(tname);
		
		if(t == null) {
			throw new RenderingException("error: XSLT transformer "+tname+" does not exist; investigate logfile for details");
		} else {
			return t;
		}
	}
	
	
	private void registerWiklet(String classname) {
		try {
			Class cls = Class.forName(classname);
			Constructor cons = cls.getConstructor(Main.class);
			
			wiklets.add((Wiklet)cons.newInstance(this.main));
			
		} catch(ClassNotFoundException ex) {
			log.error("could not register wiklet "+classname+" (no class definition found)");
		} catch(NoSuchMethodException ex) {
			log.error("could not register wiklet "+classname+" (class definition contains errors)",ex);
		} catch(InstantiationException ex) {
			log.error("could not register wiklet "+classname+" (class is abstract)",ex);
		} catch(IllegalAccessException ex) {
			log.error("could not register wiklet "+classname+ " (access not allowed)",ex);
		} catch(InvocationTargetException ex) {
			log.error("could not register wiklet "+classname+" ("+ex.getCause().getLocalizedMessage()+")");
		}

	}
	
	
	public Document createDOM(Article a) throws RenderingException {
		
		if(a.getBaseDom() == null) {
			long start = System.currentTimeMillis();

			processToXML(a);
			
			long end = System.currentTimeMillis();

			//log.debug("transformation from structured text to XML took "+(end-start)+" ms");
		}
		
		
		
		return a.getBaseDom();
	}
	
	/**
	 * Transform an article in structured text to a suitable HTML representation.
	 * 
	 * This method is a wrapper method for the three pipeline transformations, and should be used
	 * by external programs.
	 * 
	 * @param a
	 * @return
	 * @throws RenderingException
	 */
	public String transform(Article a) throws RenderingException {
		long start = System.currentTimeMillis();
		
		Document d;

		if(a.getBaseDom() == null) {
			createDOM(a);
		}
		
		if(a.getExtendedDom() == null) {
			d = processWiklets(a.getBaseDom(false),a);		
			a.setExtendedDom(d);
			lastdoc = d;
		} else {
			d = a.getExtendedDom();
		}
		
		String result =  processToHTML(d,a);
		
		long end = System.currentTimeMillis();
		
		log.debug("transformation from structured text to HTML took "+(end-start)+" ms");
		
		return result;
	}

	/**
	 * Transform an article in structured text to a mozile editable HTML representation.
	 * 
	 * This method is a wrapper method for the three pipeline transformations, and should be used
	 * by external programs.
	 * 
	 * @param a
	 * @return
	 * @throws RenderingException
	 */
	public String transformMozile(Article a) throws RenderingException {
		long start = System.currentTimeMillis();
		
		Document d;

		if(a.getBaseDom() == null) {
			createDOM(a);
		}
		
		if(a.getExtendedDom() == null) {
			d = processWiklets(a.getBaseDom(false),a);		
			a.setExtendedDom(d);
			lastdoc = d;
		} else {
			d = a.getExtendedDom();
		}
		
		String result =  processToMozile(d,a);
		
		long end = System.currentTimeMillis();
		
		log.debug("transformation from structured text to Mozile-HTML took "+(end-start)+" ms");
		
		return result;
	}
	
	
	/**
	 * Transform an article in structured text to an HTML representation suitable as a tooltip.
	 * 
	 * This method is a wrapper method for three pipeline transformations. It differs from transform()
	 * in that it does not apply the complex "Wiklets" and only renders the structured text.
	 * 
	 * @param a
	 * @return
	 * @throws RenderingException
	 */
	public String transformTooltip(Article a) throws RenderingException {
		
		long start = System.currentTimeMillis();
		
		Document d;
		
		if(a.getBaseDom() == null) {
			createDOM(a);
		}
		
		d = a.getBaseDom(true);
		
		//d = processToXML(a);
		
		//a.setDom(d);
		
		String result =  processToTooltip(d,a);
		
		long end = System.currentTimeMillis();
		
		log.debug("transformation from structured text to tooltip took "+(end-start)+" ms");
		
		return result;		
	}
	
	/**
	 * Transform an article in structured text to an HTML representation suitable as a search preview.
	 * 
	 * This method is a wrapper method for three pipeline transformations. It differs from transform()
	 * in that it does not apply the complex "Wiklets" and only renders the structured text.
	 * 
	 * @param a
	 * @return
	 * @throws RenderingException
	 */
	public String transformSearchResult(Article a) throws RenderingException {
		
		long start = System.currentTimeMillis();
		
		if(a.getBaseDom() == null) {
			createDOM(a);
		}
		
		String result =  processToSearchResult(a.getBaseDom(true),a);
		
		long end = System.currentTimeMillis();
		
		log.debug("transformation from structured text to search result took "+(end-start)+" ms");
		
		return result;		
	}

	/**
	 * Transform an article in structured text to an HTML representation suitable as a preview in the
	 * recent changes listing.
	 * 
	 * This method is a wrapper method for three pipeline transformations. It differs from transform()
	 * in that it does not apply the complex "Wiklets" and only renders the structured text.
	 * 
	 * @param a
	 * @return
	 * @throws RenderingException
	 */
	public String transformRecentChangesResult(Article a) throws RenderingException {
		
		long start = System.currentTimeMillis();
		
		if(a.getBaseDom() == null) {
			createDOM(a);
		}
		
		String result =  processToRecentChangesResult(a.getBaseDom(true),a);
		
		long end = System.currentTimeMillis();
		
		log.debug("transformation from structured text to recent changes result took "+(end-start)+" ms");
		
		return result;		
	}
	
	/**
	 * Transform an article in structured text to an HTML representation suitable as a preview in the
	 * recent changes listing.
	 * 
	 * This method is a wrapper method for three pipeline transformations. It differs from transform()
	 * in that it does not apply the complex "Wiklets" and only renders the structured text.
	 * 
	 * @param a
	 * @return
	 * @throws RenderingException
	 */
	public String transformRecentChangesResultRSS(Article a) throws RenderingException {
		
		long start = System.currentTimeMillis();
		
		if(a.getBaseDom() == null) {
			createDOM(a);
		}
		
		String result =  processToRecentChangesResultRSS(a.getBaseDom(true),a);
		
		long end = System.currentTimeMillis();
		
		log.debug("transformation from structured text to recent changes result took "+(end-start)+" ms");
		
		return result;		
	}

	/**
	 * Transform an article in structured text to an HTML representation suitable as an editable form.
	 * 
	 * This method is a wrapper method for three pipeline transformations. It differs from transform()
	 * in that it renders form content as form rather than as values.
	 * 
	 * @param a
	 * @return
	 * @throws RenderingException
	 */
	public String transformForm(Article a) throws RenderingException {
		
		long start = System.currentTimeMillis();
		
		Document d;
		
		if(a.getBaseDom() == null) {
			createDOM(a);
		}
		
		if(a.getExtendedDom() == null) {
			d = processWiklets(a.getBaseDom(false),a);		
			a.setExtendedDom(d);
			lastdoc = d;
		} else {
			d = a.getExtendedDom();
		}
		
		String result =  processToForm(d,a);
		
		long end = System.currentTimeMillis();
		
		log.debug("transformation from structured text to form took "+(end-start)+" ms");
		
		return result;
	}
	
	
	/**
	 * Transform an article in structured text to Wiki interchange XML format
	 * 
	 * This method is a wrapper method for the three pipeline transformations, and should be used
	 * by external programs.
	 * 
	 * @param a
	 * @return
	 * @throws RenderingException
	 */
	public String transformExchange(Article a) throws RenderingException {
		long start = System.currentTimeMillis();
		
		Document d;
		
		if(a.getBaseDom() == null) {
			createDOM(a);
		}
		
		if(a.getExtendedDom() == null) {
			d = processWiklets(a.getBaseDom(false),a);		
			a.setExtendedDom(d);
			lastdoc = d;
		} else {
			d = a.getExtendedDom();
		}

		d.normalizeDocument();
		
		try {
			XMLWriter writer = new CanonicalXMLWriter(d);
			
			String result = writer.writeToString();

			long end = System.currentTimeMillis();
			
			log.debug("transformation from structured text to XML interchange took "+(end-start)+" ms");
			
			return result;
		} catch(WikiException ex) {
			log.error(ex);
			throw new RenderingException(ex);
		}
			
	}

	
	/**
	 * Create the references view out of an article; currently, this is done separately, but we
	 * should consider doing this in the main html form.
	 * @param a
	 * @return
	 * @throws RenderingException
	 */
	public String transformReferences(Article a) throws RenderingException {
		long start = System.currentTimeMillis();
		
		Document d;
		
		if(a.getBaseDom() == null) {
			createDOM(a);
		}
		
		if(a.getExtendedDom() == null) {
			d = processWiklets(a.getBaseDom(false),a);		
			a.setExtendedDom(d);
			lastdoc = d;
		} else {
			d = a.getExtendedDom();
		}
		
		String result =  processToReferences(d,a);
		
		long end = System.currentTimeMillis();
		
		log.debug("transformation from structured text to HTML references view took "+(end-start)+" ms");
		
		return result;
		
	}

	
	
	/**
	 * Format SPARQL results using the defined stylesheet
	 * @param a
	 * @return
	 * @throws RenderingException
	 */
	public String transformSparql(Document d) throws RenderingException {
		StringWriter out = new StringWriter();
		try {
			Transformer t = getTransformer(XSLT_SPARQL);
			
			DOMSource ds = new DOMSource(d);
			ds.setSystemId(d.getDocumentURI()!=null?d.getDocumentURI():"http://ikewiki.srfg.at/Internal");
			
			t.transform(ds,new StreamResult(out));

			return out.toString();
			
		} catch(TransformerException ex) {
			log.error("error while applying XML->HTML transformation",ex);
			throw new RenderingException(ex);
		}
		
	}
	
	
	public Article transformHTMLToArticle(String user, String language, String html) throws RenderingException {
		
		//String content = "<div xmlns=\""+Namespace.NS_XHTML+"\" MozileID=\""+qtitle+"\">"+html+"</div>";
		
		try {
			Document html_dom = parseXML(html);

			Element e_editor = html_dom.getElementById("editor");
			String qtitle = e_editor.getAttribute("qtitle");

			
			WikiResource base_resource = main.getResourceByQTitle(qtitle);
			
			Document d = processHTMLToWIF(base_resource.getTitle(),qtitle,user,language,html_dom);
			d.normalize();
			
			Article a = new Article(base_resource.getNamespace(),base_resource.getTitle(),"transformed from XHTML",d,language,new Date());
			
			a.setText(processWIFToStructured(d));
			
			return a;
		} catch(DBException ex) {
			log.error(ex);
			throw new RenderingException(ex);
		} catch(Throwable t) {
			log.error(t);
			throw new RenderingException(t);
			
		}
		
	}
	
	/**
	 * Transform a wiki interchange format (WIF) fragment to the corresponding HTML fragment.
	 * 
	 * @param fragment
	 * @param context
	 * @return
	 * @throws RenderingException
	 */
	public String transformWIFFragment(String fragment, WikiResource context) throws RenderingException {
		long start = System.currentTimeMillis();
		Document d = parseXML(fragment);
		d = processWiklets(d,context,general_wiklets);		
		String result =  processToMozile(d,context);
		long end = System.currentTimeMillis();
		
		log.debug("transformation of WIF fragment took "+(end-start)+" ms");
		
		return result;
	}

	
	
	protected String processWIFToStructured(Document wif) throws RenderingException {
		if(wif == null) {
			throw new RenderingException("Document not available in WIF format");
		}
		WikipediaSerializer serializer = new WikipediaSerializer(wif);
		return serializer.serialize();
	}

	
	
	/**
	 * Transform an article from Mozile-generated XHTML to wiki interchange format.
	 * 
	 * @param d
	 * @param qtitle
	 * @param html
	 * @return
	 * @throws RenderingException
	 */
	protected Document processHTMLToWIF(String title, String qual_title, String user, String language, Document html_dom) throws RenderingException {
		
		
		String stylesheet = XSLT_HTML_TO_WIF;
		
		try {
			Transformer t = getTransformer(stylesheet);
			
			DOMSource ds = new DOMSource(html_dom);
			DOMResult dr = new DOMResult();
			
			ds.setSystemId(html_dom.getDocumentURI()!=null?html_dom.getDocumentURI():"http://ikewiki.srfg.at/Internal");
			

			t.transform(ds,dr);

			Document result = (Document)dr.getNode();

			Element date = result.createElementNS(Namespace.NS_IKEWIKI_CORE,"date");
			date.appendChild(result.createTextNode(df.format(new Date())));
			result.getDocumentElement().insertBefore(date,result.getDocumentElement().getFirstChild());

			Element lang = result.createElementNS(Namespace.NS_IKEWIKI_CORE,"language");
			lang.appendChild(result.createTextNode(language));
			result.getDocumentElement().insertBefore(lang,result.getDocumentElement().getFirstChild());
			
			
			Element author = result.createElementNS(NS_IKEWIKI_CORE,"author");
			if(user != null)
				author.appendChild(result.createTextNode(user));
			result.getDocumentElement().insertBefore(author,result.getDocumentElement().getFirstChild());
			
			
			Element qtitle = result.createElementNS(NS_IKEWIKI_EXT,"qtitle");
			qtitle.appendChild(result.createTextNode(qual_title));
			result.getDocumentElement().insertBefore(qtitle,result.getDocumentElement().getFirstChild());

			Element etitle = result.createElementNS(NS_IKEWIKI_CORE,"title");
			etitle.appendChild(result.createTextNode(title));
			result.getDocumentElement().insertBefore(etitle,result.getDocumentElement().getFirstChild());


			return result;
		} catch(TransformerException ex) {
			log.error("error while applying transformation ("+stylesheet+")",ex);
			throw new RenderingException(ex);
		}
		
	}
	
	
	
	
	protected Document processToXML(Article a) throws RenderingException {
		
		if(a.getBaseDom() == null) {
			WikipediaXMLRenderEngine engine;
			
			try {
				engine = new WikipediaXMLRenderEngine(main,a);
			} catch(ParserConfigurationException ex) {
				log.error("could not initialise XML document builder ",ex);
				throw new RenderingException(ex);
			}
			
			
			try {

				
				Document d = engine.parseDocument();
				
				
				Element date = d.createElementNS(NS_IKEWIKI_CORE,"date");
				date.appendChild(d.createTextNode(df.format(a.getTimestamp())));
				d.getDocumentElement().insertBefore(date,d.getDocumentElement().getFirstChild());

				Element author = d.createElementNS(NS_IKEWIKI_CORE,"author");
				if(a.getCreator() != null)
					author.appendChild(d.createTextNode(a.getCreator()));
				d.getDocumentElement().insertBefore(author,d.getDocumentElement().getFirstChild());
				
				Element qtitle = d.createElementNS(NS_IKEWIKI_EXT,"qtitle");
				qtitle.appendChild(d.createTextNode(a.getQualifiedTitle()));
				d.getDocumentElement().insertBefore(qtitle,d.getDocumentElement().getFirstChild());
				
				Element title = d.createElementNS(NS_IKEWIKI_CORE,"title");
				title.appendChild(d.createTextNode(a.getTitle()));
				d.getDocumentElement().insertBefore(title,d.getDocumentElement().getFirstChild());
				

				
				a.setBaseDom(d);
				
				return a.getBaseDom();
			} catch(TokenMgrError ex) {
				// parsing may never fail, regardless of what kind of exception is thrown; 
				// it is necessary to recover gracefully and give the user the opportunity to modify the
				// content of the page
				
				log.error("error while parsing article "+a.getQualifiedTitle(),ex);
				throw new RenderingException(ex);
			} catch(Exception ex) {
				// parsing may never fail, regardless of what kind of exception is thrown; 
				// it is necessary to recover gracefully and give the user the opportunity to modify the
				// content of the page
				
				log.error("error while parsing article "+a.getQualifiedTitle(),ex);
				throw new RenderingException(ex);
			}
		} else {
			return a.getBaseDom();
		}
	}
	
	/**
	 * This method applies "wiklets" to a document, and returns an "enriched" document that contains
	 * additional information besides the page layout gathered from the structured text.
	 * 
	 * Wiklets include:
	 * - checking whether an internal link exists
	 * - checking whether an internal link is typed (otherwise an "externalise" button will be shown)
	 * 
	 * @param d       the DOM document in Wiki XML exchange format
	 * @param context the article resource this document represents
	 * @return
	 * @throws RenderingException
	 */
	protected Document processWiklets(Document d, WikiResource context) throws RenderingException {
		Document e = d;
		for(Wiklet w : wiklets) {
			e = w.apply(e,context);
		}		
		return e;
	}
	
	/**
	 * This method applies "wiklets" to a document, and returns an "enriched" document that contains
	 * additional information besides the page layout gathered from the structured text.
	 * 
	 * This method differs from processWiklets(Document,WikiResource) in that it only applies the wiklets
	 * whose class name is contained in the Collection given as third argument.
	 * 
	 * Wiklets include:
	 * - checking whether an internal link exists
	 * - checking whether an internal link is typed (otherwise an "externalise" button will be shown)
	 * 
	 * @param d            the DOM document in Wiki XML exchange format
	 * @param context      the article resource this document represents
	 * @param wiklet_names only apply wiklets whose class name is in this collection
	 * @return
	 * @throws RenderingException
	 */
	protected Document processWiklets(Document d, WikiResource context, Collection<String> wiklet_names) throws RenderingException {
		Document e = d;
		for(Wiklet w : wiklets) {
			if(wiklet_names.contains(w.getClass().getName())) {
				e = w.apply(e,context);
			}
		}		
		return e;
	}

	
	
	/**
	 * Generic transformation of DOM document to another representation; used by subsequent methods.
	 * @param d Document to transform
	 * @param context the WikiResource that is the context of this process
	 * @param stylesheet the name of the stylesheet to apply
	 * @return a string representing the transformed document
	 * @throws RenderingException
	 */
	protected String processStylesheet(Document d, WikiResource context, String stylesheet) throws RenderingException {
		StringWriter out = new StringWriter();
		
		try {
			Transformer t = getTransformer(stylesheet);
			t.setParameter("base_uri",main.getBaseUri());
			
			DOMSource ds = new DOMSource(d);
			ds.setSystemId(d.getDocumentURI()!=null?d.getDocumentURI():"http://ikewiki.srfg.at/Internal");
			
			t.transform(ds,new StreamResult(out));

			return out.toString();
			
		} catch(TransformerException ex) {
			log.error("error while applying transformation ("+stylesheet+")",ex);
			throw new RenderingException(ex);
		}
		
		
	}
	
	
	
	/**
	 * This method applies the transformation from the (enriched) XML representation of the page to
	 * the actual HTML representation that will be rendered in the user's browser. See also
	 * transform-html.xsl.
	 * 
	 * @param d       the DOM document in Wiki XML exchange format
	 * @param context the article resource this document represents
	 * @return
	 * @throws RenderingException
	 */
	protected String processToHTML(Document d, WikiResource context) throws RenderingException {
		return processStylesheet(d,context,XSLT_HTML);
	}

	/**
	 * This method applies the transformation from the (enriched) XML representation of the page to
	 * the mozile-editable HTML representation that will be rendered in the user's browser. See also
	 * transform-mozile.xsl.
	 * 
	 * @param d       the DOM document in Wiki XML exchange format
	 * @param context the article resource this document represents
	 * @return
	 * @throws RenderingException
	 */
	protected String processToMozile(Document d, WikiResource context) throws RenderingException {
		return processStylesheet(d,context,XSLT_MOZILE);
	}
	
	
	
	/**
	 * This method applies the transformation from the (enriched or not) XML representation of the page to
	 * the tooltip representation that will be rendered in the user's browser if he hovers the mouse over a link. See 
	 * also transform-tooltip.xsl.
	 * 
	 * Tooltips used to be created using transformHTML before. However, for tooltips it is not necessary to have all
	 * the semantic information available, it just degrades performance. We thus create the tooltips separately.
	 * 
	 * @param d       the DOM document in Wiki XML exchange format
	 * @param context the article resource this document represents
	 * @return
	 * @throws RenderingException
	 */
	protected String processToTooltip(Document d, WikiResource context) throws RenderingException {
		return processStylesheet(d,context,XSLT_TOOLTIP);
	}
	
	/**
	 * This method applies the transformation from the (enriched or not) XML representation of the page to
	 * the form editing representation that allows to change certain values of the content using a form. See 
	 * also transform-form.xsl.
	 * 
	 * @param d       the DOM document in Wiki XML exchange format
	 * @param context the article resource this document represents
	 * @return
	 * @throws RenderingException
	 */
	protected String processToForm(Document d, WikiResource context) throws RenderingException {
		return processStylesheet(d,context,XSLT_FORM);
	}
	
	/**
	 * This method applies the transformation from the (enriched or not) XML representation of the page to
	 * the form editing representation that allows to change certain values of the content using a form. See 
	 * also transform-form.xsl.
	 * 
	 * @param d       the DOM document in Wiki XML exchange format
	 * @param context the article resource this document represents
	 * @return
	 * @throws RenderingException
	 */
	protected String processToReferences(Document d, WikiResource context) throws RenderingException {
		return processStylesheet(d,context,XSLT_REFERENCES);
	}

	
	/**
	 * This method applies the transformation from the (enriched or not) XML representation of the page to
	 * the search result representation that will be displayed in search results.
	 * 	 * 
	 * @param d       the DOM document in Wiki XML exchange format
	 * @param context the article resource this document represents
	 * @return
	 * @throws RenderingException
	 */
	protected String processToSearchResult(Document d, WikiResource context) throws RenderingException {
		return processStylesheet(d,context,XSLT_SEARCH);
	}


	/**
	 * This method applies the transformation from the (enriched or not) XML representation of the page to
	 * the recent changes result representation that will be displayed in the HTML recent changes listing.
	 * 	 * 
	 * @param d       the DOM document in Wiki XML exchange format
	 * @param context the article resource this document represents
	 * @return
	 * @throws RenderingException
	 */
	protected String processToRecentChangesResult(Document d, WikiResource context) throws RenderingException {
		return processStylesheet(d,context,XSLT_RECENT);
	}


	/**
	 * This method applies the transformation from the (enriched or not) XML representation of the page to
	 * the recent changes result representation that will be displayed in the RSS recent changes listing.
	 * 	 * 
	 * @param d       the DOM document in Wiki XML exchange format
	 * @param context the article resource this document represents
	 * @return
	 * @throws RenderingException
	 */
	protected String processToRecentChangesResultRSS(Document d, WikiResource context) throws RenderingException {
		return processStylesheet(d,context,XSLT_RECENT_RSS);
	}
	
	
	/**
	 * Retrieve a list of all links occurring in an article.
	 * @param a
	 * @return
	 * @throws DBException
	 */
	public List<WikiLink> getReferences(Article a) throws DBException {
		try {
			if(a.getExtendedDom() == null) {
				transform(a);
			}
			
			List<WikiLink> references = new LinkedList<WikiLink>();
			
			NodeList nl = a.getExtendedDom().getElementsByTagNameNS(Namespace.NS_IKEWIKI_CORE,"intlink");
			
			for(int i = 0; i<nl.getLength(); i++) {
				Element link = (Element)nl.item(i);
				
				//if(! (link.getAttribute("create").toLowerCase().equals("yes") || link.getAttribute("create").toLowerCase().equals("true")) ) {
					references.add(new WikiLink(a,null,main.getResourceByQTitle(link.getAttribute("target")),""));
				//}
			}
			
			return references;
		} catch(RenderingException ex) {
			log.warn(ex);
			return Collections.emptyList();
		}
	}
	
	/**
	 * Parse an XML document represented as a String into a DOM structure.
	 * @param content the String representation of the XML document
	 * @return the corresponding DOM structure
	 * @throws RenderingException
	 */
	public Document parseXML(String content) throws RenderingException {		
		try {
			long start = System.currentTimeMillis();

			Document d = builder.parse(new InputSource(new StringReader(content.trim())));

			long end = System.currentTimeMillis();

			
			log.debug("XML content parsed in "+(end-start)+" ms");
			
			return d;
		} catch(SAXException ex) {
			log.error("error parsing XML content of article",ex);
			return null;
		} catch(IOException ex) {
			// never happens for ByteArrayInputStream
			log.fatal("IO error parsing XML content of article",ex);
			return null;
		} catch(Exception ex) {
			log.fatal("other error parsing XML content of article",ex);
			return null;
		}
	}

	
	
}
