/**
 *  Copyright (C) 2005-2006
 *  Sebastian Schaffert and Rupert Westenthaler
 *  Salzburg Research Forschungsgesellschaft m.b.H.
 *
 *  IkeWiki is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the
 *  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *  Boston, MA 02111-1307, USA.
 *
 *  WikipediaXMLRenderEngineBase.java
 */

package at.srfg.ikewiki.render.engine.wikipedia;

import java.io.ByteArrayInputStream;
import java.io.OutputStream;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import at.srfg.ikewiki.Main;
import at.srfg.ikewiki.data.Article;
import at.srfg.ikewiki.data.WikiException;
import at.srfg.ikewiki.data.WikiLink;
import at.srfg.ikewiki.data.WikiResource;
import at.srfg.ikewiki.misc.StandardXMLWriter;
import at.srfg.ikewiki.misc.XMLWriter;

/**
 * @author sschaffe
 *
 */
public abstract class WikipediaXMLRenderEngineBase {
	
	public static final String NS_IKEWIKI_CORE = "http://ikewiki.srfg.at/syntax/1.0/core";
	public static final String NS_IKEWIKI_EXT  = "http://ikewiki.srfg.at/syntax/1.0/ext";
	
	private static Log log = LogFactory.getLog(WikipediaXMLRenderEngineBase.class);
	
	
	/** HTML tags that are recognized as valid by the parser */
	protected static String[] ALLOWED_HTML = {
		"b", "big", "blockquote", "br", "caption", "center", "cite", "code", 
		"dd", "div", "dl", "dt", "em", "font", "h1", "h2", "h3", "h4", "h5", "h6", 
		"hr", "i", "li", "ol", "p", "pre", "s", "small", "strike", "strong", "sub",
		"sup", "table", "td", "th", "tr", "tt", "u", "ul", "var"
	};
	
	/** HTML tags that may be empty/without closing tag */
	protected static String[] ALLOWED_EMPTY_HTML = {
		"br", "hr", "p"
	};
	
	// match HTML opening tags; tag name in first group, attributes in second group
	protected static Pattern p_html_open_tag  = Pattern.compile("<(\\p{Alpha}[\\p{Alnum}-_:]*)(\\p{Space}+([^/>]*))?/?>");
	
	// match HTML closing tags; tag name in first group
	protected static Pattern p_html_close_tag = Pattern.compile("</(\\p{Alpha}[\\p{Alnum}-_:]*)>");
	
	// match HTML attributes of the form key = "value" or key = value; key is group 1, value is group 2 or 3
	protected static Pattern p_html_attribute = Pattern.compile("([\\p{Alpha}-_:][\\p{Alnum}-_:]*)(?:\\p{Space}*=\\p{Space}*(?:\"([^\"]+)\"|([^\\p{Space}]+)))?");
	
	
	protected Main main;
	protected WikiResource context;
	
	protected List<WikiLink> links;
	
	// a document builder used for constructing new DOM documents
	protected DocumentBuilder builder;
	
	// the DOM document used in the current parsing process 
	protected Document doc;
	
	// hash tables for quickly looking up allowed html tags
	protected HashSet<String> allowed_html       = new HashSet<String>(Arrays.asList(ALLOWED_HTML));
	protected HashSet<String> allowed_empty_html = new HashSet<String>(Arrays.asList(ALLOWED_EMPTY_HTML));
	
	
	/**
	 * 
	 */
	public WikipediaXMLRenderEngineBase() {
		super();
	}
	
	
	protected void init(Main main, Article context) throws ParserConfigurationException {
		this.main = main;
		this.links = new LinkedList<WikiLink>();
		this.context = context;
		
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance(); 
		factory.setNamespaceAware(true);
		
		
		builder = factory.newDocumentBuilder();
		
	}
	
	
	public Document createDocument() {
		Document d = builder.newDocument();
		if(context != null)
			d.setDocumentURI(context.getUri());
		
		Element root = d.createElementNS(NS_IKEWIKI_CORE,"page");
		d.appendChild(root);
		
		return d;
	}
	
	
	public Document getDocument() {
		return doc;
	}
	
	protected Element parseXML(String content) throws ParseException {
		try {
			ByteArrayInputStream is = new ByteArrayInputStream(content.getBytes());
			Document form = builder.parse(is);
			return (Element)doc.importNode(form.getDocumentElement(),true);			
		} catch(Exception ex) {
			log.error("error while parsing XML content",ex);
			throw new ParseException("could not parse XML content; maybe it is not well-formed?");
		}
	}
	
	protected static void printDocument(Document d, OutputStream out) {
		try {
			XMLWriter writer = new StandardXMLWriter(d);

			writer.writeToStream(out);
		}
		catch (WikiException e) {
			e.printStackTrace();
		}
	}
	
	protected static String endTrim(String s) {
		int last = s.length() - 1;
		while(last > 0 && Character.isWhitespace(s.charAt(last))) {
			last --;
		}
		return s.substring(0,last+1);
	}
	
	protected Collection<Attr> parseAttributes(String attributes) {
		if(attributes != null && attributes.trim().length() > 0) {
			HashSet<Attr> result = new HashSet<Attr>();
			
			Matcher m = p_html_attribute.matcher(attributes);
			
			while(m.find()) {
				String key = m.group(1);
				String value = m.group(2) == null ? m.group(3) : m.group(2);
				
				Attr att = doc.createAttribute(key);
				if(value != null) {
					att.setValue(value);
				} else {
					att.setValue(key);
				}
				
				result.add(att);
				
				log.debug("Attribute: "+key+" = \""+value+"\"");
			}
			return result;
		} else {
			return Collections.emptySet();
		}
	}
	
	protected int countOccurrences(String s, char c) {
		int count = 0;
		for(int i=0;i<s.length();i++) {
			if(s.charAt(i) == c) 
				count++;
		}
		return count;
	}
}
