/**
 *  Copyright (C) 2005
 *  Sebastian Schaffert and Rupert Westenthaler
 *  Salzburg Research Forschungsgesellschaft m.b.H.
 *
 *  IkeWiki is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU Lesser General Public
 * 	License as published by the Free Software Foundation; either
 * 	version 2 of the License, or (at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * 	Lesser General Public License for more details.
 *
 * 	You should have received a copy of the GNU Lesser General Public
 * 	License along with this library; if not, write to the
 * 	Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * 	Boston, MA 02111-1307, USA.
 * 
 */
package at.srfg.ikewiki.render.wiklet;

import java.util.Collection;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import at.srfg.ikewiki.Main;
import at.srfg.ikewiki.data.Namespace;
import at.srfg.ikewiki.data.WikiResource;
import at.srfg.ikewiki.db.DBException;

/**
 * The OWL Class Wiklet adds functionality to the rendering of OWL and RDFS classes.
 * It displays as a view at the bottom of the page and currently displays the properties
 * that are applicable to classes of this type
 * 
 * @author sschaffe
 *
 */
public class OWLClassWiklet extends AbstractWiklet implements Wiklet {

	private static Log log = LogFactory.getLog(OWLClassWiklet.class);

	/**
	 * @param main
	 */
	public OWLClassWiklet(Main main) {
		super(main);
	}

	/* (non-Javadoc)
	 * @see at.srfg.ikewiki.render.wiklet.Wiklet#apply(org.w3c.dom.Document, at.srfg.ikewiki.data.WikiResource)
	 */
	public Document apply(Document d, WikiResource context) {
		try {
			// if the resource is a class, add a iw:view element to the document and add information 
			// about relevant properties and sub-/superclasses
			if(main.getStorageBackend().isClass(context)) {
				Element view = d.createElementNS(Namespace.NS_IKEWIKI_EXT,"view");
				view.setAttribute("position","after");
				view.setAttribute("wiklet","OWLClassWiklet");
				d.getDocumentElement().appendChild(view);

				// superclasses
				for(WikiResource cls : main.getStorageBackend().getSuperClasses(context,true)) {
					Element sc_elem = d.createElementNS(Namespace.NS_IKEWIKI_WIKLET_CV,"superclass");
					sc_elem.setAttribute("name",cls.getQualifiedTitle());
					sc_elem.setAttribute("uri",cls.getUri());
					
					view.appendChild(sc_elem);
				}
				
				// subclasses
				for(WikiResource cls : main.getStorageBackend().getSubClasses(context,true)) {
					Element sc_elem = d.createElementNS(Namespace.NS_IKEWIKI_WIKLET_CV,"subclass");
					sc_elem.setAttribute("name",cls.getQualifiedTitle());
					sc_elem.setAttribute("uri",cls.getUri());
					
					view.appendChild(sc_elem);
				}

				// outgoing properties
				for(WikiResource prop : main.getStorageBackend().getPropertiesByDomain(context,false)) {
					Element sc_elem = d.createElementNS(Namespace.NS_IKEWIKI_WIKLET_CV,"property");
					sc_elem.setAttribute("type","outgoing");
					sc_elem.setAttribute("name",prop.getQualifiedTitle());
					sc_elem.setAttribute("uri",prop.getUri());

					Collection<WikiResource> direct_types = main.getStorageBackend().getDirectTypes(prop);
					// list the direct types of this property
					for(WikiResource type : main.getStorageBackend().getTypes(prop)) {
						if(direct_types.contains(type)) {
							Element ptype = d.createElementNS(Namespace.NS_IKEWIKI_EXT,"type");
							ptype.appendChild(d.createTextNode(type.getQualifiedTitle()));
							ptype.setAttribute("direct","true");
							sc_elem.appendChild(ptype);
						} else {
							Element ptype = d.createElementNS(Namespace.NS_IKEWIKI_EXT,"type");
							ptype.appendChild(d.createTextNode(type.getQualifiedTitle()));
							ptype.setAttribute("direct","false");
							sc_elem.appendChild(ptype);
						}						
					}
					
					// list the range of this property
					for(WikiResource range : main.getStorageBackend().getRange(prop)) {
						Element prange = d.createElementNS(Namespace.NS_IKEWIKI_WIKLET_CV,"range");
						prange.appendChild(d.createTextNode(range.getQualifiedTitle()));
						
						sc_elem.appendChild(prange);						
					}
					
					view.appendChild(sc_elem);					
				}
				
				// incoming properties
				for(WikiResource prop : main.getStorageBackend().getPropertiesByRange(context,false)) {
					Element sc_elem = d.createElementNS(Namespace.NS_IKEWIKI_WIKLET_CV,"property");
					sc_elem.setAttribute("type","incoming");
					sc_elem.setAttribute("name",prop.getQualifiedTitle());
					sc_elem.setAttribute("uri",prop.getUri());

					Collection<WikiResource> direct_types = main.getStorageBackend().getDirectTypes(prop);
					// list the direct types of this property
					for(WikiResource type : main.getStorageBackend().getTypes(prop)) {
						if(direct_types.contains(type)) {
							Element ptype = d.createElementNS(Namespace.NS_IKEWIKI_EXT,"type");
							ptype.appendChild(d.createTextNode(type.getQualifiedTitle()));
							ptype.setAttribute("direct","true");
							sc_elem.appendChild(ptype);
						} else {
							Element ptype = d.createElementNS(Namespace.NS_IKEWIKI_EXT,"type");
							ptype.appendChild(d.createTextNode(type.getQualifiedTitle()));
							ptype.setAttribute("direct","false");
							sc_elem.appendChild(ptype);
						}						
					}
					
					// list the range of this property
					for(WikiResource domain : main.getStorageBackend().getDomain(prop)) {
						Element prange = d.createElementNS(Namespace.NS_IKEWIKI_WIKLET_CV,"domain");
						prange.appendChild(d.createTextNode(domain.getQualifiedTitle()));
						
						sc_elem.appendChild(prange);						
					}
					
					view.appendChild(sc_elem);					
				}
				
			}
		} catch(DBException ex) {
			log.error(ex);
		}
		return d;
	}

}
