/**
 * 
 */
package at.srfg.semweb.reasoner;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.util.Collection;

import jpl.Atom;
import jpl.JPL;
import jpl.Query;
import jpl.Term;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import at.srfg.semweb.reasoner.data.PrologNode;
import at.srfg.semweb.reasoner.data.PrologResource;
import at.srfg.semweb.reasoner.data.PrologTriple;

/**
 * Interface to the SWI Prolog engine and Prolog-based OWL Lite reasoner.
 * 
 * In order to use this class on Unix, you need to preload the libjlp.so of
 * SWI Prolog, otherwise there will be symbol lookup errors because the Unix Java does not
 * make symbols of a native library available to other libraries. The following environment variables 
 * need to be set:
 * 
 *  LD_PRELOAD=/path/to/libjlp.so
 *  LD_LIBRARY_PATH=$JAVA_HOME/jre/lib/i386:$JAVA_HOME/jre/lib/i386/client/:$PLBASE/lib/$ARCH
 *  
 * LD_LIBRARY path needs to be set because otherwise, java is no longer able to find its own
 * libraries. 
 * 
 *  alternatively, we can start with 
 *  -Djava.library.path=$PLBASE/lib/$ARCH:$JAVA_HOME/jre/lib/i386:$JAVA_HOME/jre/lib/i386/client
 *  
 * 
 * @author sschaffe
 *
 */
public class PrologEngine {

	private static Log log = LogFactory.getLog(PrologEngine.class);

	/**
	 * Path to the Prolog jpl library; needs to be passed to the VM as argument for the moment.
	 */
	private String prologPath;
	
	
	/**
	 * 
	 */
	public PrologEngine(String prologPath) {
		super();
		this.prologPath = prologPath;
		
		initProlog();
	}

	
	private void initProlog() {
		
		//System.setProperty("java.library.path",this.prologPath);

		JPL.init();		
		log.info("SWI Prolog initialised");
		
		consult("reasoner.pl");
		log.info("reasoner initialised");
	}

	/**
	 * Consult a resource relative to the PrologEngine class. Makes use of the
	 * getResourceAsStream function and creates a temporary file that is then loaded into 
	 * SWI Prolog 
	 * @param name
	 */
	public void consult(String name) {
		InputStream is = this.getClass().getResourceAsStream(name);
		try {
			consult(is);
		} catch(IOException ex) {
			log.error("consultation of resource "+name+" failed",ex);
		}		
	}
	
	/**
	 * Consult a resource given as input stream. Creates a temporary file that is
	 * then loaded into SWI Prolog.
	 * 
	 * @param is the input stream from where to read the Prolog source code
	 * @throws IOException
	 */
	public void consult(InputStream is) throws IOException {
		File tmp = File.createTempFile("at.srfg.semweb.reasoner",".pl");
		
		BufferedReader in = new BufferedReader(new InputStreamReader(is));
		PrintWriter out = new PrintWriter(new FileWriter(tmp));
		
		String line = in.readLine();
		while(line != null) {
			out.println(line);
			line = in.readLine();
		}
		out.close();
		in.close();
		
		consult(tmp);
		
		tmp.delete();
	}
	
	/**
	 * Consult the contents of the given file and load them into SWI Prolog.
	 * 
	 * @param file
	 */
	public void consult(File file) {
	    Query q1 =
	        new Query(
	            "consult",
	            new Term[] {new Atom(file.getAbsolutePath())}
	        );
	    boolean result = q1.hasSolution();
	    
	    log.info("consultation of Prolog file "+file.getName() + (result?" succeeded":" failed"));
	}

	
	
	
	/**
	 * Load an ontology or RDF file in RDF/XML syntax relative to the PrologEngine class. Makes use of the
	 * getResourceAsStream function and creates a temporary file that is then loaded into 
	 * SWI Prolog 
	 * @param name
	 */
	public void loadOntology(String name) {
		loadOntology(name,this.getClass());
	}
	

	/**
	 * Load an ontology or RDF file in RDF/XML syntax relative to the provided class. Makes use of the
	 * getResourceAsStream function and creates a temporary file that is then loaded into 
	 * SWI Prolog 
	 * @param name
	 */
	public void loadOntology(String name, Class ref_class) {
		InputStream is = ref_class.getResourceAsStream(name);
		try {
			loadOntology(is);
		} catch(IOException ex) {
			log.error("loading of ontology "+name+" failed",ex);
		}		
	}
	
	/**
	 * Load an ontology
	 * @param is
	 * @throws IOException
	 */
	public void loadOntology(InputStream is) throws IOException {
		File tmp = File.createTempFile("at.srfg.semweb.reasoner",".rdf");
		
		BufferedReader in = new BufferedReader(new InputStreamReader(is));
		PrintWriter out = new PrintWriter(new FileWriter(tmp));
		
		String line = in.readLine();
		while(line != null) {
			out.println(line);
			line = in.readLine();
		}
		out.close();
		in.close();
		
		loadOntology(tmp);
		
		tmp.delete();
	}

	public void loadOntology(File file) {
		Query q1 =
			new Query(
					"rdf_load",
					new Term[] {new Atom(file.getAbsolutePath())}
			);
		
		boolean result = q1.hasSolution();
		
		log.info("loading of ontology "+file.getName()+(result?" succeeded":" failed"));
	}
	
	
	public void loadDOLCE() {
		String base = "/home/wastl/tmp/DLP397/";
		
		String[] files = { 
				"Common_397.owl", "DLP_397.owl", "DOLCE-Lite_397.owl", "ExtDnS_397.owl",
				"FunctPart_397.owl", "Information_397.owl", "Modal_397.owl", "Plans_397.owl",
				"Social_397.owl", "SpatialRel_397.owl", "TemporalRel_397.owl"
		};
		
		for(String f : files) {
			File file = new File(base + f);
			loadOntology(file);
		}
	}
	

	public Collection<PrologTriple> listTriples(PrologResource subject, PrologResource predicate, PrologNode object) {
		
		return null;
	}
	
	
	/* (non-Javadoc)
	 * @see java.lang.Object#finalize()
	 */
	@Override
	protected void finalize() throws Throwable {
		JPL.halt();
		
		super.finalize();
	}


	public static void main(String[] args) {
		// set log level for the whole application
		Logger main_logger = Logger.getLogger("at.srfg.semweb.reasoner");
		main_logger.setLevel(Level.DEBUG);

		PrologEngine engine = new PrologEngine("/opt/sschaffe/apps/lib/pl-5.6.8/lib/i686-linux/libjpl.so");
		try {
			engine.loadOntology("rdf-schema.rdf",Class.forName("at.srfg.ikewiki.installer.InitOntologies"));
			engine.loadOntology("owl.rdfs",Class.forName("at.srfg.ikewiki.installer.InitOntologies"));
			engine.loadDOLCE();
		} catch(Exception ex) {
			ex.printStackTrace();
		}
		
	}
	
}
