% (c)2006 Sebastian Schaffert, Salzburg Research Forschungsgesellschaft

:- use_module(library('semweb/rdf_db')).

% allow namespace expansion in triples
:- rdf_meta
	triple(r,r,o),
	triple(r,r,o,@),
	type(r,r),
	type(r,r,@),
	subClass(r,r),
	subClass(r,r,@),
	subProperty(r,r),
	subProperty(r,r,@),
	equivalentClass(r,r),
	equivalentClass(r,r,@),
	class(r),
	property(r),
	objectProperty(r),
	datatypeProperty(r),
	annotationProperty(r),
	domain(r,r),
	range(r,r),
	rdf_list(r,t),
	rdf_list_member(r,r).



% register relevant namespaces


% the predicate triple/4:
% - arguments 1-3 are subject, predicate, object of the rdf triple
% - the argument 4 indicates whether a triple is directly contained in the store
%   or inferred by a rule

% the predicate triple/3 simply throws away the information about directness
triple(X,Y,Z) :- triple(X,Y,Z,_).



% every rdf statement as a direct triple
triple(X,Y,Z,direct) :- rdf(X,Y,Z).

% types are added as (indirect) triples
triple(X,'http://www.w3.org/1999/02/22-rdf-syntax-ns#type',Z,indirect) :-
	type(X,Z,indirect),
	\+ rdf(X,rdf:'type',Z) . % avoid redundancies


% subclasses are added as (indirect) triples
triple(X,'http://www.w3.org/2000/01/rdf-schema#subClassOf',Y,indirect) :-
	subClass(X,Y,indirect),
	\+ rdf(X,rdfs:'subClassOf',Y) . % avoid redundancies



% equivalences are added as (indirect) triples
triple(X,'http://www.w3.org/2002/07/owl#equivalentClass',Y,indirect) :-
	equivalentClass(X,Y,indirect),
	\+ rdf(X,owl:'equivalentClass',Y) . % avoid redundancies
	
% subproperties are added as (indirect) triples
triple(X,'http://www.w3.org/2000/01/rdf-schema#subPropertyOf',Y,indirect) :-
	subProperty(X,Y,indirect), \+ rdf(X,rdfs:'subPropertyOf',Y) .


% equivalent properties are added as (indirect) triples
triple(X,'http://www.w3.org/2002/07/owl#equivalentProperty',Y,indirect) :-
	equivalentProperty(X,Y,indirect), \+ rdf(X,owl:'equivalentProperty',Y) .



% TransitiveProperty
triple(S,P,O,indirect) :-
	transitiveProperty(S,P,O,indirect), \+ rdf(S,P,O) . % avoid duplicates

% SymmetricProperty
triple(S,P,O,indirect) :-
	symmetricProperty(S,P,O,indirect), \+ rdf(S,P,O) . % avoid duplicates


% inverse
triple(P,'http://www.w3.org/2002/07/owl#inverseOf',Q,indirect) :-
	inverse(P,Q,indirect), \+ rdf(P,owl:inverseOf,Q).

%disjointWith
triple(X,'http://www.w3.org/2002/07/owl#disjointWith',Y,indirect) :-
	disjointWith(X,Y,indirect), \+ rdf(X,owl:disjointWith,Y).


% sameAs
triple(X,'http://www.w3.org/2002/07/owl#sameAs',Y,D) :-
	sameAs(X,Y,D), \+ rdf(X,owl:sameAs,Y).




%%%%%%%%%%%%%%%%%%%%
% consequences


% sameAs: if S is the same as S1, then S has all the relations of S1
triple(S,P,O,indirect) :-
	sameAs(S,S1),
	triple(S1,P,O),
	not( rdf(S,P,O) ).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% type

type(X,Y) :- type(X,Y,_).
	
% direct type definition
type(X,Z,direct) :-
	rdf(X,rdf:'type',Z).


% every resource has type rdfs:resource
type(X,'http://www.w3.org/2000/01/rdf-schema#Resource',indirect) :-
	isResource(X).

% indirect type definition via subclass
type(X,Z,indirect) :-
	rdf(X,rdf:'type',Y),
	subClass(Y,Z).


% indirect type definition via subproperty
type(X,Z,indirect) :-
	rdf(X,rdf:'type',Y),
	subProperty(Y,Z).


% indirect type definition via domain
type(X,Z,indirect) :-
	rdf(X,P,_), domain(P,Z).

% indirect type definition via range
type(X,Z,indirect) :-
	rdf(_,P,X), isResource(X), range(P,Z).



% for equivalentClass, intersectionOf not needed, covered by subClass

% indirect type definition via intersection
%type(X,Z,indirect) :-
%	rdf(X,rdf:'type',I),
%	rdf(I,owl:'intersectionOf',List),
%	rdf_list_member(List,Z).
	
%type(X,Z,indirect) :-
%	rdf(X,rdf:'type',I),
%	subClass(I,J),
%	rdf(J,owl:'intersectionOf',List),
%	rdf_list_member(List,Z).


% indirect type definition via union (if an individual has a type that
% participates in a union, then it also has the union as type)

% unionOf is expensive at the moment!
/*
type(X,U,indirect) :-
	rdf(X,rdf:'type',C),
	rdf(U,owl:'unionOf',List),
	rdf_list_member(List,C).

type(X,U,indirect) :-
	rdf(X,rdf:'type',C),
	rdf(U,owl:'unionOf',List),
	subClass(C,D),
	rdf_list_member(List,D).
*/

% type restriction by allValuesFrom:
% if a class C restricts a property P to allValuesFrom some class D, I is an
% instance of C, there exists a triple I,P,J, then J has type D
typ(J,D,indirect) :-
	rdf(I,rdf:'type',C),
	rdf(I,P,J),
	allValuesFrom(C,P,D).
	

typ(J,D,indirect) :-
	rdf(I,rdf:'type',C),
	subClass(C,C1),
	rdf(I,P,J),
	allValuesFrom(C1,P,D).

typ(J,D,indirect) :-
	rdf(I,rdf:'type',C),
	rdf(I,P,J),
	subProperty(P,P1),
	allValuesFrom(C,P1,D).

typ(J,D,indirect) :-
	rdf(I,rdf:'type',C),
	subClass(C,C1),
	rdf(I,P,J),
	subProperty(P,P1),
	allValuesFrom(C1,P1,D).


typ(J,D,indirect) :-
	rdf(I,rdf:'type',C),
	transitiveProperty(I,P,J),
	allValuesFrom(C,P,D).

typ(J,D,indirect) :-
	rdf(I,rdf:'type',C),
	subClass(C,C1),
	transitiveProperty(I,P,J),
	allValuesFrom(C1,P,D).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% subClassOf

subClass(X,Y) :- subClass(X,Y,_).

subClass(X,Y,direct) :-
	rdf(X,rdfs:'subClassOf',Y).


subClass(X,Z,indirect) :-
	rdf(X,rdfs:'subClassOf',Y),
	Z \= X,
	subClass(Y,Z).

subClass(X,Y,indirect) :-
	rdf(X,owl:'equivalentClass',Y).

subClass(X,Y,indirect) :-
	rdf(Y,owl:'equivalentClass',X).



% subclass via intersection
subClass(X,Z,indirect) :-
	rdf(X,owl:'intersectionOf',List),
	rdf_list_member(List,Z).
	



/* adding owl:Thing and owl:Nothing here is a serious performance drawback
   and probably does not add much information, we leave it out for the moment */
/* 
% owl:Thing is the superclass of all classes
subClass(X,'http://www.w3.org/2002/07/owl#Thing',indirect) :-
	class(X).

% owl:Nothing is a subclass of all classes
subClass('http://www.w3.org/2002/07/owl#Nothing',X,indirect) :-
	class(X).
*/

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% equivalentClass

equivalentClass(X,Y) :- equivalentClass(X,Y,_).

% direct
equivalentClass(X,Y,direct) :-
	rdf(X,owl:'equivalentClass',Y).

equivalentClass(X,Y,indirect) :-
	rdf(Y,owl:'equivalentClass',X).

% transitive
equivalentClass(X,Z,indirect) :-
	rdf(X,owl:'equivalentClass',Y),
	Z \= X,
	equivalentClass(Y,Z).

/*
% by subclassing
equivalentClass(X,Y,indirect) :-
	subClass(X,Y), subClass(Y,X).
*/

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% subPropertyOf

subProperty(X,Y) :- subProperty(X,Y,_).


subProperty(X,Y,direct) :-
	rdf(X,rdfs:'subPropertyOf',Y).

% subPropertyOf is transitive
subProperty(X,Z,indirect) :-
	rdf(X,rdfs:'subPropertyOf',Y),
	Z \= X,
	subProperty(Y,Z).
	
% equivalent properties are also subproperties
subProperty(X,Z,indirect) :-
	rdf(X,owl:'equivalentProperty',Z).

subProperty(X,Z,indirect) :-
	rdf(Z,owl:'equivalentProperty',X).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% equivalentProperty

equivalentProperty(P,Q) :- equivalentProperty(P,Q,_).

equivalentProperty(P,Q,direct) :-
	rdf(P,owl:'equivalentProperty',Q).
	
equivalentProperty(P,Q,indirect) :-
	rdf(Q,owl:'equivalentProperty',P).


% transitive
equivalentProperty(P,R,indirect) :-
	rdf(P,owl:'equivalentProperty',Q),
	R \= P,
	equivalentProperty(Q,R).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% domain
domain(P,D) :- domain(P,D,_).

domain(P,D,direct) :-
	rdf(P,rdfs:'domain',D).

% domain restrictions also restrict the domain to the superclasses of
% C
domain(P,D,indirect) :-
	rdf(P,rdfs:'domain',C),
	isResource(C), subClass(C,D).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% range
range(P,D) :- range(P,D,_).

range(P,D,direct) :-
	rdf(P,rdfs:'range',D).

% range restrictions also restrict the range to the superclasses of
% C
range(P,D,indirect) :-
	rdf(P,rdfs:'range',C),
	isResource(C), subClass(C,D).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% TransitiveProperty
transitiveProperty(S,P,O,direct) :-
	type(P,owl:'TransitiveProperty'),
	rdf(S,P,O).


transitiveProperty(S,P,O,indirect) :-
	type(P,owl:'TransitiveProperty'),
	rdf(S,P,Z),
	transitiveProperty(Z,P,O).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SymmetricProperty

symmetricProperty(S,P,O,direct) :-
	type(P,owl:'SymmetricProperty'),
	rdf(S,P,O).

symmetricProperty(S,P,O,indirect) :-
	type(P,owl:'SymmetricProperty'),
	rdf(O,P,S).

	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% inverseOf

inverse(P,Q) :- inverse(P,Q,_).

inverse(P,Q,direct) :-
	rdf(P,owl:'inverseOf',Q).

inverse(P,Q,indirect) :-
	rdf(Q,owl:'inverseOf',P).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% disjointWith

disjointWith(X,Y) :- disjointWith(X,Y,_).

% direct
disjointWith(X,Y,direct) :-
	rdf(X,owl:'disjointWith',Y).

disjointWith(X,Y,indirect) :-
	rdf(Y,owl:'disjointWith',X).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% sameAs

sameAs(X,Y) :- sameAs(X,Y,_).

sameAs(X,Y,direct) :-
	rdf(X,owl:'sameAs',Y).

sameAs(X,Y,direct) :-
	rdf(X,owl:'sameIndividualAs',Y).


sameAs(X,Y,indirect) :-
	rdf(Y,owl:'sameAs',X).

sameAs(X,Y,indirect) :-
	rdf(Y,owl:'sameIndividualAs',X).

sameAs(X,Z,indirect) :-
	rdf(X,owl:'sameAs',Y),
	X \= Z,
	sameAs(Y,Z).

sameAs(X,Z,indirect) :-
	rdf(X,owl:'sameIndividualAs',Y),
	X \= Z,
	sameAs(Y,Z).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% allValuesFrom
allValuesFrom(C,P,D) :- allValuesFrom(C,P,D,_).

allValuesFrom(C,P,D,direct) :-
	rdf(C,owl:'onProperty',P),
	rdf(C,owl:'allValuesFrom',D).


/*

triple(S,P,O,indirect) :-
	rdf(S1,owl:'sameAs',S),
	triple(S1,P,O).
*/

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% convenienve predicates

class(X) :-
	type(X,owl:'Class').

class(X) :-
	type(X,rdfs:'Class').



property(X) :-
	type(X,rdf:'Property').

objectProperty(X) :-
	type(X,owl:'ObjectProperty').

annotationProperty(X) :-
	type(X,owl:'AnnotationProperty').

datatypeProperty(X) :-
	type(X,owl:'DatatypeProperty').



rdf_list(R,[H|T]) :-
	type(R,rdf:'List'),
	rdf(R,rdf:'first',H),
	rdf(R,rdf:'rest',R2),
	rdf_list(R2,T).

rdf_list(R,[H]) :-
	type(R,rdf:'List'),
	rdf(R,rdf:'first',H),
	rdf(R,rdf:'rest',rdf:'nil').


rdf_list(rdf:'nil',[]).

% check whether E is a member of the RDF list L
rdf_list_member(L,E) :-
	type(L,rdf:'List'),
	rdf(L,rdf:'first',E).
	
rdf_list_member(L,E) :-
	type(L,rdf:'List'),
	rdf(L,rdf:'rest',L2),
	rdf_list_member(L2,E).

rdf_list_member(rdf:'nil',_) :- fail.



% the following predicates cannot be used for enumerating, they are just checks
isResource(X) :- X \= literal(_).

isAnonResource(X) :- atom_prefix(X,'__').

isUriResource(X) :- isResource(X), \+ isAnonResource(X).

isLiteral(literal(_)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% integrity constraints

% if X and Y are disjoint classes, and R is an instance of X, then R is not an instance of Y
check_disjoint :-
	triple(X,owl:'disjointWith',Y),
	triple(R,rdf:'type',X),
	not( triple(R,rdf:'type',Y) ).

% if an individual T is of type C, and there exists a property P with domain C and some range D,
% and C has a minCardinality restriction of n on P, then there need to exist at least n
% different individuals of type D that are in relation P with T
check_cardinality :-
	triple(T,rdf:'type',C),
	property(P),
	domain(P,C),
	range(P,D),
	triple(C,owl:'minCardinality',N). % TODO ...


check_different.



owl :- rdf_load('/home/wastl/work/eclipse/IkeWiki/WEB-INF/src/at/srfg/ikewiki/installer/owl.rdfs').

dolce :-
	owl,
	rdf_load('/home/wastl/tmp/DLP397/DLP_397.owl'),
	rdf_load('/home/wastl/tmp/DLP397/Common_397.owl'),
	rdf_load('/home/wastl/tmp/DLP397/DOLCE-Lite_397.owl'),
	rdf_load('/home/wastl/tmp/DLP397/ExtDnS_397.owl'),
	rdf_load('/home/wastl/tmp/DLP397/FunctPart_397.owl'),
	rdf_load('/home/wastl/tmp/DLP397/Information_397.owl'),
	rdf_load('/home/wastl/tmp/DLP397/Modal_397.owl'),
	rdf_load('/home/wastl/tmp/DLP397/Plans_397.owl'),
	rdf_load('/home/wastl/tmp/DLP397/Social_397.owl'),
	rdf_load('/home/wastl/tmp/DLP397/SpatialRel_397.owl'),
	rdf_load('/home/wastl/tmp/DLP397/TemporalRel_397.owl').


sample :-
	rdf_load('/home/wastl/tmp/sample.owl').

test1(L) :- setof(f(S,P,O),triple(S,P,O),L).

test2(L) :- setof(f(S,P,O),triple(S,P,O,indirect),L).


% Local Variables:
% mode: prolog
% End
	