/**
 *  Copyright (C) 2005
 *  Sebastian Schaffert and Rupert Westenthaler
 *  Salzburg Research Forschungsgesellschaft m.b.H.
 *
 *  IkeWiki is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU Lesser General Public
 * 	License as published by the Free Software Foundation; either
 * 	version 2 of the License, or (at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * 	Lesser General Public License for more details.
 *
 * 	You should have received a copy of the GNU Lesser General Public
 * 	License along with this library; if not, write to the
 * 	Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * 	Boston, MA 02111-1307, USA.
 * 
 */


/* necessary:
 * createlang plpgsql ontowiki
 */

/* a function to update the modification and user fields of pages on change */
CREATE FUNCTION f_set_modification () RETURNS trigger AS '
BEGIN
	NEW.modification    := ''now'';
	NEW.modification_by := current_user;
	RETURN NEW;
END;
' LANGUAGE plpgsql;


CREATE SEQUENCE user_ids;

CREATE TABLE users (
	id  		INT NOT NULL PRIMARY KEY DEFAULT nextval('user_ids'),
	login	 	CHAR(80) NOT NULL,
	password 	CHAR(80) NOT NULL, /* md5 hash */
	first_name 	CHAR(80) NOT NULL,
	last_name   CHAR(80) NOT NULL,
	email 		CHAR(255) NOT NULL,
	first_login TIMESTAMP NOT NULL DEFAULT current_timestamp,
	last_login 	TIMESTAMP NOT NULL DEFAULT current_timestamp,
	login_count INT NOT NULL DEFAULT 0
);	

CREATE UNIQUE INDEX idx_users_login ON users(login);


CREATE SEQUENCE role_ids;

CREATE TABLE roles (
	id  		INT NOT NULL PRIMARY KEY DEFAULT nextval('role_ids'),
	name 		CHAR(80) NOT NULL,
	description TEXT,	
);	


CREATE TABLE user_roles (
	user_id 	INT NOT NULL REFERENCES users(id) ON DELETE CASCADE,
	role_id 	INT NOT NULL REFERENCES roles(id) ON DELETE CASCADE
);	

CREATE INDEX idx_user_roles_user ON user_roles(user_id);


/**
 * A table to store URIs and IDs; referenced by the triples_internal table.
 */
CREATE SEQUENCE uri_ids;

CREATE TABLE uris (
	id		 	INT NOT NULL PRIMARY KEY DEFAULT nextval('uri_ids'),
	uri 		CHAR(255) NOT NULL
);	

CREATE UNIQUE INDEX idx_uris_uri ON uris(uri);


CREATE SEQUENCE namespace_ids;

CREATE TABLE namespaces (
	id 				INT NOT NULL DEFAULT nextval('namespace_ids'),
	uri 			CHAR(255) NOT NULL,
	prefix 			CHAR(255),
	PRIMARY KEY(id)
);
CREATE UNIQUE INDEX idx_namespaces_uri ON namespaces(uri);
CREATE UNIQUE INDEX idx_namespaces_prefix ON namespaces(prefix);	

INSERT INTO namespaces (uri,prefix) VALUES ('http://ikewiki.srfg.at/Internal#','');
INSERT INTO namespaces (uri,prefix) VALUES ('http://ikewiki.srfg.at/Special#','Special');
INSERT INTO namespaces (uri,prefix) VALUES ('http://ikewiki.srfg.at/Help#','Help');
INSERT INTO namespaces (uri,prefix) VALUES ('http://ikewiki.srfg.at/User#','User');



/**
 * The table pages contains the "static" meta-information about a Wiki page.
 * 
 * As the text content of the page is subject to change, it is not stored here, but
 * instead via an intermediate table "revision".
 *
 * The uri column is computed automatically and should never be set explicitly.
 */
CREATE FUNCTION f_set_page_uri () RETURNS trigger AS '
DECLARE
    namespace_rec RECORD;
    uri_rec RECORD;
    full_uri varchar;
BEGIN
	/* need to check for URI and insert reference instead! */

	SELECT INTO namespace_rec uri,prefix FROM namespaces WHERE id=NEW.namespace;

	full_uri    := namespace_rec.uri || NEW.title ;

	SELECT INTO uri_rec * FROM uris WHERE uris.uri = full_uri;
	IF NOT FOUND THEN
		INSERT INTO uris (uri) VALUES (full_uri);
		NEW.uri := currval(''uri_ids'');
	ELSE
		NEW.uri := uri_rec.id;
	END IF;

	RETURN NEW;
END;
' LANGUAGE plpgsql;



CREATE SEQUENCE page_ids;

CREATE TABLE pages (
	id				INT NOT NULL DEFAULT nextval('page_ids'),
	namespace		INT REFERENCES namespaces(id) ON DELETE SET NULL,
	title			CHAR(255) NOT NULL,
	uri				INT REFERENCES uris(id) ON DELETE SET NULL,
	cur_revision	INT,
	creator         INT REFERENCES users(id) ON DELETE SET NULL,
	creation		TIMESTAMP NOT NULL DEFAULT current_timestamp,
	modification		TIMESTAMP NOT NULL DEFAULT current_timestamp,
	modification_by	CHAR(80) DEFAULT current_user,
	PRIMARY KEY(id)
);

CREATE UNIQUE INDEX idx_pages_namespace_title ON pages(namespace,title);
CREATE UNIQUE INDEX idx_pages_uri ON pages(uri);

CREATE TRIGGER tr_pages_modification BEFORE INSERT OR UPDATE ON pages
		FOR EACH ROW EXECUTE PROCEDURE f_set_modification();
	
CREATE TRIGGER tr_pages_uri BEFORE INSERT OR UPDATE ON pages
		FOR EACH ROW EXECUTE PROCEDURE f_set_page_uri();


/**
 * Store the text contents
 */
CREATE SEQUENCE text_ids; 
 
CREATE TABLE texts (
	id				INT NOT NULL DEFAULT nextval('text_ids'),
	language		CHAR(12),
	text			TEXT,
	text_xml			TEXT,
	PRIMARY KEY(id)
);	

		
/*
 * Store the revision meta-information about pages.
 *
 * references a page description and a text; if updated, the cur_revision link in the page
 * is also updated.
 */	
CREATE SEQUENCE revision_ids;

CREATE TABLE revisions (
	id				INT NOT NULL DEFAULT nextval('revision_ids'),
	page			INT NOT NULL REFERENCES pages(id) ON DELETE CASCADE,
	text			INT NOT NULL REFERENCES texts(id) ON DELETE CASCADE,
	creator         INT REFERENCES users(id) ON DELETE SET NULL,
	comment			TEXT,	
	deleted			BOOLEAN DEFAULT FALSE,
	timestamp		TIMESTAMP NOT NULL DEFAULT current_timestamp,
	creation		TIMESTAMP NOT NULL DEFAULT current_timestamp,
	modification		TIMESTAMP NOT NULL DEFAULT current_timestamp,
	modification_by	CHAR(80) DEFAULT current_user,
	PRIMARY KEY(id)
);
	
CREATE INDEX idx_revisions_page ON revisions(page);
CREATE INDEX idx_revisions_timestamp ON revisions(timestamp);

/*
 * Update the revision pointer in the pages table when a new revision is inserted.
 */
CREATE FUNCTION f_update_revisions() RETURNS trigger AS '
BEGIN
	UPDATE pages SET cur_revision=NEW.id WHERE id=NEW.page;
	RETURN NEW;
END;
' LANGUAGE plpgsql;
		
CREATE TRIGGER	tr_update_revisions AFTER INSERT OR UPDATE ON revisions
		FOR EACH ROW EXECUTE PROCEDURE f_update_revisions();
		

CREATE VIEW current_page AS (
	SELECT P.id AS page_id, R.id AS revision_id, N.prefix AS ns_prefix, N.uri AS ns_uri, (SELECT login FROM users WHERE users.id = R.creator) AS creator, title, (N.uri || title) AS uri, R.timestamp, R.comment, T.language, T.text
	FROM pages P, revisions R, texts T, namespaces N
	WHERE R.deleted = FALSE
	  AND R.page = P.id
	  AND R.text = T.id
	  AND P.cur_revision = R.id
	  AND P.namespace = N.id
);


CREATE VIEW page_revisions AS (
	SELECT P.id AS page_id, R.id AS revision_id, N.prefix AS ns_prefix, N.uri AS ns_uri, (SELECT login FROM users WHERE users.id = R.creator) AS creator, title, (N.uri || title) AS uri, R.timestamp, R.comment, T.language, T.text
	FROM pages P, revisions R, texts T, namespaces N
	WHERE R.deleted = FALSE
	  AND R.page = P.id
	  AND R.text = T.id
	  AND P.namespace = N.id
);


CREATE FUNCTION f_rank_page(pattern varchar,title varchar,text varchar) RETURNS float(24) AS '
DECLARE
	pattern 	ALIAS FOR $1;
	title	 	ALIAS FOR $2;
	text	 	ALIAS FOR $3;
	title_pos	integer;
	text_pos	integer;
	rank		float(24);
BEGIN
	title_pos := position(pattern in title);
	text_pos  := position(pattern in text);
	rank := 0.0;
	
	IF title_pos > 0 THEN
		rank := rank + 1000 * 1/title_pos;
	END IF;
	
	IF text_pos > 0 THEN
		rank := rank + 100 * 1/text_pos;
	END IF;
	RETURN rank;
END;
' LANGUAGE plpgsql;



CREATE SEQUENCE image_ids;

CREATE TABLE images (
	id 				INT NOT NULL DEFAULT nextval('image_ids'),
	namespace 		INT REFERENCES namespaces(id) ON DELETE SET NULL,
	title	 		CHAR(255) NOT NULL,
	uri 			INT REFERENCES uris(id) ON DELETE SET NULL,
	comment 			TEXT,
	mimetype 		CHAR(80) NOT NULL,
	image		 	BYTEA NOT NULL,
	creator      	INT REFERENCES users(id) ON DELETE SET NULL,
	creation 		TIMESTAMP NOT NULL DEFAULT current_timestamp,
	modification	 	TIMESTAMP NOT NULL DEFAULT current_timestamp,
	modification_by CHAR(80) DEFAULT current_user,
	PRIMARY KEY(id)
);

CREATE UNIQUE INDEX idx_images_namespace_title ON images(namespace,title);

CREATE TRIGGER tr_images_modification BEFORE INSERT OR UPDATE ON images
		FOR EACH ROW EXECUTE PROCEDURE f_set_modification();

CREATE TRIGGER tr_images_uri BEFORE INSERT OR UPDATE ON images
		FOR EACH ROW EXECUTE PROCEDURE f_set_page_uri();


CREATE VIEW current_image AS (
	SELECT I.id AS image_id, N.prefix AS ns_prefix, N.uri AS ns_uri, (SELECT login FROM users WHERE users.id = I.creator) AS creator, title, (N.uri || title) AS uri, I.creation AS timestamp, I.mimetype, I.comment, I.image
	FROM images I, namespaces N
	WHERE I.namespace = N.id
);


CREATE SEQUENCE variable_ids;

CREATE TABLE variables (
	id 				INT NOT NULL DEFAULT nextval('variable_ids'),
	name 			CHAR(80) NOT NULL,
	value 			TEXT DEFAULT ''
);

CREATE INDEX idx_variables_name ON variables(name);




CREATE SEQUENCE metadata_journal_ids;
CREATE SEQUENCE metadata_journal_transaction_ids;

/* add subject/predicate/object types? */
CREATE TABLE metadata_journal (
	id 				INT NOT NULL DEFAULT nextval('metadata_journal_ids'),
	transaction 	INT NOT NULL,
	timestamp		TIMESTAMP NOT NULL DEFAULT current_timestamp,
	user        	INT REFERENCES users(id) ON DELETE SET NULL,
	action	 		CHAR(1),
	persistent	 	BOOLEAN NOT NULL,
	subject	 		CHAR(255), /* value is interpreted depending on type */
	predicate 		CHAR(255), /* value is interpreted depending on type */
	object	 		CHAR(255), /* value is interpreted depending on type */
	object_overflow	TEXT,
	PRIMARY KEY(id)
);

/* these are needed to determine the history for a resource; therefore we index over the field and its type */
CREATE INDEX metadata_journal_subject_idx ON metadata_journal(subject);
CREATE INDEX metadata_journal_object_idx ON metadata_journal(object);
CREATE INDEX metadata_journal_predicate_idx ON metadata_journal(predicate);



/* permissions, roles */
CREATE SEQUENCE permission_ids;
CREATE TABLE permissions (
	id 				INT NOT NULL DEFAULT nextval('permission_ids'),
	name 			CHAR(255) NOT NULL,
	comment 			TEXT,
	PRIMARY KEY(id)
);

/* object column has format type::identifier, e.g.:
   wiklet::at.srfg.ikewiki.render.wiklet.SparqlWiklet
   resource::http://ikewiki.srfg.at/Internal#FrontPage
   namespace::http://ikewiki.srfg.at/Internal#
   
   Idea for code: we need a method resolveObject(Class x) which returns the object identifier
*/   
   

CREATE SEQUENCE role_permission_ids;
CREATE TABLE role_permissions (
	id 				INT NOT NULL DEFAULT nextval('role_permission_ids'),
	role			INT NOT NULL REFERENCES roles(id) ON DELETE CASCADE,
	object  		CHAR(255) NOT NULL,
	permission		INT NOT NULL REFERENCES permissions(id) ON DELETE CASCADE,
	PRIMARY KEY(id)
);	

CREATE SEQUENCE user_permission_ids;
CREATE TABLE user_permissions (
	id 				INT NOT NULL DEFAULT nextval('role_permission_ids'),
	user			INT NOT NULL REFERENCES userss(id) ON DELETE CASCADE,
	object  		CHAR(255) NOT NULL,
	permission		INT NOT NULL REFERENCES permissions(id) ON DELETE CASCADE,
	PRIMARY KEY(id)
);	


CREATE_VIEW all_user_permissions AS (
	( SELECT login,object,permission 
	  FROM user_permissions, users 
	  WHERE user_permissions.user = users.id )
	UNION
	( SELECT login,object,permission
	  FROM role_permissions,roles,user_roles,users
	  WHERE role_permissions.role = roles.id
	    AND user_roles.role = roles.id
	    AND user_roles.user = users.id
	 )
);	 


CREATE USER ontowiki WITH PASSWORD 'ontowiki';
GRANT ALL ON DATABASE ontowiki TO ontowiki;
GRANT ALL ON current_page, page_revisions, revisions, pages, texts, namespaces, revision_ids, page_ids, text_ids, uris, uri_ids, namespace_ids TO ontowiki;
GRANT ALL ON users, user_ids, image_ids, images, current_image, variable_ids, variables TO ontowiki;
/* GRANT ALL ON triple_ids, triples, triples_internal TO ontowiki; */
